/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef __FRAME_H__
#define __FRAME_H__

#include <stdint.h>

#include <bundle.h>
#include <frame_types.h>

#ifdef __cplusplus
extern "C" {
#endif

/**
 * @brief The frame handle.
 * @since_tizen 5.5
 */
typedef struct frame_s *frame_h;

/**
 * @brief The native buffer handle.
 * @since_tizen 5.5
 */
typedef void *frame_native_buffer_h;

/**
 * @brief Enumeration for the frame type.
 * @since_tizen 5.5
 */
typedef enum {
	FRAME_TYPE_REMOTE_SURFACE_NATIVE_BUFFER,        /**< The remote surface: The native buffer */
	FRAME_TYPE_REMOTE_SURFACE_IMAGE_FILE,           /**< The remote surface: The image file */
	FRAME_TYPE_SPLASH_SCREEN_IMAGE,                 /**< The splash screen: The image */
	FRAME_TYPE_SPLASH_SCREEN_EDJE,                  /**< The splash screen: The EDJE */
} frame_type_e;

/**
 * @brief Enumeration for the direction of the frame.
 * @since_tizen 5.5
 */
typedef enum {
	FRAME_DIRECTION_FORWARD,  /**< The direction that is from the caller to the other application. */
	FRAME_DIRECTION_BACKWARD, /**< The direction that is from the other application to the caller. */
} frame_direction_e;

/**
 * @brief Gets the native buffer.
 * @since_tizen 5.5
 * @detils If the type of the frame is not FRAME_TYPE_REMOTE_SURFACE_NATIVE_BUFFER,
 *         this function returns a negative error value that is FRAME_BROKER_ERROR_INVALID_PARAMETER.
 * @remarks The @a buffer should not be released.
 *
 * @param[in]   handle          The frame handle
 * @param[out]  buffer          The native buffer
 * @return      @c 0 on success,
 *              otherwise a negative error value
 *
 * @retval #FRAME_BROKER_ERROR_NONE Successful
 * @retval #FRAME_BROKER_ERROR_INVALID_PARAMETER Invalid parameter
 *
 * @see frame_native_buffer_get_raw()
 * @see frame_native_buffer_get_width()
 * @see frame_native_buffer_get_height()
 * @see frame_native_buffer_get_bbp()
 */
int frame_get_native_buffer(frame_h handle, frame_native_buffer_h *buffer);

/**
 * @brief Gets the image file information.
 * @since_tizen 5.5
 * @detils If the type of the frame is not FRAME_TYPE_REMOTE_SURFACE_IMAGE_FILE,
 *         this function returns a negative error value that is FRAME_BROKER_ERROR_INVALID_PARAMETER.
 * @remarks The @a fd should not be released using close().
 *
 * @param[in]   handle          The frame handle
 * @param[out]  fd              The file descriptor
 * @param[out]  size            The size of the frame buffer
 * @return      @c 0 on success,
 *              otherwise a negative error value
 *
 * @retval #FRAME_BROKER_ERROR_NONE Successful
 * @retval #FRAME_BROKER_ERROR_INVALID_PARAMETER Invalid parameter
 */
int frame_get_image_file(frame_h handle,
		int32_t *fd, uint32_t* size);

/**
 * @brief Gets the file path of the splash screen.
 * @since_tizen 5.5
 * @detils To use this function, the type of the frame has to be FRAME_TYPE_SPLASH_SCREEN_IMAGE or FRAME_TYPE_SPLASH_SCREEN_EDJE.
 *         If it's not, this function returns a negative error value that is FRAME_BROKER_ERROR_INVALID_PARAMETER.
 * @remarks The @a file_path should be released using free()
 *
 * @param[in]   handle          The frame handle
 * @param[out]  file_path       The file path
 * @return      @c 0 on success,
 *              otherwise a negative error value
 *
 * @retval #FRAME_BROKER_ERROR_NONE Successful
 * @retval #FRAME_BROKER_ERROR_INVALID_PARAMETER Invalid parameter
 * @retval #FRAME_BROKER_ERROR_OUT_OF_MEMORY Out of memory
 */
int frame_get_file_path(frame_h handle,
		char **file_path);

/**
 * @brief Gets the file group of the splash screen.
 * @since_tizen 5.5
 * @detils To use this function, the type of the frame has to be FRAME_TYPE_SPLASH_SCREEN_EDJE.
 *         If it's not, this function returns a negative error value that is FRAME_BROKER_ERROR_INVALID_PARAMETER.
 * @remarks The @a file_group should be released using free()
 *
 * @param[in]   handle          The frame handle
 * @param[out]  file_group      The file group
 * @return      @c 0 on success,
 *              otherwise a negative error value
 *
 * @retval #FRAME_BROKER_ERROR_NONE Successful
 * @retval #FRAME_BROKER_ERROR_INVALID_PARAMETER Invalid parameter
 * @retval #FRAME_BROKER_ERROR_OUT_OF_MEMORY Out of memory
 */
int frame_get_file_group(frame_h handle,
		char **file_group);

/**
 * @brief Gets the type of the frame.
 * @since_tizen 5.5
 *
 * @param[in]   handle          The frame handle
 * @param[out]  type            The type
 * @return      @c 0 on success,
 *              otherwise a negative error value
 *
 * @retval #FRAME_BROKER_ERROR_NONE Successful
 * @retval #FRAME_BROKER_ERROR_INVALID_PARAMETER Invalid parameter
 *
 * @see frame_type_e
 */
int frame_get_type(frame_h handle, frame_type_e *type);

/**
 * @brief Gets the direction of the frame.
 * @since_tizen 5.5
 *
 * @param[in]   handle          The frame handle
 * @param[out]  direction       The direction
 * @return      @c 0 on success,
 *              otherwise a negative error value
 *
 * @retval #FRAME_BROKER_ERROR_NONE Successful
 * @retval #FRAME_BROKER_ERROR_INVALID_PARAMETER Invalid parameter
 *
 * @see frame_direction_e
 */
int frame_get_direction(frame_h handle, frame_direction_e *direction);

/**
 * @brief Gets the position X of the frame.
 * @since_tizen 5.5
 *
 * @param[in]   handle          The frame handle
 * @param[out]  x               The position X
 * @return      @c 0 on success,
 *              otherwise a negative error valuea
 *
 * @retval #FRAME_BROKER_ERROR_NONE Successful
 * @retval #FRAME_BROKER_ERROR_INVALID_PARAMETER Invalid parameter
 */
int frame_get_position_x(frame_h handle, int32_t *x);

/**
 * @brief Gets the position Y of the frame.
 * @since_tizen 5.5
 *
 * @param[in]   handle          The frame handle
 * @param[out]  y               The position Y
 * @return      @c 0 on success,
 *              otherwise a negative error value
 *
 * @retval #FRAME_BROKER_ERROR_NONE Successful
 * @retval #FRAME_BROKER_ERROR_INVALID_PARAMETER Invalid parameter
 */
int frame_get_position_y(frame_h handle, int32_t *y);

/**
 * @brief Gets the extra data of the frame.
 * @since_tizen 5.5
 * @remarks The @a extra_data should not be released.
 *
 * @param[in]   handle          The frame handle
 * @param[out]  extra_data      The bundle object
 * @return      @c 0 on success,
 *              otherwise a negative error value
 *
 * @retval #FRAME_BROKER_ERROR_NONE Successful
 * @retval #FRAME_BROKER_ERROR_INVALID_PARAMETER Invalid parameter
 */
int frame_get_extra_data(frame_h handle, bundle **extra_data);

/**
 * @brief Gets the raw data of the native buffer.
 * @since_tizen 5.5
 *
 * @param[in]   handle          The native buffer handle
 * @param[out]  raw             The raw data
 * @return      @c 0 on success,
 *              otherwise a negative error value
 *
 * @retval #FRAME_BROKER_ERROR_NONE Successful
 * @retval #FRAME_BROKER_ERROR_INVALID_PARAMETER Invalid parameter
 */
int frame_native_buffer_get_raw(frame_native_buffer_h handle, void **raw);

/**
 * @brief Gets the width of the native buffer.
 * @since_tizen 5.5
 *
 * @param[in]   handle          The native buffer handle
 * @param[out]  width           The width of the native buffer
 * @return      @c 0 on success,
 *              otherwise a negative error value
 *
 * @retval #FRAME_BROKER_ERROR_NONE Successful
 * @retval #FRAME_BROKER_ERROR_INVALID_PARAMETER Invalid parameter
 */
int frame_native_buffer_get_width(frame_native_buffer_h handle,
		uint32_t *width);

/**
 * @brief Gets the height of the native buffer.
 * @since_tizen 5.5
 *
 * @param[in]   handle          The native buffer handle
 * @param[out]  height          The height of the native buffer
 * @return      @c 0 on success,
 *              otherwise a negative error value
 *
 * @retval #FRAME_BROKER_ERROR_NONE Successful
 * @retval #FRAME_BROKER_ERROR_INVALID_PARAMETER Invalid parameter
 */
int frame_native_buffer_get_height(frame_native_buffer_h handle,
		uint32_t *height);
/**
 * @brief Gets the bpp of the native buffer.
 * @since_tizen 5.5
 *
 * @param[in]   handle          The native buffer handle
 * @param[out]  bpp             The bpp of the native buffer
 * @return      @c 0 on success,
 *              otherwise a negative error value
 *
 * @retval #FRAME_BROKER_ERROR_NONE Successful
 * @retval #FRAME_BROKER_ERROR_INVALID_PARAMETER Invalid parameter
 */
int frame_native_buffer_get_bpp(frame_native_buffer_h handle,
		uint32_t *bpp);

#ifdef __cplusplus
}
#endif

#endif /* __FRAME_H__ */
