/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef __FRAME_BROKER_H__
#define __FRAME_BROKER_H__

#include <Ecore_Wl2.h>
#include <app_control.h>
#include <frame.h>
#include <frame_context.h>
#include <frame_types.h>
#include <tizen.h>

#ifdef __cplusplus
extern "C" {
#endif

/**
 * @brief The frame broker handle.
 * @since_tizen 5.5
 */
typedef struct frame_broker_s *frame_broker_h;

/**
 * @brief Creates the frame broker handle.
 * @since_tizen 5.5
 * @remarks The @a handle should be release using frame_broker_destroy() function.
 *
 * @param[in]   wl2_win         The Ecore Wl2 Window
 * @param[in]   callback        The set of callback functions to handle frame context lifecycle events
 * @param[in]   user_data       The user data to be passed to the callback functions
 * @param[out]  handle          The frame broker handle
 * @return      @c 0 on success,
 *              otherwise a negative error value
 *
 * @retval #FRAME_BROKER_ERROR_NONE Successful
 * @retval #FRAME_BROKER_ERROR_INVALID_PARAMETER Invalid parameter
 * @retval #FRAME_BROKER_ERROR_OUT_OF_MEMORY Out of memory
 * @retval #FRAME_BROKER_ERROR_IO_ERROR I/O error
 *
 * @remarks This function is only available for platform level signed applications.
 *
 * @see frame_broker_destroy()
 */
int frame_broker_create(Ecore_Wl2_Window *wl2_win,
		frame_context_lifecycle_callback_s *callback,
		void *user_data,
		frame_broker_h *handle);

/**
 * @brief Destroyes the frame broker handle.
 * @since_tizen 5.5
 *
 * @param[in]   handle          The frame broker handle
 * @return      @c 0 on success,
 *              otherwise a negative error value
 *
 * @retval #FRAME_BROKER_ERROR_NONE Successful
 * @retval #FRAME_BROKER_ERROR_INVALID_PARAMETER Invalid parameter
 */
int frame_broker_destroy(frame_broker_h handle);

/**
 * @brief Sends the launch request asynchronously.
 * @since_tizen 5.5
 *
 * @privlevel public
 * @privilege %http:://tizen.org/privilege/appmanager.launch
 *
 * @param[in]   handle          The frame broker handle
 * @param[in]   app_control     The app_control handle
 * @param[in]   result_cb       The callback function to be called when the result is delivered
 * @param[in]   reply_cb        The callback function to be called when the reply is delivered
 * @param[in]   uset_data       The user data to be passed to the callback function
 * @return      @c 0 on success,
 *              otherwise a negative error value
 *
 * @retval #FRAME_BROKER_ERROR_NONE Successful
 * @retval #FRAME_BROKER_ERROR_APP_NOT_FOUND The application to run the given launch request is not found
 * @retval #FRAME_BROKER_ERROR_INVALID_PARAMETER Invalid parameter
 * @retval #FRAME_BROKER_ERROR_IO_ERROR I/O error
 * @retval #FRAME_BROKER_ERROR_LAUNCH_REJECTED Failed to launch the application
 * @retval #FRAME_BROKER_ERROR_PERMISSION_DENIED Permission denied
 *
 * @see app_control_send_launch_request_async()
 * @see app_control_result_cb()
 * @see app_control_reply_cb()
 */
int frame_broker_send_launch_request(frame_broker_h handle,
		app_control_h app_control,
		app_control_result_cb result_cb,
		app_control_reply_cb reply_cb,
		void *user_data);

/**
 * @brief Sends the launch request to the provider asynchronously.
 * @since_tizen 5.5
 *
 * @privlevel public
 * @privilege %http:://tizen.org/privilege/appmanager.launch
 *
 * @param[in]   handle          The frame broker handle
 * @param[in]   app_control     The app_control handle
 * @param[in]   result_cb       The callback function to be called when the result is delivered
 * @param[in]   reply_cb        The callback function to be called when the reply is delivered
 * @param[in]   uset_data       The user data to be passed to the callback function
 * @return      @c 0 on success,
 *              otherwise a negative error value
 *
 * @retval #FRAME_BROKER_ERROR_NONE Successful
 * @retval #FRAME_BROKER_ERROR_APP_NOT_FOUND The application to run the given launch request is not found
 * @retval #FRAME_BROKER_ERROR_INVALID_PARAMETER Invalid parameter
 * @retval #FRAME_BROKER_ERROR_IO_ERROR I/O error
 * @retval #FRAME_BROKER_ERROR_LAUNCH_REJECTED Failed to launch the application
 * @retval #FRAME_BROKER_ERROR_PERMISSION_DENIED Permission denied
 *
 * @see app_control_send_launch_request_async()
 * @see app_control_result_cb()
 * @see app_control_reply_cb()
 */
int frame_broker_send_launch_request_to_provider(frame_broker_h handle,
		app_control_h app_control,
		app_control_result_cb result_cb,
		app_control_reply_cb reply_cb,
		void *user_data);

#ifdef __cplusplus
}
#endif

#endif /* __FRAME_BROKER_H__ */
