/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef __FRAME_CONTEXT_H__
#define __FRAME_CONTEXT_H__

#include <frame.h>
#include <frame_types.h>

#ifdef __cplusplus
extern "C" {
#endif

/**
 * @brief The frame context handle.
 * @since_tizen 5.5
 */
typedef struct frame_context_s *frame_context_h;

/**
 * @brief Enumeration for the frame context errors.
 * @since_tizen 5.5
 */
typedef enum {
	FRAME_CONTEXT_ERROR_NONE,          /**< Successful */
	FRAME_CONTEXT_ERROR_DISQUALIFIED,  /**< Disqualified */
	FRAME_CONTEXT_ERROR_WRONG_REQUEST, /**< Wrong request */
} frame_context_error_e;

/**
 * @brief Called when the frame context is created.
 * @since_tizen 5.5
 * @remarks The @a context should not be released. The @a context can be used only in the callback.
 *
 * @param[in]   context         The frame context handle
 * @param[in]   user_data       The user data passed from frame_broker_create() function
 */
typedef void (*frame_context_create_cb)(frame_context_h context,
		void *user_data);

/**
 * @brief Called when the frame context is resumed.
 * @since_tizen 5.5
 * @details When the frame has been prepared, this callback function is called.
 *          The caller can start animations. To notify that the animation is started, the caller should call frame_context_start_animation() function.
 *          After the animation is finished, the caller should call frame_context_finish_animation() function to notify.
 * @remarks The @a context and the @a frame should not be released. The @a context and the @a frame can be used only in the callback.
 *
 * @param[in]   context         The frame context handle
 * @param[in]   frame           The frame handle
 * @param[in]   user_data       The user data passed from frame_broker_create() function
 *
 * @see frame_context_start_animation()
 * @see frame_context_finish_animation()
 */
typedef void (*frame_context_resume_cb)(frame_context_h context,
		frame_h frame,
		void *user_data);
/**
 * @brief Called when the frame context is paused.
 * @since_tizen 5.5
 * @details When the animation has been finished, this callback function is called.
 * @remarks The @a context should not be released. The @a context can be used only in the callback.
 *
 * @param[in]   context         The frame context handle
 * @param[in]   user_data       The user data passed from frame_broker_create() function
 */
typedef void (*frame_context_pause_cb)(frame_context_h context,
		void *user_data);

/**
 * @brief Called when the frame context is destroyed.
 * @since_tizen 5.5
 * @remarks The @a context should not be released. The @a context can be used only in the callback.
 *
 * @param[in]   context         The frame context handle
 * @param[in]   user_data       The user data passed from frame_broker_create() function
 */
typedef void (*frame_context_destroy_cb)(frame_context_h context,
		void *user_data);

/**
 * @brief Called when the error is occurred.
 * @since_tizen 5.5
 * @remarks The @a context should not be released. The @a context can be used only in the callback.
 *
 * @param[in]   context         The frame context handle
 * @param[in]   error           The error
 * @param[in]   user_data       The user data passed from frame_broker_create() function
 *
 * @see frame_context_error_e
 */
typedef void (*frame_context_error_cb)(frame_context_h context,
		frame_context_error_e error,
		void *user_data);

/**
 * @brief Called when the frame context is updated.
 * @since_tizen 5.5
 * @details When the frame has been updated, this callback function is called.
 * @remarks The @a context and the @a frame should not be released. The @a context and the @a frame can be used only in the callback.
 *
 * @param[in]   context         The frame context handle
 * @param[in]   frame           The frame handle
 * @param[in]   user_data       The user data passed from frame_broker_create() function
 *
 * @see frame_context_resume_cb()
 * @see frame_context_start_animation()
 * @see frame_context_finish_animation()
 */
typedef void (*frame_context_update_cb)(frame_context_h context,
		frame_h frame,
		void *user_data);

/**
 * @brief The structure type containing the set of callback functions for lifecycle of a frame context.
 * @since_tizen 5.5
 *
 * @see frame_broker_create()
 * @see frame_context_create_cb()
 * @see frame_context_resume_cb()
 * @see frame_context_pause_cb()
 * @see frame_context_destroy_cb()
 * @see frame_context_error_cb()
 */
typedef struct {
	frame_context_create_cb create;   /**< The callback function called after the frame context is created. */
	frame_context_resume_cb resume;   /**< The callback function called when the frame is prepared. */
	frame_context_pause_cb pause;     /**< The callback function called when the animation is finished. */
	frame_context_destroy_cb destroy; /**< The callback function called before the frame context is destroyed. */
	frame_context_update_cb update;   /**< The callback function called when the frame is updated. */
	frame_context_error_cb error;     /**< The callback function called when the error is occurred. */
} frame_context_lifecycle_callback_s;

/**
 * @brief Notifies that the animation is started.
 * @since_tizen 5.5
 *
 * @param[in]   handle          The frame context handle
 * @return      @c 0 on success,
 *              otherwise a negative error value
 *
 * @retval #FRAME_BROKER_ERROR_NONE Successful
 * @retval #FRAME_BROKER_ERROR_INVALID_PARAMETER Invalid parameter
 * @retval #FRAME_BROKER_ERROR_IO_ERROR I/O error
 *
 * @see frame_context_finish_animation()
 */
int frame_context_start_animation(frame_context_h handle);

/**
 * @brief Notifies that the animation is finished.
 * @since_tizen 5.5
 *
 * @param[in]   handle          The frame context handle
 * @return      @c 0 on success,
 *              otherwise a negative error value
 *
 * @retval #FRAME_BROKER_ERROR_NONE Successful
 * @retval #FRAME_BROKER_ERROR_INVALID_PARAMETER Invalid parameter
 * @retval #FRAME_BROKER_ERROR_IO_ERROR I/O error
 *
 * @see frame_context_start_animation()
 */
int frame_context_finish_animation(frame_context_h handle);

#ifdef __cplusplus
}
#endif

#endif /* __FRAME_CONTEXT_H__ */
