/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <screen_connector_launcher_service.h>

#include "frame.h"
#include "frame_private.h"
#include "log_private.h"
#include "private.h"

struct frame_s {
	screen_connector_launcher_service_image_h image;
	bundle *extra_data;
};

static frame_type_e __convert_type(
		screen_connector_launcher_service_image_type_e type)
{
	switch (type) {
	case SCREEN_CONNECTOR_LAUNCHER_SERVICE_IMAGE_TYPE_REMOTE_SURFACE_TBM:
		return FRAME_TYPE_REMOTE_SURFACE_NATIVE_BUFFER;
	case SCREEN_CONNECTOR_LAUNCHER_SERVICE_IMAGE_TYPE_REMOTE_SURFACE_IMAGE_FILE:
		return FRAME_TYPE_REMOTE_SURFACE_IMAGE_FILE;
	case SCREEN_CONNECTOR_LAUNCHER_SERVICE_IMAGE_TYPE_SPLASH_SCREEN_IMAGE:
		return FRAME_TYPE_SPLASH_SCREEN_IMAGE;
	default:
		return FRAME_TYPE_SPLASH_SCREEN_EDJE;
	}
}

static frame_direction_e __convert_direction(
		screen_connector_launcher_service_direction_e direction)
{
	if (direction == SCREEN_CONNECTOR_LAUNCHER_SERVICE_DIRECTION_FORWARD)
		return FRAME_DIRECTION_FORWARD;

	return FRAME_DIRECTION_BACKWARD;
}

int frame_create(screen_connector_launcher_service_image_h image,
		frame_h *handle)
{
	struct frame_s *frame;
	char *shared_widget_info = NULL;

	if (!image || !handle) {
		_E("Invalid parameter");
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;
	}

	frame = calloc(1, sizeof(struct frame_s));
	if (!frame) {
		_E("Out of memory");
		return FRAME_BROKER_ERROR_OUT_OF_MEMORY;
	}

	screen_connector_launcher_service_image_get_shared_widget_info(image,
			&shared_widget_info);
	_D("shared_widget_info(%s)", shared_widget_info);
	if (shared_widget_info && shared_widget_info[0] != '\0' &&
			strcmp(shared_widget_info, "None") != 0) {
		frame->extra_data = bundle_decode(
				(bundle_raw *)shared_widget_info,
				strlen(shared_widget_info));
	} else {
		frame->extra_data = bundle_create();
	}
	free(shared_widget_info);

	if (!frame->extra_data) {
		_E("Failed to create extra data");
		frame_destroy(frame);
		return FRAME_BROKER_ERROR_OUT_OF_MEMORY;
	}

	frame->image = image;

	*handle = frame;

	return FRAME_BROKER_ERROR_NONE;
}

int frame_destroy(frame_h handle)
{
	if (!handle) {
		_E("Invalid parameter");
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;
	}

	bundle_free(handle->extra_data);
	free(handle);

	return FRAME_BROKER_ERROR_NONE;
}

API int frame_get_native_buffer(frame_h handle, frame_native_buffer_h *buffer)
{
	tbm_surface_info_s *surface_info = NULL;
	frame_type_e type;
	int ret;

	if (!handle || !buffer) {
		_E("Invalid parameter");
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;
	}

	frame_get_type(handle, &type);
	if (type != FRAME_TYPE_REMOTE_SURFACE_NATIVE_BUFFER) {
		_E("Invalid parameter. type(%d)", type);
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;
	}

	ret = screen_connector_launcher_service_image_get_tbm_surface_info(
			handle->image, &surface_info);
	if (ret != 0) {
		_E("Failed to get tbm surface info. error(%d)", ret);
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;
	}

	*buffer = (frame_native_buffer_h)surface_info;

	return ret;
}

API int frame_get_image_file(frame_h handle,
		int32_t *fd, uint32_t *size)
{
	frame_type_e type;
	int ret;

	if (!handle || !fd || !size) {
		_E("Invalid parameter");
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;
	}

	frame_get_type(handle, &type);
	if (type != FRAME_TYPE_REMOTE_SURFACE_IMAGE_FILE) {
		_E("Invalid parameter. type(%d)", type);
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;
	}

	ret = screen_connector_launcher_service_image_get_image_file(
			handle->image, fd, size);
	return ret;
}

API int frame_get_file_path(frame_h handle, char **file_path)
{
	frame_type_e type;
	int ret;

	if (!handle || !file_path) {
		_E("Invalid parameter");
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;
	}

	frame_get_type(handle, &type);
	if (type != FRAME_TYPE_SPLASH_SCREEN_IMAGE &&
			type != FRAME_TYPE_SPLASH_SCREEN_EDJE) {
		_E("Invalid parameter. type(%d)", type);
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;
	}

	ret = screen_connector_launcher_service_image_get_file_path(
			handle->image, file_path);
	return ret;
}

API int frame_get_file_group(frame_h handle, char **file_group)
{
	frame_type_e type;
	int ret;

	if (!handle || !file_group) {
		_E("Invalid parameter");
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;
	}

	frame_get_type(handle, &type);
	if (type != FRAME_TYPE_SPLASH_SCREEN_EDJE) {
		_E("Invalid parameter. type(%d)", type);
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;
	}

	ret = screen_connector_launcher_service_image_get_file_group(
			handle->image, file_group);
	return ret;
}

API int frame_get_type(frame_h handle, frame_type_e *type)
{
	screen_connector_launcher_service_image_type_e image_type;
	int ret;

	if (!handle || !type) {
		_E("Invalid parameter");
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;
	}

	ret = screen_connector_launcher_service_image_get_type(handle->image,
			&image_type);
	if (ret != 0)
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;

	*type = __convert_type(image_type);

	return FRAME_BROKER_ERROR_NONE;
}

API int frame_get_direction(frame_h handle, frame_direction_e *direction)
{
	screen_connector_launcher_service_direction_e scls_direction;
	int ret;

	if (!handle || !direction) {
		_E("Invalid parameter");
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;
	}

	ret = screen_connector_launcher_service_image_get_direction(
			handle->image, &scls_direction);
	if (ret != 0)
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;

	*direction = __convert_direction(scls_direction);

	return FRAME_BROKER_ERROR_NONE;
}

API int frame_get_position_x(frame_h handle, int32_t *x)
{
	int ret;

	if (!handle || !x) {
		_E("Invalid parameter");
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;
	}

	ret = screen_connector_launcher_service_image_get_position_x(
			handle->image, x);
	if (ret != 0)
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;

	return FRAME_BROKER_ERROR_NONE;
}

API int frame_get_position_y(frame_h handle, int32_t *y)
{
	int ret;

	if (!handle || !y) {
		_E("Invalid parameter");
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;
	}

	ret = screen_connector_launcher_service_image_get_position_y(
			handle->image, y);
	if (ret != 0)
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;

	return FRAME_BROKER_ERROR_NONE;
}

API int frame_get_extra_data(frame_h handle, bundle **extra_data)
{
	if (!handle || !extra_data) {
		_E("Invalid parameter");
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;
	}

	*extra_data = handle->extra_data;

	return FRAME_BROKER_ERROR_NONE;
}

API int frame_native_buffer_get_raw(frame_native_buffer_h handle,
		void **raw)
{
	tbm_surface_info_s *info;

	if (!handle || !raw) {
		_E("Invalid parameter");
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;
	}

	info = (tbm_surface_info_s *)handle;
	*raw = info->planes[0].ptr;

	return FRAME_BROKER_ERROR_NONE;
}

API int frame_native_buffer_get_width(frame_native_buffer_h handle,
		uint32_t *width)
{
	tbm_surface_info_s *info;

	if (!handle || !width) {
		_E("Invalid parameter");
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;
	}

	info = (tbm_surface_info_s *)handle;
	*width = info->width;

	return FRAME_BROKER_ERROR_NONE;
}

API int frame_native_buffer_get_height(frame_native_buffer_h handle,
		uint32_t *height)
{
	tbm_surface_info_s *info;

	if (!handle || !height) {
		_E("Invalid parameter");
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;
	}

	info = (tbm_surface_info_s *)handle;
	*height = info->height;

	return FRAME_BROKER_ERROR_NONE;
}

API int frame_native_buffer_get_bpp(frame_native_buffer_h handle,
		uint32_t *bpp)
{
	tbm_surface_info_s *info;

	if (!handle || !bpp) {
		_E("Invalid parameter");
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;
	}

	info = (tbm_surface_info_s *)handle;
	*bpp = info->bpp;

	return FRAME_BROKER_ERROR_NONE;
}
