/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <glib.h>
#include <app_control_internal.h>
#include <aul_launcher_service.h>
#include <aul_running_context.h>
#include <screen_connector_launcher_service.h>
#include <uuid/uuid.h>

#include "frame_broker.h"
#include "frame_broker_private.h"
#include "frame_context_private.h"
#include "frame_private.h"
#include "log_private.h"
#include "private.h"

struct frame_broker_s {
	char *name;
	Ecore_Wl2_Window *wl2_win;
	frame_context_lifecycle_callback_s callback;
	void *user_data;
	aul_launcher_service_h als;
	screen_connector_launcher_service_h scls;
	frame_context_h context;
	frame_context_h pre_context;
};

static GList *__win_list;

static gint __compare_win(gconstpointer a, gconstpointer b)
{
	Ecore_Wl2_Window *win_a = (Ecore_Wl2_Window *)a;
	Ecore_Wl2_Window *win_b = (Ecore_Wl2_Window *)b;

	return win_a != win_b;
}

static bool __exist_win(Ecore_Wl2_Window *win)
{
	GList *found;

	found = g_list_find_custom(__win_list, win, __compare_win);
	if (found)
		true;

	return false;
}

static void __add_win(Ecore_Wl2_Window *win)
{
	__win_list = g_list_append(__win_list, win);
}

static void __remove_win(Ecore_Wl2_Window *win)
{
	__win_list = g_list_remove(__win_list, win);
}

static char *__generate_uuid(void)
{
	char uuid[37];
	uuid_t u;

	uuid_generate(u);
	uuid_unparse(u, uuid);
	return strdup(uuid);
}

static uint32_t __generate_serial(void)
{
	static gint serial;

	g_atomic_int_inc(&serial);

	return (uint32_t)serial;
}

static void __check_pre_context(frame_broker_h broker)
{
	if (broker->context) {
		_D("Destroy previous context");
		frame_context_destroy(broker->context);
	}

	broker->context = broker->pre_context;
	broker->pre_context = NULL;

	frame_context_on_create(broker->context);
}

static void __aul_launcher_service_cb(const char *app_id,
		const char *inst_id,
		const int pid,
		const uint32_t serial,
		void *user_data)
{
	frame_broker_h broker = user_data;
	frame_context_h context;
	int ctx_pid = -1;
	uint32_t ctx_serial = 0;
	int ret;

	_D("app_id(%s), inst_id(%s), pid(%d), serial(%u)",
			app_id, inst_id, pid, serial);

	if (broker->pre_context) {
		__check_pre_context(broker);
		return;
	}

	if (broker->context) {
		frame_context_get_pid(broker->context, &ctx_pid);
		frame_context_get_serial(broker->context, &ctx_serial);
		if (ctx_pid == pid && ctx_serial == serial) {
			_D("Frame context already exists");
			return;
		}
	}

	ret = frame_context_create(broker, app_id, inst_id, pid, serial,
			&broker->callback, broker->user_data,
			&context);
	if (ret != FRAME_BROKER_ERROR_NONE) {
		_E("Failed to create frame context. error(%d)", ret);
		return;
	}

	if (broker->context) {
		_D("Destroy previous context");
		frame_context_destroy(broker->context);
	}

	broker->context = context;
	frame_context_on_create(context);
}

static void __scls_prepare_cb(screen_connector_launcher_service_image_h image,
		uint32_t serial,
		void *user_data)
{
	frame_broker_h broker = user_data;
	frame_context_h context = broker->context;
	frame_h frame;
	int ret;

	_D("[__SCLS__] Prepare");
	if (context == NULL) {
		if (broker->pre_context) {
			__check_pre_context(broker);
		} else {
			_E("Invalid context");
			return;
		}
	}

	frame_context_set_serial(context, serial);

	ret = frame_create(image, &frame);
	if (ret != FRAME_BROKER_ERROR_NONE) {
		_E("Failed to create frame. error(%d)", ret);
		frame_context_on_error(context,
				FRAME_CONTEXT_ERROR_WRONG_REQUEST);
		return;
	}

	frame_context_on_resume(context, frame);
}

static void __scls_stop_cb(uint32_t serial, void *user_data)
{
	frame_broker_h broker = user_data;
	frame_context_h context = broker->context;

	_D("[__SCLS__] Stop");
	if (context == NULL) {
		if (broker->pre_context) {
			__check_pre_context(broker);
		} else {
			_E("Invalid context");
			return;
		}
	}

	_D("Destroy context");
	frame_context_destroy(context);
}

static frame_context_error_e __convert_error(
		screen_connector_launcher_service_error_e error)
{
	switch (error) {
	case SCREEN_CONNECTOR_LAUNCHER_SERVICE_ERROR_DISQUALIFIED:
		return FRAME_CONTEXT_ERROR_DISQUALIFIED;
	case SCREEN_CONNECTOR_LAUNCHER_SERVICE_ERROR_WRONG_REQUEST:
		return FRAME_CONTEXT_ERROR_WRONG_REQUEST;
	default:
		return FRAME_CONTEXT_ERROR_NONE;
	}
}

static void __scls_error_cb(screen_connector_launcher_service_error_e error,
		uint32_t serial,
		void *user_data)
{
	frame_broker_h broker = user_data;
	frame_context_h context = broker->context;

	_D("[__SCLS__] Error");
	if (context == NULL) {
		if (broker->pre_context) {
			__check_pre_context(broker);
		} else {
			_E("Invalid context");
			return;
		}
	}

	frame_context_on_error(context, __convert_error(error));
}

static void __scls_cleanup_cb(uint32_t serial, void *user_data)
{
	frame_broker_h broker = user_data;
	frame_context_h context = broker->context;
	frame_direction_e direction;
	frame_h frame;
	int ret;

	_D("[__SCLS__] Clean up");
	if (context == NULL) {
		_E("Invalid context");
		return;
	}

	frame_context_on_pause(context);

	ret = frame_context_get_frame(context, &frame);
	if (ret != FRAME_BROKER_ERROR_NONE) {
		_W("Failed to get frame");
		return;
	}

	ret = frame_get_direction(frame, &direction);
	if ((ret == FRAME_BROKER_ERROR_NONE &&
				direction == FRAME_DIRECTION_BACKWARD) ||
			frame == NULL) {
		_D("Destroy context");
		frame_context_destroy(context);
	}
}

static void __scls_reset_cb(uint32_t serial, void *user_data)
{
	frame_broker_h broker = user_data;
	frame_context_h context = broker->context;

	_W("[__SCLS__] Reset");
	frame_context_set_frame(context, NULL);
}

API int frame_broker_create(Ecore_Wl2_Window *wl2_win,
		frame_context_lifecycle_callback_s *callback,
		void *user_data,
		frame_broker_h *handle)
{
	screen_connector_launcher_service_h scls;
	screen_connector_launcher_service_ops ops = {
		.prepare = __scls_prepare_cb,
		.stop = __scls_stop_cb,
		.error = __scls_error_cb,
		.cleanup = __scls_cleanup_cb,
		.reset = __scls_reset_cb
	};
	struct frame_broker_s *broker;
	int ret;

	if (!wl2_win || !callback || !handle) {
		_E("Invalid parameter");
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;
	}

	if (!callback->create || !callback->resume ||
			!callback->pause || !callback->destroy ||
			!callback->update || !callback->error) {
		_E("Invalid parameter");
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;
	}

	if (__exist_win(wl2_win)) {
		_E("Already exists. wl2_win(%p)", wl2_win);
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;
	}

	broker = calloc(1, sizeof(struct frame_broker_s));
	if (!broker) {
		_E("Out of memory");
		return FRAME_BROKER_ERROR_OUT_OF_MEMORY;
	}

	broker->name = __generate_uuid();
	if (!broker->name) {
		_E("Failed to generate uuid");
		frame_broker_destroy(broker);
		return FRAME_BROKER_ERROR_OUT_OF_MEMORY;
	}

	ret = aul_launcher_service_create(broker->name,
			__aul_launcher_service_cb,
			broker, &broker->als);
	if (ret != AUL_R_OK) {
		_E("Failed to create aul launcher service. error(%d)", ret);
		frame_broker_destroy(broker);
		return FRAME_BROKER_ERROR_OUT_OF_MEMORY;
	}

	ret = aul_launcher_service_listen(broker->als);
	if (ret != AUL_R_OK) {
		_E("Failed to listen aul launcher service. error(%d)", ret);
		frame_broker_destroy(broker);
		return FRAME_BROKER_ERROR_IO_ERROR;
	}

	ret = screen_connector_launcher_service_create(wl2_win, &scls);
	if (ret != 0) {
		_E("Failed to create scls evas. error(%d)", ret);
		frame_broker_destroy(broker);
		return FRAME_BROKER_ERROR_IO_ERROR;
	}

	screen_connector_launcher_service_set_ops(scls, &ops, broker);

	broker->scls = scls;
	broker->wl2_win = wl2_win;
	broker->callback = *callback;
	broker->user_data = user_data;

	__add_win(wl2_win);

	*handle = broker;

	return FRAME_BROKER_ERROR_NONE;
}

API int frame_broker_destroy(frame_broker_h handle)
{
	if (!handle) {
		_E("Invalid parameter");
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;
	}

	__remove_win(handle->wl2_win);

	if (handle->context)
		frame_context_destroy(handle->context);

	screen_connector_launcher_service_destroy(handle->scls);
	aul_launcher_service_destroy(handle->als);
	free(handle->name);
	free(handle);

	return FRAME_BROKER_ERROR_NONE;
}

static int __send_launch_request(frame_broker_h handle,
		app_control_h app_control,
		app_control_result_cb result_cb,
		app_control_reply_cb reply_cb,
		void *user_data,
		bool shared_widget)
{
	aul_running_context_h context = NULL;
	char buf[12];
	uint32_t serial;
	char *app_id;
	char *inst_id = NULL;
	char *comp_id = NULL;
	int pid = -1;
	int ret;

	if (!handle || !app_control || !result_cb) {
		_E("Invalid parameter");
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;
	}

	ret = app_control_get_app_id(app_control, &app_id);
	if (ret != APP_CONTROL_ERROR_NONE) {
		_E("Failed to get app id. error(%d)", ret);
		return ret;
	}

	app_control_get_component_id(app_control, &comp_id);
	app_control_get_instance_id(app_control, &inst_id);

	ret = aul_running_context_create(app_id, comp_id, inst_id, &context);
	if (ret != AUL_R_OK)
		_W("Failed to create running context. error(%d)", ret);

	free(comp_id);
	free(inst_id);
	inst_id = NULL;

	aul_running_context_get_inst_id(context, (const char **)&inst_id);
	aul_running_context_get_pid(context, &pid);

	if (shared_widget) {
		ret = frame_broker_launch_with_shared_widget(handle, app_id,
				inst_id, pid, &serial);
	} else {
		ret = frame_broker_launch(handle, app_id, inst_id, pid,
				&serial);
	}

	if (inst_id && pid > 0) {
		frame_context_destroy(handle->pre_context);
		handle->pre_context = NULL;

		frame_context_create(handle, app_id, inst_id, pid, serial,
				&handle->callback, handle->user_data,
				&handle->pre_context);
	}

	aul_running_context_destroy(context);
	free(app_id);
	if (ret != FRAME_BROKER_ERROR_NONE) {
		frame_context_destroy(handle->pre_context);
		handle->pre_context = NULL;
		return ret;
	}

	snprintf(buf, sizeof(buf), "%u", serial);
	ret = app_control_add_extra_data(app_control,
			AUL_K_LAUNCHER_SERVICE_SERIAL, buf);
	if (ret != APP_CONTROL_ERROR_NONE) {
		_E("Failed to add extra data. error(%d)", ret);
		frame_context_destroy(handle->pre_context);
		handle->pre_context = NULL;
		return ret;
	}

	ret = app_control_add_extra_data(app_control,
			AUL_K_LAUNCHER_SERVICE, handle->name);
	if (ret != APP_CONTROL_ERROR_NONE) {
		_E("Failed to add extra data. error(%d)", ret);
		frame_context_destroy(handle->pre_context);
		handle->pre_context = NULL;
		return ret;
	}

	ret = app_control_send_launch_request_async(app_control,
			result_cb, reply_cb, user_data);
	if (ret != APP_CONTROL_ERROR_NONE) {
		_E("Failed to send launch request. error(%d)", ret);
		frame_context_destroy(handle->pre_context);
		handle->pre_context = NULL;
		return ret;
	}

	return FRAME_BROKER_ERROR_NONE;
}

API int frame_broker_send_launch_request(frame_broker_h handle,
		app_control_h app_control,
		app_control_result_cb result_cb,
		app_control_reply_cb reply_cb,
		void *user_data)
{
	return __send_launch_request(handle, app_control, result_cb, reply_cb,
			user_data, false);
}

API int frame_broker_send_launch_request_to_provider(
		frame_broker_h handle,
		app_control_h app_control,
		app_control_result_cb result_cb,
		app_control_reply_cb reply_cb,
		void *user_data)
{
	return __send_launch_request(handle, app_control, result_cb, reply_cb,
			user_data, true);
}

void frame_broker_set_frame_context(frame_broker_h handle,
		frame_context_h context)
{
	if (!handle) {
		_E("Invalid parameter");
		return;
	}

	handle->context = context;
}

int frame_broker_launch(frame_broker_h handle, const char *app_id,
		const char *inst_id, int pid, uint32_t *serial)
{
	int ret;

	if (!handle || !app_id || pid < -1 || !serial) {
		_E("Invalid parameter");
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;
	}

	*serial = __generate_serial();
	ret = screen_connector_launcher_service_launch(handle->scls,
			app_id, inst_id ? inst_id : "", pid, *serial);
	if (ret != 0) {
		_E("Failed to launch scls evas. error(%d)", ret);
		return FRAME_BROKER_ERROR_IO_ERROR;
	}

	return FRAME_BROKER_ERROR_NONE;
}

int frame_broker_launch_with_shared_widget(frame_broker_h handle,
                const char *app_id, const char *inst_id, int pid,
		uint32_t *serial)
{
	int ret;

	if (!handle || !app_id ||  pid < -1 || !serial) {
		_E("Invalid parameter");
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;
	}

	*serial = __generate_serial();
	ret = screen_connector_launcher_service_launch_with_shared_widget(
			handle->scls, app_id,
			inst_id ? inst_id : "", pid, *serial);
	if (ret != 0) {
		_E("Failed to launch scls evas with shared widget. error(%d)",
				ret);
		return FRAME_BROKER_ERROR_IO_ERROR;
	}

	return FRAME_BROKER_ERROR_NONE;
}

int frame_broker_launching(frame_broker_h handle, uint32_t serial)
{
	int ret;

	if (!handle || !serial) {
		_E("Invalid parameter");
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;
	}

	ret = screen_connector_launcher_service_launching(
			handle->scls, serial);
	if (ret != 0) {
		_E("Failed to notify that the animation has been started");
		return FRAME_BROKER_ERROR_IO_ERROR;
	}

	return FRAME_BROKER_ERROR_NONE;
}

int frame_broker_launch_cancel(frame_broker_h handle, uint32_t serial)
{
	int ret;

	if (!handle || !serial) {
		_E("Invalid parameter");
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;
	}

	ret = screen_connector_launcher_service_launch_cancel(
			handle->scls, serial);
	if (ret != 0) {
		_E("Failed to notify that the animation has been canceled.");
		return FRAME_BROKER_ERROR_IO_ERROR;
	}

	return FRAME_BROKER_ERROR_NONE;
}

int frame_broker_launch_done(frame_broker_h handle, uint32_t serial)
{
	int ret;

	if (!handle || !serial) {
		_E("Invalid parameter");
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;
	}

	ret = screen_connector_launcher_service_launch_done(
			handle->scls, serial);
	if (ret != 0) {
		_E("Failed to notify that the animation has been finished.");
		return FRAME_BROKER_ERROR_IO_ERROR;
	}

	return FRAME_BROKER_ERROR_NONE;
}
