/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <glib.h>

#include "frame_broker.h"
#include "frame_broker_private.h"
#include "frame_context.h"
#include "frame_context_private.h"
#include "frame_private.h"
#include "log_private.h"
#include "private.h"

struct frame_context_s {
	frame_broker_h broker;
	frame_context_state_e state;
	frame_context_lifecycle_callback_s callback;
	void *user_data;
	guint idle_tag;
	char *app_id;
	char *inst_id;
	int pid;
	bool started;
	uint32_t serial;
	frame_h frame;
};

int frame_context_create(frame_broker_h broker,
		const char *app_id,
		const char *inst_id,
		int pid,
		uint32_t serial,
		frame_context_lifecycle_callback_s *callback,
		void *user_data,
		frame_context_h *handle)
{
	struct frame_context_s *context;

	if (!broker || !app_id || !inst_id || pid < 1 || !callback || !handle) {
		_E("Invalid parameter");
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;
	}

	context = calloc(1, sizeof(struct frame_context_s));
	if (!context) {
		_E("Out of memory");
		return FRAME_BROKER_ERROR_OUT_OF_MEMORY;
	}

	context->app_id = strdup(app_id);
	if (!context->app_id) {
		_E("Failed to duplicate application ID");
		frame_context_destroy(context);
		return FRAME_BROKER_ERROR_OUT_OF_MEMORY;
	}

	context->inst_id = strdup(inst_id);
	if (!context->inst_id) {
		_E("Failed to duplicate instance ID");
		frame_context_destroy(context);
		return FRAME_BROKER_ERROR_OUT_OF_MEMORY;
	}

	context->broker = broker;
	context->state = FRAME_CONTEXT_STATE_NONE;
	context->pid = pid;
	context->serial = serial;
	context->callback = *callback;
	context->user_data = user_data;

	*handle = context;

	return FRAME_BROKER_ERROR_NONE;
}

int frame_context_destroy(frame_context_h handle)
{
	if (!handle) {
		_E("Invalid parameter");
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;
	}

	frame_broker_set_frame_context(handle->broker, NULL);

	if (handle->idle_tag)
		g_source_remove(handle->idle_tag);

	frame_context_on_pause(handle);
	frame_context_on_destroy(handle);

	if (handle->frame)
		frame_destroy(handle->frame);

	free(handle->inst_id);
	free(handle->app_id);
	free(handle);

	return FRAME_BROKER_ERROR_NONE;
}

API int frame_context_start_animation(frame_context_h handle)
{
	int ret;

	if (!handle || !handle->serial) {
		_E("Invalid parameter");
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;
	}

	if (handle->started) {
		_W("Already started");
		return FRAME_BROKER_ERROR_NONE;
	}

	ret = frame_broker_launching(handle->broker, handle->serial);
	if (ret != 0) {
		_E("Failed to notify that the animation is started. error(%d)",
				ret);
		return FRAME_BROKER_ERROR_IO_ERROR;
	}

	handle->started = true;

	return FRAME_BROKER_ERROR_NONE;
}

static gboolean __context_destroy_cb(gpointer data)
{
	frame_context_h context = data;

	context->idle_tag = 0;
	frame_context_destroy(context);

	return G_SOURCE_REMOVE;
}

static gboolean __context_pause_cb(gpointer data)
{
	frame_context_h context = data;
	frame_direction_e direction;

	context->idle_tag = 0;
	frame_context_on_pause(context);

	_W("frame(%p)", context->frame);
	frame_get_direction(context->frame, &direction);
	if (direction == FRAME_DIRECTION_BACKWARD) {
		_D("frame_context_destroy()");
		frame_context_destroy(context);
	}

	return G_SOURCE_REMOVE;
}

API int frame_context_finish_animation(frame_context_h handle)
{
	int ret = 0;

	if (!handle || !handle->serial) {
		_E("Invalid parameter");
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;
	}

	if (handle->started) {
		ret = frame_broker_launch_done(handle->broker, handle->serial);
		handle->started = false;
		if (!handle->idle_tag) {
			handle->idle_tag = g_idle_add(__context_pause_cb,
					handle);
		}
	} else {
		ret = frame_broker_launch_cancel(handle->broker,
				handle->serial);
		if (!handle->idle_tag) {
			handle->idle_tag = g_idle_add(__context_destroy_cb,
					handle);
		}
	}

	if (ret != 0) {
		_E("Failed to notify that the animation is finished. error(%d)",
				ret);
		return FRAME_BROKER_ERROR_IO_ERROR;
	}

	return FRAME_BROKER_ERROR_NONE;
}

int frame_context_get_pid(frame_context_h handle, int *pid)
{
	if (!handle || !pid) {
		_E("Invalid parameter");
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;
	}

	*pid = handle->pid;

	return FRAME_BROKER_ERROR_NONE;
}

int frame_context_set_serial(frame_context_h handle, uint32_t serial)
{
	if (!handle) {
		_E("Invalid parameter");
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;
	}

	_D("serial(%u) is changed to %u", handle->serial, serial);
	handle->serial = serial;

	return FRAME_BROKER_ERROR_NONE;
}

int frame_context_get_serial(frame_context_h handle, uint32_t *serial)
{
	if (!handle || !serial) {
		_E("Invalid parameter");
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;
	}

	*serial = handle->serial;

	return FRAME_BROKER_ERROR_NONE;
}

int frame_context_set_frame(frame_context_h handle, frame_h frame)
{
	if (!handle) {
		_E("Invalid parameter");
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;
	}

	if (handle->frame)
		frame_destroy(handle->frame);

	_W("prev frame(%p), new frame(%p)", handle->frame, frame);

	handle->frame = frame;

	return FRAME_BROKER_ERROR_NONE;
}

int frame_context_get_frame(frame_context_h handle, frame_h *frame)
{
	if (!handle || !frame) {
		_E("Invalid parameter");
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;
	}

	*frame = handle->frame;

	return FRAME_BROKER_ERROR_NONE;
}

int frame_context_set_state(frame_context_h handle,
                frame_context_state_e state)
{
	if (!handle) {
		_E("Invalid parameter");
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;
	}

	handle->state = state;

	return FRAME_BROKER_ERROR_NONE;
}

int frame_context_get_state(frame_context_h handle,
                frame_context_state_e *state)
{
	if (!handle || !state) {
		_E("Invalid parameter");
		return FRAME_BROKER_ERROR_INVALID_PARAMETER;
	}

	*state = handle->state;

	return FRAME_BROKER_ERROR_NONE;
}

void frame_context_on_create(frame_context_h handle)
{
	if (!handle) {
		_E("Invalid parameter");
		return;
	}

	if (handle->state != FRAME_CONTEXT_STATE_NONE)
		return;

	handle->callback.create(handle, handle->user_data);
	handle->state = FRAME_CONTEXT_STATE_CREATED;
}

void frame_context_on_resume(frame_context_h handle, frame_h frame)
{
	if (!handle) {
		_E("Invalid parameter");
		return;
	}

	frame_context_set_frame(handle, frame);

	if (handle->state == FRAME_CONTEXT_STATE_RESUMED) {
		handle->callback.update(handle, frame, handle->user_data);
		return;
	}

	if (handle->state != FRAME_CONTEXT_STATE_CREATED &&
			handle->state != FRAME_CONTEXT_STATE_PAUSED)
		return;

	handle->callback.resume(handle, frame, handle->user_data);
	handle->state = FRAME_CONTEXT_STATE_RESUMED;
}

void frame_context_on_pause(frame_context_h handle)
{
	if (!handle) {
		_E("Invalid parameter");
		return;
	}

	if (handle->state != FRAME_CONTEXT_STATE_RESUMED)
		return;

	handle->callback.pause(handle, handle->user_data);
	handle->state = FRAME_CONTEXT_STATE_PAUSED;
}

void frame_context_on_destroy(frame_context_h handle)
{
	if (!handle) {
		_E("Invalid parameter");
		return;
	}

	if (handle->state == FRAME_CONTEXT_STATE_NONE)
		return;

	handle->callback.destroy(handle, handle->user_data);
	handle->state = FRAME_CONTEXT_STATE_DESTROYED;
}

void frame_context_on_error(frame_context_h handle, frame_context_error_e error)
{
	if (!handle) {
		_E("Invalid parameter");
		return;
	}

	handle->callback.error(handle, error, handle->user_data);
}
