/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef __FRAME_PROVIDER_H__
#define __FRAME_PROVIDER_H__

#include <stdint.h>

#include <Ecore_Wl2.h>
#include <bundle.h>
#include <frame_provider_types.h>

#ifdef __cplusplus
extern "C" {
#endif

/**
 * @brief The frame provider handle.
 * @since_tizen 5.5
 */
typedef struct frame_provider_s *frame_provider_h;

/**
 * @brief Called when the show event is occurred.
 * @since_tizen 5.5
 * @details When the provider should show the object to share with the caller, this callback function is called.
 *          After showing the object, the caller should call frame_provider_notify_show_status() function to notify.
 * @remarks The @a handle should not be released.
 *
 * @param[in]   handle          The frame provider handle
 * @param[in]   user_data       The user data passed from frame_provider_create() function
 *
 * @see frame_provider_create()
 * @see frame_provider_notify_show_status()
 */
typedef void (*frame_provider_show_cb)(frame_provider_h handle,
		void *user_data);

/**
 * @brief Called when the hide event is occurred.
 * @since_tizen 5.5
 * @details When the provider should hide the object to share with the caller, this callback function is called.
 *          After hiding the object, the caller should call frame_provider_notify_hide_status() function to notify.
 * @remarks The @a handle should not be released.
 *
 * @param[in]   handle          The frame provider handle
 * @param[in]   user_data       The user data passed from frame_provider_create() function
 *
 * @see frame_provider_create()
 * @see frame_provider_notify_hide_status()
 */
typedef void (*frame_provider_hide_cb)(frame_provider_h handle,
		void *user_data);

/**
 * @brief The structure type containing the set of callback functions for events of a frame provider.
 * @since_tizen 5.5
 *
 * @see frame_provider_create()
 * @see frame_provider_show_cb()
 * @see frame_provider_hide_cb()
 */
typedef struct {
	frame_provider_show_cb show; /**< The callback function called when the show event is occurred. */
	frame_provider_hide_cb hide; /**< The callback function called when the hide event is occurred. */
} frame_provider_event_callback_s;

/**
 * @brief Creates the frame provider handle.
 * @since_tizen 5.5
 * @remarks The @a handle should be released using frame_provider_destroy().
 *
 * @param[in]   wl2_win         The Ecore Wl2 Window
 * @param[in]   callback        The set of callback functions to handle frame provider events
 * @param[in]   user_data       The user data to be passed to the callback functions
 * @param[out]  handle          The frame provider handle
 * @return      @c 0 on success,
 *              otherwise a negative error value
 *
 * @retval #FRAME_ERROR_NONE Successful
 * @retval #FRAME_ERROR_INVALID_PARAMETER Invalid parameter
 * @retval #FRAME_ERROR_OUT_OF_MEMORY Out of memory
 *
 * @see frame_provider_destroy()
 * @see frame_provider_event_callback_s
 */
int frame_provider_create(Ecore_Wl2_Window *wl2_win,
		frame_provider_event_callback_s *callback,
		void *user_data,
		frame_provider_h *handle);

/**
 * @brief Destroyes the frame provider handle.
 * @since_tizen 5.5
 *
 * @param[in]   handle          The frame provider handle
 * @return      @c 0 on success,
 *              otherwise a negative error value
 *
 * @retval #FRAME_ERROR_NONE Successful
 * @retval #FRAME_ERROR_INVALID_PARAMETER Invalid parameter
 */
int frame_provider_destroy(frame_provider_h handle);

/**
 * @brief Notifies that the object is prepared to show.
 * @since_tizen 5.5
 *
 * @param[in]   handle          The frame provider handle
 * @param[in]   extra_data      The bundle object
 * @return      @c 0 on success,
 *              otherwise a negative error value
 *
 * @retval #FRAME_ERROR_NONE Successful
 * @retval #FRAME_ERROR_INVALID_PARAMETER Invalid parameter
 * @retval #FRAME_ERROR_INVALID_OPERATION Invalid operation
 * @retval #FRAME_ERROR_IO_ERROR I/O error
 */
int frame_provider_notify_show_status(frame_provider_h handle,
		bundle *extra_data);

/**
 * @brief Notifies that the object is prepared to hide.
 * @since_tizen 5.5
 *
 * @param[in]   handle          The frame provider handle
 * @param[in]   extra_data      The bundle object
 * @return      @c 0 on success,
 *              otherwise a negative error value
 *
 * @retval #FRAME_ERROR_NONE Successful
 * @retval #FRAME_ERROR_INVALID_PARAMETER Invalid parameter
 * @retval #FRAME_ERROR_INVALID_OPERATION Invalid operation
 * @retval #FRAME_ERROR_IO_ERROR I/O error
 */
int frame_provider_notify_hide_status(frame_provider_h handle,
		bundle *extra_data);

#ifdef __cplusplus
}
#endif

#endif /* __FRAME_PROVIDER_H__ */
