/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <screen_connector_shared_widget_launch.h>
#include <Ecore_Wl2.h>

#include "frame_provider.h"
#include "frame_provider_types.h"
#include "log_private.h"
#include "private.h"

struct frame_provider_s {
	Ecore_Wl2_Window *wl2_win;
	screen_connector_shared_widget_launch_h scswl;
	screen_connector_shared_widget_launch_prepare_state_e state;
	uint32_t serial;
	frame_provider_event_callback_s callback;
	void *user_data;
};

static GList *__win_list;

static gint __compare_win(gconstpointer a, gconstpointer b)
{
	Ecore_Wl2_Window *win_a = (Ecore_Wl2_Window *)a;
	Ecore_Wl2_Window *win_b = (Ecore_Wl2_Window *)b;

	return win_a != win_b;
}

static bool __exist_win(Ecore_Wl2_Window *win)
{
	GList *found;

	found = g_list_find_custom(__win_list, win, __compare_win);
	if (found)
		true;

	return false;
}

static void __add_win(Ecore_Wl2_Window *win)
{
	__win_list = g_list_append(__win_list, win);
}

static void __remove_win(Ecore_Wl2_Window *win)
{
	__win_list = g_list_remove(__win_list, win);
}

static void __scswl_prepare_cb(
		screen_connector_shared_widget_launch_prepare_state_e state,
		uint32_t serial,
		void *user_data)
{
	frame_provider_h provider = user_data;

	provider->state = state;
	provider->serial = serial;

	if (state == SCREEN_CONNECTOR_SHARED_WIDGET_LAUNCH_PREPARE_STATE_WIDGET_SHOW)
		provider->callback.show(provider, provider->user_data);
	else
		provider->callback.hide(provider, provider->user_data);
}

API int frame_provider_create(Ecore_Wl2_Window *wl2_win,
		frame_provider_event_callback_s *callback,
		void *user_data,
		frame_provider_h *handle)
{
	struct frame_provider_s *provider;
	int ret;

	if (!wl2_win || !callback || !handle) {
		_E("Invalid parameter");
		return FRAME_PROVIDER_ERROR_INVALID_PARAMETER;
	}

	if (__exist_win(wl2_win)) {
		_E("Already exists. wl2_win(%p)", wl2_win);
		return FRAME_PROVIDER_ERROR_INVALID_PARAMETER;
	}

	provider = calloc(1, sizeof(struct frame_provider_s));
	if (!provider) {
		_E("Out of memory");
		return FRAME_PROVIDER_ERROR_OUT_OF_MEMORY;
	}

	ret = screen_connector_shared_widget_launch_create(wl2_win,
			&provider->scswl);
	if (ret != 0) {
		_E("Failed to create scswl. error(%d)", ret);
		frame_provider_destroy(provider);
		return FRAME_PROVIDER_ERROR_OUT_OF_MEMORY;
	}

	screen_connector_shared_widget_launch_prepare_cb_set(provider->scswl,
			__scswl_prepare_cb, provider);

	provider->wl2_win = wl2_win;
	provider->callback = *callback;
	provider->user_data = user_data;

	__add_win(wl2_win);

	*handle = provider;

	return FRAME_PROVIDER_ERROR_NONE;
}

API int frame_provider_destroy(frame_provider_h handle)
{
	if (!handle) {
		_E("Invalid parameter");
		return FRAME_PROVIDER_ERROR_NONE;
	}

	__remove_win(handle->wl2_win);

	screen_connector_shared_widget_launch_destroy(handle->scswl);
	free(handle);

	return FRAME_PROVIDER_ERROR_NONE;
}

static int __notify_status(frame_provider_h handle,
		screen_connector_shared_widget_launch_prepare_state_e state,
		bundle *extra_data)
{
	bundle_raw *raw;
	int len;
	int ret;

	if (!handle || !handle->serial) {
		_E("Invalid parameter");
		return FRAME_PROVIDER_ERROR_INVALID_PARAMETER;
	}

	if (handle->state != state) {
		_E("Invalid operation");
		return FRAME_PROVIDER_ERROR_INVALID_OPERATION;
	}

	if (extra_data) {
		ret = bundle_encode(extra_data, &raw, &len);
		if (ret != BUNDLE_ERROR_NONE) {
			_E("Failed to encode extra data. error(%d)", ret);
			if (ret == BUNDLE_ERROR_OUT_OF_MEMORY)
				return FRAME_PROVIDER_ERROR_OUT_OF_MEMORY;

			return FRAME_PROVIDER_ERROR_INVALID_PARAMETER;
		}
	} else {
		raw = (bundle_raw *)strdup("None");
	}

	ret = screen_connector_shared_widget_launch_prepare_done(handle->scswl,
			(const char *)raw, state, handle->serial);
	free((void *)raw);
	if (ret != 0) {
		_E("Failed to notify provider status. error(%d)", ret);
		return FRAME_PROVIDER_ERROR_IO_ERROR;
	}

	return FRAME_PROVIDER_ERROR_NONE;
}

API int frame_provider_notify_show_status(frame_provider_h handle,
		bundle *extra_data)
{
	return __notify_status(handle,
		SCREEN_CONNECTOR_SHARED_WIDGET_LAUNCH_PREPARE_STATE_WIDGET_SHOW,
		extra_data);
}

API int frame_provider_notify_hide_status(frame_provider_h handle,
		bundle *extra_data)
{
	return __notify_status(handle,
		SCREEN_CONNECTOR_SHARED_WIDGET_LAUNCH_PREPARE_STATE_WIDGET_HIDE,
		extra_data);
}
