/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef MOCK_GDBUS_H_
#define MOCK_GDBUS_H_

#include "mock.h"
#include <glib.h>

#ifdef __cplusplus
extern "C" {
#endif

typedef void GDBusConnection;
typedef void GCancellable;
typedef enum
{
	G_BUS_TYPE_STARTER = -1,
	G_BUS_TYPE_NONE = 0,
	G_BUS_TYPE_SYSTEM  = 1,
	G_BUS_TYPE_SESSION = 2
} GBusType;
typedef struct _GError GError;
typedef char gchar;
typedef short gshort;
typedef long glong;
typedef int gint;
typedef gint gboolean;
typedef unsigned char guchar;
typedef unsigned short gushort;
typedef unsigned long gulong;
typedef unsigned int guint;

typedef unsigned short guint16;
typedef unsigned int guint32;

typedef void* gpointer;
typedef const void *gconstpointer;
typedef guint32 GQuark;
typedef int GDBusMessage;


/**
 * GBusNameOwnerFlags:
 * @G_BUS_NAME_OWNER_FLAGS_NONE: No flags set.
 * @G_BUS_NAME_OWNER_FLAGS_ALLOW_REPLACEMENT: Allow another message bus connection to claim the name.
 * @G_BUS_NAME_OWNER_FLAGS_REPLACE: If another message bus connection owns the name and have
 * specified #G_BUS_NAME_OWNER_FLAGS_ALLOW_REPLACEMENT, then take the name from the other connection.
 * @G_BUS_NAME_OWNER_FLAGS_DO_NOT_QUEUE: Do not place message bus connection in a queue to own the name.
 *
 * Flags used in g_bus_own_name().
 *
 * Since: 2.26
 */
typedef enum
{
	G_BUS_NAME_OWNER_FLAGS_NONE = 0,                    /*< nick=none >*/
	G_BUS_NAME_OWNER_FLAGS_ALLOW_REPLACEMENT = (1<<0),  /*< nick=allow-replacement >*/
	G_BUS_NAME_OWNER_FLAGS_REPLACE = (1<<1),            /*< nick=replace >*/
	G_BUS_NAME_OWNER_FLAGS_DO_NOT_QUEUE = (1<<2)        /*< nick=do-not-queue >*/
} GBusNameOwnerFlags;
typedef enum
{
	G_BUS_NAME_WATCHER_FLAGS_NONE = 0,
	G_BUS_NAME_WATCHER_FLAGS_AUTO_START = (1<<0)
} GBusNameWatcherFlags;

typedef enum
{
	G_DBUS_SEND_MESSAGE_FLAGS_NONE = 0,
	G_DBUS_SEND_MESSAGE_FLAGS_PRESERVE_SERIAL = (1<<0)
} GDBusSendMessageFlags;

typedef enum /*< flags >*/
{
	G_DBUS_SIGNAL_FLAGS_NONE = 0,
	G_DBUS_SIGNAL_FLAGS_NO_MATCH_RULE = (1<<0),
	G_DBUS_SIGNAL_FLAGS_MATCH_ARG0_NAMESPACE = (1<<1),
	G_DBUS_SIGNAL_FLAGS_MATCH_ARG0_PATH = (1<<2)
} GDBusSignalFlags;

typedef void (*GBusNameAcquiredCallback) (GDBusConnection *connection,
		const gchar     *name,
		gpointer         user_data);

typedef void (*GBusNameLostCallback) (GDBusConnection *connection,
		const gchar     *name,
		gpointer         user_data);

typedef void (*GBusNameAppearedCallback) (GDBusConnection *connection,
		const gchar     *name,
		const gchar     *name_owner,
		gpointer         user_data);

typedef void (*GBusNameVanishedCallback) (GDBusConnection *connection,
		const gchar     *name,
		gpointer         user_data);
typedef void            (*GDestroyNotify)       (gpointer       data);
typedef void (*GDBusSignalCallback) (GDBusConnection  *connection,
		const gchar      *sender_name,
		const gchar      *object_path,
		const gchar      *interface_name,
		const gchar      *signal_name,
		GVariant         *parameters,
		gpointer          user_data);

DECLARE_FAKE_VALUE_FUNC(GDBusConnection *, g_bus_get_sync, GBusType,
		GCancellable *, GError **);
DECLARE_FAKE_VALUE_FUNC(guint, g_bus_own_name_on_connection,
		GDBusConnection *, const gchar *, GBusNameOwnerFlags,
		GBusNameAcquiredCallback, GBusNameLostCallback, gpointer,
		GDestroyNotify);
DECLARE_FAKE_VALUE_FUNC(gboolean, g_dbus_connection_emit_signal,
		GDBusConnection *, const gchar *, const gchar *,
		const gchar *, const gchar *, GVariant *, GError **);
DECLARE_FAKE_VALUE_FUNC(guint, g_bus_watch_name,
		GBusType, const gchar *, GBusNameWatcherFlags,
		GBusNameAppearedCallback, GBusNameVanishedCallback, gpointer,
		GDestroyNotify);

DECLARE_FAKE_VALUE_FUNC(guint, g_bus_watch_name_on_connection,
		GDBusConnection *, const gchar *, GBusNameWatcherFlags,
		GBusNameAppearedCallback, GBusNameVanishedCallback, gpointer,
		GDestroyNotify);

DECLARE_FAKE_VALUE_FUNC(GDBusMessage *, g_dbus_message_new_method_call,
		const gchar *, const gchar *, const gchar *, const gchar *);

DECLARE_FAKE_VOID_FUNC(g_dbus_message_set_body, GDBusMessage  *,
		GVariant *);

DECLARE_FAKE_VALUE_FUNC(GDBusMessage *, g_dbus_connection_send_message_with_reply_sync,
		GDBusConnection *, GDBusMessage *, GDBusSendMessageFlags, gint,
		volatile guint32 *, GCancellable *, GError **);
DECLARE_FAKE_VOID_FUNC(g_object_unref, gpointer);
DECLARE_FAKE_VALUE_FUNC(GVariant *, g_dbus_message_get_body, GDBusMessage *);
DECLARE_FAKE_VOID_FUNC(g_dbus_connection_signal_unsubscribe, GDBusConnection *,
		guint);

DECLARE_FAKE_VOID_FUNC(g_bus_unwatch_name, guint);

DECLARE_FAKE_VALUE_FUNC(guint, g_dbus_connection_signal_subscribe,
		GDBusConnection *, const gchar *, const gchar *,
		const gchar *, const gchar *, const gchar *,
		GDBusSignalFlags, GDBusSignalCallback,
		gpointer, GDestroyNotify);

#ifdef __cplusplus
}
#endif
#endif  /* MOCK_GDBUS_H_ */
