/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <glib.h>
#include <gtest/gtest.h>
#include <gmock/gmock.h>
#include <widget_viewer_evas.h>
#include <widget_viewer_evas_extension.h>
#include <widget_errno.h>

#include "unittest/mock/gio/gio.h"
#include "unittest/mock/cynara_mock.h"
#include "unittest/mock/evas_mock.h"
#include "unittest/mock/system_info_mock.h"
#include "unittest/mock/aul_mock.h"
#include "unittest/mock/widget_service_mock.h"
#include "unittest/mock/screen_connector_mock.h"

int __fake_system_info_get_platform_bool(const char* key, bool* value) {
  *value = true;
  return 0;
}

int __cynara_check(cynara* c, const char* client, const char* client_session,
    const char* user,
    const char* privilege) {
  return CYNARA_API_ACCESS_ALLOWED;
}

Evas_Object* __elm_win_add(Evas_Object* parent, const char *name,
    Elm_Win_Type type) {
  return (Evas_Object*)g_hash_table_new_full(
      g_str_hash, g_str_equal, free, nullptr);
}

int __aul_app_get_appid_bypid(int pid, char *appid, int len) {
  snprintf(appid, len, "%s", "org.tizen.test");
  return 0;
}

int __widget_service_get_widget_max_count(const char *widget_id) {
  return 5;
}

int __widget_instance_create(const char *widget_id, char **instance_id) {
  *instance_id = strdup("test@org.tizen.test");
  return 0;
}

Eina_Bool __elm_layout_file_set(Eo *obj, const char *file, const char *group) {
  return EINA_TRUE;
}

void *__evas_object_data_get(const Evas_Object *eo_obj, const char *key) {
  GHashTable* tbl = (GHashTable*)eo_obj;
  return g_hash_table_lookup(tbl, key);
}

void __evas_object_data_set(Evas_Object *eo_obj, const char *key,
    const void *data) {
  GHashTable* tbl = (GHashTable*)eo_obj;
  g_hash_table_insert(tbl, strdup(key), (gpointer)data);
}

void __evas_object_event_callback_add(Evas_Object *eo_obj,
    Evas_Callback_Type type, Evas_Object_Event_Cb func, const void *data) {
}

void __elm_object_signal_callback_add(Evas_Object *obj, const char *emission,
    const char *source, Edje_Signal_Cb func, void *data) {
}

Evas_Object *__elm_layout_add(Evas_Object *parent) {
  return (Evas_Object*)g_hash_table_new_full(
      g_str_hash, g_str_equal, free, nullptr);
}

void __elm_object_part_content_set(Evas_Object *obj, const char *part,
    Evas_Object *content) {
}

Evas_Object *__elm_object_part_content_get(const Evas_Object *obj,
    const char *part) {
  return nullptr;
}

int __aul_widget_write_log(const char *tag, const char *format, va_list ap) {
  return 0;
}

int __screen_connector_toolkit_evas_init(Evas_Object *win,
                                   screen_connector_screen_type_e type) {
  return 0;
}

namespace {
class WVE : public ::testing::Test {
 public:
  virtual void SetUp() {
    system_info_get_platform_bool_fake.custom_fake =
        __fake_system_info_get_platform_bool;
    cynara_check_fake.custom_fake = __cynara_check;
    elm_win_add_fake.custom_fake = __elm_win_add;
    aul_app_get_appid_bypid_fake.custom_fake = __aul_app_get_appid_bypid;
    widget_service_get_widget_max_count_fake.custom_fake =
        __widget_service_get_widget_max_count;
    widget_instance_create_fake.custom_fake = __widget_instance_create;

    elm_layout_add_fake.custom_fake = __elm_layout_add;
    elm_layout_file_set_fake.custom_fake = __elm_layout_file_set;
    evas_object_data_set_fake.custom_fake = __evas_object_data_set;
    evas_object_data_get_fake.custom_fake = __evas_object_data_get;
    evas_object_event_callback_add_fake.custom_fake =
        __evas_object_event_callback_add;
    elm_object_signal_callback_add_fake.custom_fake =
        __elm_object_signal_callback_add;
    elm_object_part_content_set_fake.custom_fake =
        __elm_object_part_content_set;
    elm_object_part_content_get_fake.custom_fake =
        __elm_object_part_content_get;

    screen_connector_toolkit_evas_init_fake.custom_fake =
        __screen_connector_toolkit_evas_init;
    aul_widget_write_log_fake.custom_fake = __aul_widget_write_log;
  }
  virtual void TearDown(){
  }
};

TEST_F(WVE, widget_viewer_evas_init_invalid)
{
  int ret = widget_viewer_evas_init(NULL);
  EXPECT_EQ(WIDGET_ERROR_INVALID_PARAMETER, ret);
}

TEST_F(WVE, widget_viewer_evas_init_fault)
{
  int ret = widget_viewer_evas_fini();
  EXPECT_EQ(WIDGET_ERROR_FAULT, ret);
}

TEST_F(WVE, widget_viewer_evas_init_fini)
{
  Evas_Object* win = elm_win_add(NULL, "Widget Viewer", ELM_WIN_BASIC);
  widget_viewer_evas_init(win);
  int ret = widget_viewer_evas_fini();
  EXPECT_EQ(WIDGET_ERROR_NONE, ret);
}

TEST_F(WVE, widget_viewer_evas_add_widget)
{
  Evas_Object* win = elm_win_add(NULL, "Widget Viewer", ELM_WIN_BASIC);
  widget_viewer_evas_init(win);
  Evas_Object* widget = widget_viewer_evas_add_widget(win,
      "org.tizen.widget", "content", 0.5);
  int ret = widget_viewer_evas_pause_widget(widget);
  EXPECT_EQ(WIDGET_ERROR_NONE, ret);

  ret = widget_viewer_evas_resume_widget(widget);
  EXPECT_EQ(WIDGET_ERROR_NONE, ret);
  EXPECT_TRUE(widget != nullptr);

  set_last_result(WIDGET_ERROR_NONE);
  widget_viewer_evas_get_content_info(widget);
  EXPECT_EQ(WIDGET_ERROR_NONE, get_last_result());

  set_last_result(WIDGET_ERROR_NONE);
  widget_viewer_evas_get_title_string(widget);
  EXPECT_EQ(WIDGET_ERROR_NOT_SUPPORTED, get_last_result());

  const char* ret_id = widget_viewer_evas_get_widget_id(widget);
  EXPECT_EQ(0, strcmp(ret_id, "org.tizen.widget"));

  set_last_result(WIDGET_ERROR_NONE);
  widget_viewer_evas_get_period(widget);
  EXPECT_EQ(WIDGET_ERROR_NONE, get_last_result());

  set_last_result(WIDGET_ERROR_NONE);
  widget_viewer_evas_cancel_click_event(widget);
  EXPECT_EQ(WIDGET_ERROR_INVALID_PARAMETER, get_last_result());

  set_last_result(WIDGET_ERROR_NONE);
  ret = widget_viewer_evas_feed_mouse_up_event(widget);
  EXPECT_EQ(WIDGET_ERROR_NONE, get_last_result());

  set_last_result(WIDGET_ERROR_NONE);
  widget_viewer_evas_disable_preview(widget);
  EXPECT_EQ(WIDGET_ERROR_NONE, get_last_result());

  widget_viewer_evas_disable_overlay_text(widget);
  EXPECT_EQ(WIDGET_ERROR_NONE, get_last_result());

  widget_viewer_evas_disable_loading(widget);
  EXPECT_EQ(WIDGET_ERROR_NONE, get_last_result());

  widget_viewer_evas_activate_faulted_widget(widget);
  EXPECT_EQ(WIDGET_ERROR_NONE, get_last_result());

  widget_viewer_evas_is_faulted(widget);
  EXPECT_EQ(WIDGET_ERROR_NONE, get_last_result());

  ret = widget_viewer_evas_freeze_visibility(widget,
      WIDGET_VISIBILITY_STATUS_SHOW_FIXED);
  EXPECT_EQ(WIDGET_ERROR_INVALID_PARAMETER, ret);

  ret = widget_viewer_evas_thaw_visibility(widget);
  EXPECT_EQ(WIDGET_ERROR_INVALID_PARAMETER, ret);

  bool is_frozon = widget_viewer_evas_is_visibility_frozen(widget);
  EXPECT_EQ(is_frozon, false);

  bool is_widget = widget_viewer_evas_is_widget(widget);
  EXPECT_EQ(is_widget, true);

  widget_viewer_evas_set_permanent_delete(widget, 1);
  EXPECT_EQ(WIDGET_ERROR_NONE, get_last_result());

  const char *ret_inst_id = widget_viewer_evas_get_widget_instance_id(widget);
  EXPECT_EQ(0, strcmp(ret_inst_id, "test@org.tizen.test"));

  ret = widget_viewer_evas_set_delayed_resuming_time(10);
  EXPECT_EQ(WIDGET_ERROR_NONE, ret);

  ret = widget_viewer_evas_launch_setup_app(widget);
  EXPECT_EQ(WIDGET_ERROR_NOT_EXIST, ret);
}

}