/*
 * Samsung API
 * Copyright (c) 2016 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Flora License, Version 1.1 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://floralicense.org/license/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef __WATCH_MANAGER_H__
#define __WATCH_MANAGER_H__

#include <tizen_type.h>
#include <Evas.h>
#include <app.h>

#ifdef __cplusplus
extern "C" {
#endif

typedef enum {
	WATCH_POLICY_HINT_EXPAND = 1,
} watch_policy_size_hint;

#define WATCH_SMART_SIGNAL_ADDED "watch,added"
#define WATCH_SMART_SIGNAL_REMOVED "watch,removed"
#define WATCH_SMART_SIGNAL_UPDATED "watch,updated"

typedef void (*watch_dead_signal_cb)(const char *appid, int pid, bool is_faulted, void *data);
extern int watch_manager_init(Evas_Object *win);
extern int watch_manager_fini();
extern int watch_manager_get_app_control(const char *app_id, app_control_h *app_control);
extern int watch_manager_send_terminate(Evas_Object *watch);
extern int watch_policy_set_size_hint(watch_policy_size_hint hint);
extern int watch_manager_enable_buffer_update(void);
extern int watch_manager_disable_buffer_update(void);
extern int watch_manager_set_resource_id(int resource_id);
extern int watch_manager_get_resource_id(Evas_Object *watch, int *resource_id);
extern int watch_manager_add_dead_signal_listener(watch_dead_signal_cb cb, void *data);
extern int watch_manager_remove_dead_signal_listener(watch_dead_signal_cb cb);

/**
 * @brief Sends request to bind watch's window with the window.
 * @details Watch's window will be shown with the viewer window automatically.
 * @since_tizen 4.0
 * @param[in] win Window for the viewer app
 * @return @c 0 on success, otherwise a negative error value
 */
extern int watch_manager_window_bind_request(Evas_Object *win);

/**
 * @brief Binds watch's window with the window.
 * @details Watch's window will be shown with the viewer window automatically.
 * @since_tizen 4.0
 * @param[in] win Window for the viewer app
 * @return @c 0 on success, otherwise a negative error value
 */
extern int watch_manager_window_bind(Evas_Object *win);

/**
 * @brief Unbinds watch's window
 * @details Watch's window will be updated as an image object.
 * @since_tizen 4.0
 * @return @c 0 on success, otherwise a negative error value
 */
extern int watch_manager_window_unbind(void);

/**
 * @brief Notify watch application that viewer application is paused.
 * @remarks Watch application can not resume until it gets viewer resumed event.
 * @since_tizen 3.0
 * @return @c 0 on success, otherwise a negative error value
 * @see #watch_manager_notify_resumed_status_of_viewer
 */
extern int watch_manager_notify_paused_status_of_viewer(void);

/**
 * @brief Notify watch application that viewer application is resumed.
 * @remarks Watch application can not resume until it gets viewer resumed event.
 * @since_tizen 3.0
 * @return @c 0 on success, otherwise a negative error value
 */
extern int watch_manager_notify_resumed_status_of_viewer(void);

/**
 * @brief Gets watch's setup appid.
 * @since_tizen 4.0
 * @remarks setup_appid should be freed.
 * @param[in] app_id A watch app id
 * @param[out] setup_appid A watch app id
 * @return @c 0 on success, otherwise a negative error value
 */
extern int watch_manager_get_setup_appid(const char *app_id, char **setup_appid);

/**
 * @brief Gets the opr of the Evas Object
 * @since_tizen 3.0
 * @param[out] opr The OPR(On Pixel Ratio)
 * @return @c 0 on success, otherwise a negative error value
 */
extern int watch_manager_get_opr(float *opr);

/**
 * @brief Cancels watch's click event.
 * @details If you call this function after feeding the mouse down event, the watch will get ON_HOLD events.
 * @since_tizen 4.0
 * @param[in] watch A watch object
 * @return @c 0 on success, otherwise a negative error value
 */
extern int watch_manager_cancel_click_event(Evas_Object *watch);

/**
 * @brief Changes the policy for sending visibilities
 * @details If you turn off, the framework will not send visibility events automatically.
 * @since_tizen 4.0
 * @remarks By default, it is enabled
 * @param[in] enable Whether it should be enabled or not
 * @return @c 0 on success, otherwise a negative error value
 */
extern int watch_manager_change_auto_visibility(bool enable);

/**
 * @brief Sends a paused event to the display server.
 * @details It sends a paused event to the display server so that it can pass on the watch application.
 * @since_tizen 4.0
 * @return @c 0 on success, otherwise a negative error value
 */
extern int watch_manager_pause(void);

/**
 * @brief Sends a resumed event to the display server.
 * @details It sends a resumed event to the display server so that it can pass on the watch application.
 * @since_tizen 4.0
 * @return @c 0 on success, otherwise a negative error value
 */
extern int watch_manager_resume(void);

#ifdef __cplusplus
}
#endif

#endif
