/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Flora License, Version 1.1 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://floralicense.org/license/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <glib.h>
#include <aul_svc.h>
#include <bundle_internal.h>
#include <app_control_internal.h>
#include <wayland-extension/tizen-remote-surface-client-protocol.h>
#include <tbm_surface.h>
#include <screen_connector_toolkit_evas.h>
#include <dlog.h>
#include <unistd.h>
#include "watch_control.h"
#include <aul.h>
#include <Elementary.h>
#include <system_info.h>
#include <tbm_surface.h>
#include <pkgmgr-info.h>
#include <aul_app_com.h>
#include <aul_screen_connector.h>
#include <Ecore_Wl2.h>
#include <gio/gio.h>
#include <aul_watch_control.h>
#include <device/display.h>
#include <device/callback.h>

#include "watch_control_private.h"

#define KEY_SCREEN_SHAPE_CIRCLE "http://tizen.org/feature/screen.shape.circle"

static int __watch_viewer_initialized = 0;
static int __watch_size_policy = WATCH_POLICY_HINT_EXPAND;

static int __default_width;
static int __default_height;
static screen_connector_toolkit_evas_h __toolkit;
static screen_connector_toolkit_evas_h __rid_toolkit;

static Evas_Object *__win;
static char *__watch_appid;
static char *viewer_appid;
static int __is_dead_signal_connected;
static aul_app_com_connection_h __conn_dead_signal;
static unsigned int __watch_rid;
static bool __is_bound;
static int __viewer_visibility = AUL_SCREEN_STATUS_PAUSE;

struct toolkit_info_s {
	char *appid;
	screen_connector_toolkit_evas_h toolkit;
};
static GList *__toolkit_list;

static GList *__dead_cbs;
struct dead_cb_s {
	watch_dead_signal_cb cb;
	void *data;
};

struct manual_render_s {
	aul_watch_control_h control_handle;
	Ecore_Event_Handler *pre_visibility;
	Ecore_Event_Handler *visibility;
	guint timer;
	guint pause_timer;
	bool started;
};

static bool __is_pended;
static int __change_viewer_visibility(int visibility, bool update_cur_state);
static int __change_visibility(int visibility);
static aul_app_com_connection_h __conn_launch_signal;
static struct manual_render_s __manual_render;

static int __listen_launch_signal(void);
static void __ignore_launch_signal(void);
static void __set_viewer_appid(void);
static void __manual_render_start(bool force_resume);
static void __manual_render_stop(void);

static bool __lcd_is_off(void)
{
	int ret = 0;
	display_state_e state = DISPLAY_STATE_NORMAL;

	ret = device_display_get_state(&state);
	if (ret < 0 || state == DISPLAY_STATE_SCREEN_OFF)
		return true;
	return false;
}

static bool __watch_is_visible(void)
{
	Evas_Object *image;
	bool is_visible = false;

	image = evas_object_data_get(__win, "tbm,watch");
	if (!image)
		return false;

	screen_connector_toolkit_evas_is_visible(image, &is_visible);

	return is_visible;
}

static gboolean __manual_render_timeout_cb(gpointer data)
{
	_W("Manual render timeout expired");
	__manual_render_stop();
	__manual_render.timer = 0;
	return G_SOURCE_REMOVE;
}

static gboolean __manual_render_pause_timeout_cb(gpointer data)
{
	int lcdoff = __lcd_is_off();

	_W("Manual render pause timer is expired, lcdoff %d", lcdoff);
	if (lcdoff)
		watch_manager_notify_paused_status_of_viewer();

	__manual_render.pause_timer = 0;
	return G_SOURCE_REMOVE;
}

static void __manual_render_set_pause_timer(void)
{
	if (__manual_render.pause_timer)
		return;

	__manual_render.pause_timer = g_timeout_add(200,
			__manual_render_pause_timeout_cb, NULL);
}

static void __manual_render_unset_pause_timer(void)
{
	if (!__manual_render.pause_timer)
		return;

	g_source_remove(__manual_render.pause_timer);
	__manual_render.pause_timer = 0;
}

static void __manual_render_unset_timer(void)
{
	if (!__manual_render.timer)
		return;

	g_source_remove(__manual_render.timer);
	__manual_render.timer = 0;
}

static void __manual_render_set_timer(void)
{
	if (__manual_render.timer)
		return;

	__manual_render.timer = g_timeout_add(200,
			__manual_render_timeout_cb, NULL);
}

static void __manual_render_start(bool force_resume)
{
	if (!__watch_is_visible()) {
		_W("Watch is not visible. Skip manual render");
		return;
	}

	if (__is_bound) {
		_W("Watch is bound cannot manual render just call tick");
		__change_viewer_visibility(AUL_SCREEN_STATUS_PRE_RESUME, false);
		watch_manager_resume();
		return;
	}

	if (__manual_render.started) {
		_W("Manual render is already started");
		return;
	}

	elm_win_norender_push(__win);
	__change_viewer_visibility(AUL_SCREEN_STATUS_PRE_RESUME, false);
	watch_manager_resume();
	__manual_render_unset_timer();
	__manual_render_set_timer();
	__manual_render.started = true;

	if (force_resume) {
		watch_manager_notify_resumed_status_of_viewer();
		if (__lcd_is_off()) {
			_W("Pause timer is started");
			__manual_render_unset_pause_timer();
			__manual_render_set_pause_timer();
		}
	}
}

static void __manual_render_stop(void)
{
	if (!__manual_render.started)
		return;

	elm_win_norender_pop(__win);
	_I("Set manual false");

	__manual_render.started = false;
	__manual_render_unset_timer();
}

static void __aul_watch_control_cb(bundle *b, void *user_data)
{
	bool v = __watch_is_visible();

	if (!v)
		return;
	__manual_render_start(true);
}

static gboolean __force_resume_cb(gpointer data)
{
	watch_manager_resume();
	return G_SOURCE_REMOVE;
}

static void __obj_move_cb(void *data, Evas *e, Evas_Object *obj,
		void *event_info)
{
	bool v = __watch_is_visible();

	if (!v)
		return;
	__manual_render_start(true);
	g_idle_add(__force_resume_cb, NULL);
}

static Eina_Bool __window_on_pre_visibility(void *data, int type, void *event)
{
        Ecore_Wl2_Event_Window_Pre_Visibility_Change *ev = event;
	Ecore_Evas *ee;
	Ecore_Wl2_Window *wl_win;
	int win_id;

        if (!ev)
                return ECORE_CALLBACK_RENEW;

	ee = ecore_evas_ecore_evas_get(evas_object_evas_get(__win));
	if (!ee)
		return ECORE_CALLBACK_RENEW;

	wl_win = ecore_evas_wayland2_window_get(ee);
	if (!wl_win)
		return ECORE_CALLBACK_RENEW;

	win_id = ecore_wl2_window_id_get(wl_win);
	if (ev->win != win_id)
		return ECORE_CALLBACK_RENEW;

        if (ev && ev->type == ECORE_WL2_WINDOW_VISIBILITY_TYPE_PRE_UNOBSCURED) {
                __manual_render_start(false);
                LOGD("[__WINDOW_PRE_VISIBILITY_CHANGE__] win_id(%d)", ev->win);
        }

        return ECORE_CALLBACK_RENEW;
}

static int __manual_render_init(void)
{
	int ret;

	ret = aul_watch_control_add_handler(__aul_watch_control_cb, NULL,
			&__manual_render.control_handle);
	if (ret < 0) {
		_E("Failed to add aul watch control handler");
		return -1;
	}

	__manual_render.pre_visibility = ecore_event_handler_add(
			ECORE_WL2_EVENT_WINDOW_PRE_VISIBILITY_CHANGE,
			__window_on_pre_visibility, NULL);
	if (!__manual_render.pre_visibility)
		_W("Failed to add pre visibility change handler");
	return 0;
}

static void __manual_render_fini(void)
{
	__manual_render_stop();

	ecore_event_handler_del(__manual_render.visibility);
	__manual_render.visibility = NULL;

	ecore_event_handler_del(__manual_render.pre_visibility);
	__manual_render.pre_visibility = NULL;

	aul_watch_control_remove_handler(__manual_render.control_handle);
	__manual_render.control_handle = NULL;
}

static void __destroy_toolkit_info(gpointer data)
{
	struct toolkit_info_s *info = data;

	if (!info)
		return;

	if (info->appid)
		free(info->appid);

	free(info);
}

static struct toolkit_info_s *__create_toolkit_info(const char *appid,
		screen_connector_toolkit_evas_h toolkit)
{
	struct toolkit_info_s *info;

	info = malloc(sizeof(struct toolkit_info_s));
	if (!info) {
		_E("Out of memory");
		return NULL;
	}

	info->appid = strdup(appid);
	if (!info->appid) {
		_E("Failed to duplicate appid(%s)", appid);
		free(info);
		return NULL;
	}

	info->toolkit = toolkit;

	return info;
}

static struct toolkit_info_s *__find_toolkit_info(const char *appid)
{
	struct toolkit_info_s *info;
	GList *iter;

	iter = __toolkit_list;
	while (iter) {
		info = (struct toolkit_info_s *)iter->data;
		if (info && !strcmp(info->appid, appid))
			return info;

		iter = g_list_next(iter);
	}

	return NULL;
}

static struct toolkit_info_s *__find_toolkit_info_by_toolkit(
		screen_connector_toolkit_evas_h toolkit)
{
	struct toolkit_info_s *info;
	GList *iter;

	iter = __toolkit_list;
	while (iter) {
		info = (struct toolkit_info_s *)iter->data;
		if (info && info->toolkit == toolkit)
			return info;

		iter = g_list_next(iter);
	}

	return NULL;
}

static void __process_pending_status(int status, unsigned int surf)
{
	int r;

	_W("Send pending event(%d) to %s:%u", status, __watch_appid, surf);
	r = aul_screen_connector_update_screen_viewer_status(status, surf);
	if (r < 0)
		_E("Failed to update screen viewer status");
}

static void __flush_pending_queue(unsigned int surf)
{
	if (!__is_pended)
		return;

	__process_pending_status(__viewer_visibility, surf);
	__is_pended = false;
}

static void __push_pending_queue(int status)
{
	__is_pended = true;
	_W("status(%d)", status);
}

static void __win_resized(void *data, Evas *e, Evas_Object *obj, void *event_info)
{
	Evas_Object *win = obj;
	Evas_Coord x, y, w, h;

	if (!win)
		return;

	evas_object_geometry_get(win, &x, &y, &w, &h);
	__default_width = w;
	__default_height = h;
}

static int __watch_viewer_init(Evas_Object *win)
{
	if (__watch_viewer_initialized)
		return 0;

	if (win == NULL) {
		_E("invalid arguments");
		return -1;
	}

	__win = win;
	screen_connector_toolkit_evas_init(__win, SCREEN_CONNECTOR_SCREEN_TYPE_WATCH);
	screen_connector_toolkit_evas_start_visibility_notify();

	__win_resized(NULL, NULL, win, NULL); /* init */
	evas_object_event_callback_add(win, EVAS_CALLBACK_RESIZE, __win_resized, NULL);

	__manual_render_init();

	__watch_viewer_initialized = 1;

	return 0;
}

static void __watch_viewer_fini()
{
	__manual_render_fini();

	if (__win) {
		evas_object_event_callback_del(__win, EVAS_CALLBACK_RESIZE, __win_resized);
		__win = NULL;
	}

	screen_connector_toolkit_evas_stop_visibility_notify();
	screen_connector_toolkit_evas_fini(SCREEN_CONNECTOR_SCREEN_TYPE_WATCH);
}

API int watch_manager_init(Evas_Object *win)
{
	__watch_viewer_init(win);
	__listen_launch_signal();

	if (!viewer_appid)
		__set_viewer_appid();

	return 0;
}

static int __watch_screen_get_width()
{
	if (__watch_size_policy == WATCH_POLICY_HINT_EXPAND)
		return __default_width;

	return 0;
}

static int __watch_screen_get_height()
{
	if (__watch_size_policy == WATCH_POLICY_HINT_EXPAND)
		return __default_height;

	return 0;
}

static void __screen_connector_toolkit_evas_added_cb(const char *appid, const char *instance_id, int pid,
	Evas_Object *image, void *data)
{
	int w, h, x, y;
	evas_object_geometry_get(image, &x, &y, &w, &h);

	_D("obj(%p) added", image);
	_D("w: %d, h: %d, x: %d y: %d", w, h, x, y);
	evas_object_smart_callback_call(__win, WATCH_SMART_SIGNAL_ADDED, image);
	evas_object_data_set(__win, "tbm,watch", image);
	evas_object_event_callback_add(image, EVAS_CALLBACK_MOVE,
			__obj_move_cb, image);
	screen_connector_toolkit_evas_get_rid(image, (int *)&__watch_rid);
	__flush_pending_queue(__watch_rid);
	__change_viewer_visibility(__viewer_visibility, false);
}

static gboolean __remove_toolkit(gpointer user_data)
{
	screen_connector_toolkit_evas_remove(
			(screen_connector_toolkit_evas_h)user_data);
	return G_SOURCE_REMOVE;
}

static void __screen_connector_toolkit_evas_removed_cb(const char *appid, const char *instance_id, int pid,
	Evas_Object *image, void *data)
{
	Evas_Object *cur_image;
	struct toolkit_info_s *info;
	GList *iter;

	_D("obj(%p) removed", image);
	evas_object_smart_callback_call(__win, WATCH_SMART_SIGNAL_REMOVED, image);
	evas_object_event_callback_del_full(image, EVAS_CALLBACK_MOVE,
			__obj_move_cb, image);
	cur_image = evas_object_data_get(__win, "tbm,watch");
	if (cur_image == image) {
		evas_object_data_set(__win, "tbm,watch", NULL);
		if (__toolkit) {
			info = __find_toolkit_info_by_toolkit(__toolkit);
			if (info) {
				__toolkit_list = g_list_remove(__toolkit_list,
						info);
				__destroy_toolkit_info(info);
			}
			g_idle_add(__remove_toolkit, __toolkit);
		}
		if (__rid_toolkit)
			g_idle_add(__remove_toolkit, __rid_toolkit);
		__toolkit = NULL;
		__rid_toolkit = NULL;
		__is_bound = false;
	}

	iter = __toolkit_list;
	while (iter) {
		info = (struct toolkit_info_s *)iter->data;
		iter = g_list_next(iter);
		if (info->toolkit != __toolkit &&
				info->toolkit != __rid_toolkit) {
			__toolkit_list = g_list_remove(__toolkit_list, info);
			g_idle_add(__remove_toolkit, info->toolkit);
			__destroy_toolkit_info(info);
		}
	}
}

static void __screen_connector_toolkit_evas_updated_cb(
		const char *appid, const char *instance_id, int pid,
	Evas_Object *image, void *data)
{
	evas_object_smart_callback_call(
			__win, WATCH_SMART_SIGNAL_UPDATED, image);
	if (__manual_render.started)
		__manual_render_stop();
	else
		LOGI("updated");
}

API int watch_manager_enable_buffer_update(void)
{
	int ret;

	if (__rid_toolkit == NULL && __toolkit == NULL) {
		_E("cannot find toolkit");
		return -1;
	}

	if (__rid_toolkit) {
		ret = screen_connector_toolkit_evas_set_changed_event_filter(
				__rid_toolkit,
				SCREEN_CONNECTOR_EVAS_CHANGED_EVENT_FILTER_NONE);
		if (ret != 0) {
			_E("fail to set filter");
			return ret;
		}

	}

	if (__toolkit) {
		ret = screen_connector_toolkit_evas_set_changed_event_filter(
				__toolkit,
				SCREEN_CONNECTOR_EVAS_CHANGED_EVENT_FILTER_NONE);
		if (ret != 0) {
			_E("fail to set filter");
			return ret;
		}

	}

	return 0;
}

API int watch_manager_disable_buffer_update(void)
{
	int ret;

	if (__rid_toolkit == NULL && __toolkit == NULL) {
		_E("cannot find toolkit");
		return -1;
	}

	if (__rid_toolkit) {
		ret = screen_connector_toolkit_evas_set_changed_event_filter(
				__rid_toolkit,
				SCREEN_CONNECTOR_EVAS_CHANGED_EVENT_FILTER_ALL);
		if (ret != 0) {
			_E("fail to set filter");
			return ret;
		}

	}

	if (__toolkit) {
		ret = screen_connector_toolkit_evas_set_changed_event_filter(
				__toolkit,
				SCREEN_CONNECTOR_EVAS_CHANGED_EVENT_FILTER_ALL);
		if (ret != 0) {
			_E("fail to set filter");
			return ret;
		}

	}

	return 0;
}

API int watch_manager_set_resource_id(int resource_id)
{
	screen_connector_toolkit_evas_ops ops;
	screen_connector_toolkit_evas_h handle;
	int ret;

	ops.added_cb = __screen_connector_toolkit_evas_added_cb;
	ops.removed_cb = __screen_connector_toolkit_evas_removed_cb;
	ops.updated_cb = __screen_connector_toolkit_evas_updated_cb;
	handle = screen_connector_toolkit_evas_add_by_rid(&ops, resource_id,
			SCREEN_CONNECTOR_SCREEN_TYPE_WATCH, NULL);
	if (handle == NULL) {
		_E("Fail screen_connector_toolkit_evas_add_by_rid");
		return -1;
	}

	if (__rid_toolkit)
		screen_connector_toolkit_evas_remove(__rid_toolkit);

	__rid_toolkit = handle;

	if (__watch_appid) {
		free(__watch_appid);
		__watch_appid = NULL;
	}
	__watch_rid = resource_id;
	ret = aul_screen_connector_get_appid_by_surface_id(resource_id,
		&__watch_appid);
	if (ret != AUL_R_OK) {
		_E("Fail to get appid of %d", resource_id);
		screen_connector_toolkit_evas_remove(handle);
		__rid_toolkit = NULL;
		return -1;
	}
	_D("appid(%s), rid(%d)", __watch_appid, resource_id);

	return 0;
}

API int watch_manager_get_resource_id(Evas_Object *watch, int *resource_id)
{
	return screen_connector_toolkit_evas_get_rid(watch, resource_id);
}

API int watch_manager_send_terminate(Evas_Object *watch)
{
	int pid;
	int r;

	screen_connector_toolkit_evas_get_pid(watch, &pid);
	_D("watch_manager_send_terminate %d", pid);
	if (pid < 1) {
		_E("failed to get pid from %p", watch);
		return -1;
	}

	r = aul_terminate_pid(pid);
	if (r < 0) {
		_E("failed to terminate pid:%d (%d)", pid, r);
		return -1;
	}

	return 0;
}

static void __set_viewer_appid(void)
{
	char appid[255] = {0,};
	int ret;

	ret = aul_app_get_appid_bypid(getpid(), appid, sizeof(appid));
	if (ret == AUL_R_OK) {
		viewer_appid = strdup(appid);
		if (!viewer_appid)
			_E("out of memory");
	}
}

API int watch_manager_get_app_control(const char *app_id, app_control_h *app_control)
{
	char buf[10];
	bundle *b = NULL;

	if (app_id == NULL) {
		_E("Invalid param");
		return -1;
	}

	app_control_create(app_control);
	app_control_set_app_id(*app_control, app_id);

	snprintf(buf, sizeof(buf), "%d", __watch_screen_get_width());
	app_control_add_extra_data(*app_control, "WATCH_WIDTH", buf);
	snprintf(buf, sizeof(buf), "%d", __watch_screen_get_height());
	app_control_add_extra_data(*app_control, "WATCH_HEIGHT", buf);
	snprintf(buf, sizeof(buf), "%d", __viewer_visibility);
	app_control_add_extra_data(*app_control, "WATCH_VIEWER_VISIBILITY", buf);

	app_control_set_operation(*app_control, APP_CONTROL_OPERATION_MAIN);

	app_control_to_bundle(*app_control, &b);
	if (b) {
		if (!viewer_appid)
			__set_viewer_appid();

		bundle_add_str(b, AUL_K_WIDGET_VIEWER, viewer_appid);
	}

	if (__watch_appid)
		free(__watch_appid);
	__watch_appid = strdup(app_id);
	if (!__watch_appid)
		_E("Out of memory");

	return 0;
}

API int watch_manager_fini()
{
	if (__toolkit_list) {
		g_list_free_full(__toolkit_list, __destroy_toolkit_info);
		__toolkit_list = NULL;
	}

	__watch_viewer_fini();

	if (__watch_appid) {
		free(__watch_appid);
		__watch_appid = NULL;
	}

	if (viewer_appid) {
		free(viewer_appid);
		viewer_appid = NULL;
	}

	if (__conn_dead_signal) {
		if (aul_app_com_leave(__conn_dead_signal) < 0)
			_E("failed to leave app com endpoint dead signal");
		__conn_dead_signal = NULL;
	}

	__ignore_launch_signal();

	return 0;
}

API int watch_policy_set_size_hint(watch_policy_size_hint hint)
{
	__watch_size_policy = hint;

	return 0;
}

static bool __is_circle(void)
{
	static bool initialized = false;
	static bool shape_circle = false;
	int r;

	if (initialized)
		return shape_circle;

	r = system_info_get_platform_bool(KEY_SCREEN_SHAPE_CIRCLE,
			&shape_circle);
	if (r != SYSTEM_INFO_ERROR_NONE) {
		_E("Failed to get system info");
		return false;
	}

	initialized = true;
	return shape_circle;
}

static float __get_opr(void *source_data, int width, int height)
{
	int *source;
	int x;
	int y;
	int idx;
	int max_rad;
	int rad;
	int pos;
	unsigned int r = 0;
	unsigned int g = 0;
	unsigned int b = 0;
	unsigned int pixel_sum;
	float opr;
	float max_opr;
	bool shape_circle = __is_circle();

	source = source_data;
	pos = width / 2;
	max_rad = (height * height) >> 2;

	for (y = 0; y < height; ++y) {
		for (x = 0; x < width; ++x) {
			if (shape_circle) {
				rad = (pos - x) * (pos - x) + (pos - y) * (pos - y);
				if (rad <= max_rad) {
					idx = y * width + x;
					r += ((source[idx] & 0x00ff0000) >> 16);
					g += ((source[idx] & 0x0000ff00) >> 8);
					b += (source[idx] & 0x000000ff);
				}
			} else {
				idx = y * width + x;
				r += ((source[idx] & 0x00ff0000) >> 16);
				g += ((source[idx] & 0x0000ff00) >> 8);
				b += (source[idx] & 0x000000ff);
			}
		}
	}

	if (shape_circle)
		max_opr = (width / 2) * (height / 2) * 3.14 * 3 * 255;
	else
		max_opr = width * height * 3 * 255;

	pixel_sum = r + g + b;
	opr = ((float)pixel_sum) / max_opr;

	return opr;
}

API int watch_manager_get_opr(float *opr)
{
	Evas_Object *image;
	tbm_surface_h surface;
	tbm_surface_info_s info = { 0, };
	Evas_Native_Surface *ns;
	void *source_data;
	int width = 0;
	int height = 0;
	int ret;

	if (opr == NULL) {
		_E("Invalid parameter");
		return -1;
	}

	if (__win == NULL) {
		_E("Window is NULL");
		return -1;
	}

	image = evas_object_data_get(__win, "tbm,watch");
	if (image == NULL) {
		_E("Failed to get image");
		return -1;
	}

	evas_object_geometry_get(image, NULL, NULL, &width, &height);
	_D("width(%d), height(%d)", width, height);

	ns = evas_object_image_native_surface_get(image);
	if (ns == NULL) {
		_E("Failed to get evas native surface");
		return -1;
	}

	surface = (tbm_surface_h)ns->data.tbm.buffer;
	if (surface == NULL) {
		_E("tbm surface is NULL");
		return -1;
	}

	ret = tbm_surface_map(surface, TBM_SURF_OPTION_READ, &info);
	if (ret != TBM_SURFACE_ERROR_NONE) {
		_E("Failed to map tbm surface - %d", ret);
		return -1;
	}

	source_data = info.planes[0].ptr;
	if (source_data == NULL) {
		_E("Failed to get source");
		tbm_surface_unmap(surface);
		return -1;
	}

	*opr = __get_opr(source_data, width, height);
	_D("On Pixel Ratio(%f)", *opr);
	tbm_surface_unmap(surface);

	return 0;
}

API int watch_manager_cancel_click_event(Evas_Object *watch)
{
	return screen_connector_toolkit_evas_send_touch_cancel(watch);
}

API int watch_manager_change_auto_visibility(bool enable)
{
	if (!enable)
		screen_connector_toolkit_evas_stop_visibility_notify();
	else
		screen_connector_toolkit_evas_start_visibility_notify();

	return 0;
}

static int __change_viewer_visibility(int visibility, bool update_cur_state)
{
	Evas_Object *watch;
	int r;

	if (update_cur_state)
		__viewer_visibility = visibility;

	if (__win == NULL) {
		_E("Window is NULL");
		__push_pending_queue(visibility);
		return 0;
	}

	if (__watch_rid == 0) {
		watch = evas_object_data_get(__win, "tbm,watch");
		if (watch == NULL || __watch_appid == NULL) {
			_E("watch is not added");
			__push_pending_queue(visibility);
			return 0;
		}

		screen_connector_toolkit_evas_get_rid(watch, (int *)&__watch_rid);
	}

	r = aul_screen_connector_update_screen_viewer_status(visibility, __watch_rid);
	if (r < 0) {
		_E("Failed to update screen viewer status");
		return -1;
	}
	_D("send watch control status to %s:%d", __watch_appid, visibility);

	return 0;
}

API int watch_manager_notify_resumed_status_of_viewer(void)
{
	return __change_viewer_visibility(AUL_SCREEN_STATUS_RESUME, true);
}

API int watch_manager_notify_paused_status_of_viewer(void)
{
	return __change_viewer_visibility(AUL_SCREEN_STATUS_PAUSE, true);
}

static int __change_visibility(int visibility)
{
	Evas_Object *image;

	image = evas_object_data_get(__win, "tbm,watch");
	if (image == NULL) {
		_E("Failed to get image");
		return -1;
	}

	return screen_connector_toolkit_evas_send_visibility(image, visibility);
}

API int watch_manager_pause(void)
{
	return __change_visibility(VISIBILITY_TYPE_FULLY_OBSCURED);
}

API int watch_manager_resume(void)
{
	return __change_visibility(VISIBILITY_TYPE_UNOBSCURED);
}

static void __on_aux_msg(void *data, Evas_Object *o, void *ev_info)
{
	Evas_Object *binding_window = (Evas_Object *)data;
	Elm_Win_Aux_Message *msg = (Elm_Win_Aux_Message *)ev_info;
	const char *key = elm_win_aux_msg_key_get(binding_window, msg);
	const char *val = elm_win_aux_msg_val_get(binding_window, msg);

	if (!strcmp(key, "tz_remote_surface_mng") && !strcmp(val, "prebind")) {
		__is_bound = true;
		LOGI("Start bind mode !!");
		evas_object_smart_callback_del(binding_window,
			"aux,msg,received", __on_aux_msg);
	}
}

API int watch_manager_window_bind_request(Evas_Object *win)
{
	if (!__toolkit || !win)
		return -1;

	LOGI("Bind request");
	elm_win_aux_hint_add(win, "wm.policy.win.msg.use", "1");
	evas_object_smart_callback_add(win, "aux,msg,received",
			__on_aux_msg, win);

	return screen_connector_toolkit_evas_bind(__toolkit, win);
}

API int watch_manager_window_bind(Evas_Object *win)
{
	if (!__toolkit || !win)
		return -1;

	LOGI("Bind");
	__is_bound = true;

	return screen_connector_toolkit_evas_bind(__toolkit, win);
}

API int watch_manager_window_unbind(void)
{
	if (!__toolkit)
		return -1;

	LOGI("Unbind");
	__is_bound = false;

	return screen_connector_toolkit_evas_unbind(__toolkit);
}

API int watch_manager_get_setup_appid(const char *app_id, char **setup_appid)
{
	int ret = 0;
	pkgmgrinfo_appinfo_h handle;
	char *tmp_appid = NULL;

	ret = pkgmgrinfo_appinfo_get_appinfo(app_id, &handle);
	if (ret != PMINFO_R_OK) {
		LOGE("fail to get app info");
		return -1;
	}

	ret = pkgmgrinfo_appinfo_get_setup_appid(handle, &tmp_appid);
	if (ret != PMINFO_R_OK) {
		LOGE("fail to get setup appid");
		pkgmgrinfo_appinfo_destroy_appinfo(handle);
		return -1;
	}

	if (tmp_appid)
		*setup_appid = strdup(tmp_appid);

	pkgmgrinfo_appinfo_destroy_appinfo(handle);
	return ret;
}

static int __dead_signal_handler(const char *endpoint, aul_app_com_result_e e,
		bundle *envelope, void *user_data)
{
	char *appid = NULL;
	char *is_faulted = NULL;
	char *pid_str = NULL;
	int pid = 0;
	struct dead_cb_s *cb_info;
	GList *iter = __dead_cbs;
	bool faulted = false;

	bundle_get_str(envelope, AUL_K_APPID, &appid);
	bundle_get_str(envelope, AUL_K_PID, &pid_str);
	bundle_get_str(envelope, AUL_K_IS_FAULT, &is_faulted);

	_D("appid %s pid %s is_faulted %s", appid, pid_str, is_faulted);
	if (pid_str)
		pid = atoi(pid_str);
	if (is_faulted && (strcmp(is_faulted, "true") == 0))
		faulted = true;

	while (iter) {
		cb_info = (struct dead_cb_s *)iter->data;
		if (cb_info && cb_info->cb)
			cb_info->cb(appid, pid, faulted, cb_info->data);

		iter = iter->next;
	}

	return 0;
}

static int __connect_dead_signal()
{
	if (__is_dead_signal_connected)
		return 0;

	if (aul_app_com_create("watch.dead", NULL,
			__dead_signal_handler, NULL, &__conn_dead_signal) < 0) {
		_E("failed to create status");
		return -1;
	}

	__is_dead_signal_connected = 1;

	return 0;
}

API int watch_manager_add_dead_signal_listener(watch_dead_signal_cb cb, void *data)
{
	int ret = 0;
	struct dead_cb_s *cb_info;

	cb_info = (struct dead_cb_s *)malloc(sizeof(struct dead_cb_s));
	if (!cb_info) {
		_E("out of memory");
		return -1;
	}

	cb_info->cb = cb;
	cb_info->data = data;

	__dead_cbs = g_list_append(__dead_cbs, cb_info);
	if (!__dead_cbs) {
		_E("out of memory");
		free(cb_info);
		return -1;
	}

	__connect_dead_signal();

	return ret;
}

API int watch_manager_remove_dead_signal_listener(watch_dead_signal_cb cb)
{
	struct dead_cb_s *cb_info;
	GList *iter = __dead_cbs;

	if (!cb) {
		_E("invalid parameter");
		return -1;
	}

	while (iter) {
		cb_info = (struct dead_cb_s *)iter->data;
		if (cb_info && cb_info->cb == cb) {
			__dead_cbs = g_list_remove_link(__dead_cbs, iter);
			free(cb_info);
			g_list_free(iter);
			return 0;
		}
		iter = iter->next;
	}

	_E("wrong argument");
	return -1;
}

static int __launch_signal_handler(const char *endpoint,
		aul_app_com_result_e res, bundle *envelope, void *user_data)
{
	char *appid = NULL;
	char *viewer = NULL;
	char *pid_str = NULL;
	screen_connector_toolkit_evas_ops ops;
	screen_connector_toolkit_evas_h toolkit;
	struct toolkit_info_s *info;
	bool exist;

	bundle_get_str(envelope, AUL_K_WIDGET_VIEWER, &viewer);
	if (viewer == NULL)
		return -1;

	if (strcmp(viewer, viewer_appid) != 0)
		return 0;

	bundle_get_str(envelope, AUL_K_APPID, &appid);
	if (appid == NULL)
		return -1;

	bundle_get_str(envelope, AUL_K_PID, &pid_str);

	_D("cur appid(%s), new appid(%s), pid(%s)",
			__watch_appid, appid, pid_str);

	exist = screen_connector_toolkit_is_exist(appid,
			SCREEN_CONNECTOR_SCREEN_TYPE_WATCH);
	if (exist) {
		info = __find_toolkit_info(appid);
		if (info) {
			__toolkit_list = g_list_remove(__toolkit_list, info);
			screen_connector_toolkit_evas_remove(info->toolkit);
			__destroy_toolkit_info(info);
		}
	} else {
		_D("Old toolkit %p", __toolkit);
	}

	ops.added_cb = __screen_connector_toolkit_evas_added_cb;
	ops.removed_cb = __screen_connector_toolkit_evas_removed_cb;
	ops.updated_cb = __screen_connector_toolkit_evas_updated_cb;
	toolkit = screen_connector_toolkit_evas_add(&ops, appid,
			SCREEN_CONNECTOR_SCREEN_TYPE_WATCH, NULL);
	if (!toolkit) {
		LOGE("Failed to add toolkit evas. appid(%s)", appid);
	} else {
		info = __create_toolkit_info(appid, toolkit);
		if (info)
			__toolkit_list = g_list_append(__toolkit_list, info);

		__toolkit = toolkit;
	}

	if (__watch_appid)
		free(__watch_appid);

	__watch_appid = strdup(appid);
	if (__watch_appid == NULL)
		_E("Out of memory");

	return 0;
}

static int __listen_launch_signal(void)
{
	int r;

	if (__conn_launch_signal)
		return 0;

	r = aul_app_com_create("watch.launch", NULL, __launch_signal_handler,
			NULL, &__conn_launch_signal);
	if (r < 0) {
		_E("Failed to listen watch.launch signal");
		return -1;
	}

	return 0;
}

static void __ignore_launch_signal(void)
{
	if (!__conn_launch_signal)
		return;

	aul_app_com_leave(__conn_launch_signal);
	__conn_launch_signal = NULL;
}
