/*
 * Copyright (c) 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Flora License, Version 1.1 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://floralicense.org/license/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#pragma once

#include <tizen_type.h>
#include <Evas.h>

#ifdef __cplusplus
extern "C" {
#endif


/**
 * @file widget_viewer_evas_extension.h
 */

/**
 * @addtogroup CAPI_WIDGET_VIEWER_EVAS_MODULE
 * @{
 */

/**
 * @brief Sets delayed resuming time.
 * @details The resuming event for widget applications will be postponed
 * @since_tizen 4.0
 * @param[in] time The time for being postponed, in milliseconds
 *            The time should be between 0 and 10000
 * @return @c 0 on success,
 *         otherwise a negative error value
 * @retval #WIDGET_ERROR_NONE If success
 * @retval #WIDGET_ERROR_INVALID_PARAMETER Invalid argument
 * @remark Delayed resuming will work only if the option is set by widget_viewer_evas_set_option()
 * @see widget_viewer_evas_set_option()
 * @see #WIDGET_VIEWER_EVAS_DELAYED_RESUME
 * */
int widget_viewer_evas_set_delayed_resuming_time(int time);

/**
 * @brief Gets the widget object's instance ID.
 * @since_tizen 3.0
 * @privilege %http://tizen.org/privilege/widget.viewer
 * @remarks The returned instance ID should not be freed.
 * @remarks The returned widget instance ID is volatile. If the device reboots or the widget's process restarts, it will be changed.\n
 *      So, you should not assume this value is a persistent one.
 * @remarks The returned string is usable only before the widget is destroyed. If it's going to be needed after that, it should be copied.
 * @remarks The specific error code can be obtained using the get_last_result() method. Error codes are described in Exception section.
 * @param[in] widget A widget object
 * @return The widget's instance ID as a string, NULL in case of errors
 * @exception #WIDGET_ERROR_NONE Successfully retrieved the widget instance ID
 * @exception #WIDGET_ERROR_NOT_SUPPORTED Not supported
 * @exception #WIDGET_ERROR_FAULT Widget is not ready
 * @exception #WIDGET_ERROR_PERMISSION_DENIED Permission denied
 * @exception #WIDGET_ERROR_INVALID_PARAMETER Invalid argument
 * @see get_last_result()
 */
const char *widget_viewer_evas_get_widget_instance_id(Evas_Object *widget);

/**
 * @brief Launches a setup application.
 * @since_tizen 3.0
 * @feature http://tizen.org/feature/shell.appwidget
 * @privilege %http://tizen.org/privilege/widget.viewer \n
 *            %http://tizen.org/privilege/appmanager.launch
 * param[in] widget A widget object
 * @return #WIDGET_ERROR_NONE on success,
 *          otherwise an error code (see #WIDGET_ERROR_XXX) on failure
 * @retval #WIDGET_ERROR_INVALID_PARAMETER Invalid arguments
 * @retval #WIDGET_ERROR_NOT_EXIST Setup application is not exists
 * @retval #WIDGET_ERROR_NONE Successfully requested
 */
int widget_viewer_evas_launch_setup_app(Evas_Object *widget);

/**
 * @}
 */

#ifdef __cplusplus
}
#endif

/* End of a file */
