/*
 * Samsung API
 * Copyright (c) 2013 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Flora License, Version 1.1 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://floralicense.org/license/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <libintl.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <glib.h>

#include <Ecore_Wl2.h>
#include <Elementary.h>
#include <Ecore.h>
#include <Ecore_Evas.h>
#include <Evas.h>
#include <Edje.h>

#include <pkgmgr-info.h>
#include <system_info.h>
#include <cynara-client.h>
#include <fcntl.h>

#include <widget_errno.h>
#include <widget_service_internal.h>
#include <widget_instance.h>
#include <aul_app_com.h>
#include <aul_widget.h>
#include <wayland-extension/tizen-remote-surface-client-protocol.h>
#include <tbm_surface.h>
#include <screen_connector_toolkit_evas.h>
#include <app_control_internal.h>

#ifdef API
#undef API
#endif
#define API __attribute__((visibility("default")))

#if defined(LOG_TAG)
#undef LOG_TAG
#endif
#define LOG_TAG "WIDGET_EVAS"
#include <dlog.h>

#include "widget_viewer_evas.h"

#if !defined(WIDGET_COUNT_OF_SIZE_TYPE)
	#define WIDGET_COUNT_OF_SIZE_TYPE 13
#endif

#if !defined(S_)
#define S_(str) dgettext("sys_string", str)
#endif

#if !defined(T_)
#define T_(str) dgettext(PKGNAME, str)
#endif

#if !defined(N_)
#define N_(str) (str)
#endif

#if !defined(_)
#define _(str) gettext(str)
#endif

#if !defined(WIDGET_VIEWER_EVAS_RESOURCE_EDJ)
#define WIDGET_VIEWER_EVAS_RESOURCE_EDJ "/usr/share/widget_viewer_evas/res/edje/widget_viewer_evas.edj"
#endif

#if !defined(WIDGET_VIEWER_EVAS_RESOURCE_OVERLAY_LOADING)
#define WIDGET_VIEWER_EVAS_RESOURCE_OVERLAY_LOADING "overlay"
#endif

#if !defined(WIDGET_VIEWER_EVAS_RESOURCE_DEFAULT_IMG)
#define WIDGET_VIEWER_EVAS_RESOURCE_DEFAULT_IMG "/usr/share/widget_viewer_evas/res/image/unknown.png"
#endif


#define DEFAULT_OVERLAY_COUNTER 2
#define DEFAULT_OVERLAY_WAIT_TIME 1.0f
#define RESEVED_AREA_FOR_GBAR_EFFECT 100

#define WIDGET_CLASS_NAME "widget"

#define WIDGET_KEEP_BUFFER    -2

#define DEFAULT_CLUSTER "user,created"
#define DEFAULT_CATEGORY "default"

#define WIDGET_INFO_TAG "__WIDGET_INFO__"
#define ERR_BUFFER_SIZE		1024

#define WIDGET_PRIVILEGE "http://tizen.org/privilege/widget.viewer"
#define WIDGET_LAUNCH_TIMEOUT	7000
#define MAX_NO_RESPONSE_COUNT 5

/*!
 * \note
 * Enable this to apply shadow effect to image object (for text widget)
 * #define SUPPORT_IMAGE_EFFECT
 */

enum {
	WIDGET_STATE_DETACHED,
	WIDGET_STATE_ATTACHED
};

int errno;

static struct info {
	int w;
	int h;
	bool initialized;
	Evas_Object *win;
	GHashTable *widget_table;
	GHashTable *instance_cnt_table;
	char *launching_instance_id;
	GList *launch_list;
	int launch_timer;
} s_info = {
	.w = 0,
	.h = 0,
	.initialized = false,
	.win = NULL,
	.widget_table = NULL,
	.instance_cnt_table = NULL,
	.launching_instance_id = NULL,
	.launch_list = NULL,
	.launch_timer = 0
};

struct pending_item {
	char *event;
	int w;
	int h;
};

struct widget_info {
	char *widget_id;
	char *instance_id;
	const char *title;
	char *content_info;
	int pid;
	double period;

	int disable_preview;
	int disable_loading;
	int disable_overlay;
	int permanent_delete;
	int visibility_freeze;
	int state;
	int cancel_click;
	bool restart;

	GQueue *event_queue;

	Evas_Object *layout;

	bool is_created;
	GQueue *pending_queue;
	bool is_faulted;
	screen_connector_toolkit_evas_h toolkit_h;
	int no_response_count;
};

typedef enum widget_fault_type {
      WIDGET_FAULT_DEACTIVATED,
      WIDGET_FAULT_PROVIDER_DISCONNECTED,
      WIDGET_FAULT_MAX = 0xFF
} widget_fault_type_e;

static void __flush_event_queue(struct widget_info *info);
static int __default_delayed_resuming_time = 100;
static bool __is_delayed_resuming_mode;
static void __destroy_widget_info(gpointer data);
static gboolean __launch_timeout_cb(gpointer user_data);
static int __find_launch_info(gconstpointer a, gconstpointer b);
static void __remove_launch_info(const char *instance_id);
static void __launch_instance(void);
static void __remove_launch_timer(void);

static void __destroy_pending_item(gpointer data)
{
	struct pending_item *item = (struct pending_item *)data;

	free(item->event);
	free(item);
}

static struct pending_item *__create_pending_item(const char *event,
		int w, int h)
{
	struct pending_item *item;

	item = calloc(1, sizeof(struct pending_item));
	if (!item) {
		LOGE("Out of memory");/* LCOV_EXCL_LINE */
		return NULL;/* LCOV_EXCL_LINE */
	}

	item->event = strdup(event);
	if (!item->event) {
		LOGE("Out of memory");/* LCOV_EXCL_LINE */
		free(item);/* LCOV_EXCL_LINE */
		return NULL;/* LCOV_EXCL_LINE */
	}

	item->w = w;
	item->h = h;

	return item;
}

static int __instance_pause(struct widget_info *info)
{
	Evas_Object *tbm_obj;

	if (!info) {
		LOGE("instance id is invalid");/* LCOV_EXCL_LINE */
		return -1;/* LCOV_EXCL_LINE */
	}

	if (!info->layout) {
		LOGE("invalid layout object");/* LCOV_EXCL_LINE */
		return -1;/* LCOV_EXCL_LINE */
	}

	tbm_obj = elm_object_part_content_get(info->layout, "tbm,widget");
	if (!tbm_obj) {
		LOGE("widget object is invalid");/* LCOV_EXCL_LINE */
		return -1;/* LCOV_EXCL_LINE */
	}

	return screen_connector_toolkit_evas_send_visibility(tbm_obj,
			VISIBILITY_TYPE_FULLY_OBSCURED);
}

static int __instance_resume(struct widget_info *info)
{
	Evas_Object *tbm_obj;

	if (!info) {
		LOGE("instance id is invalid");/* LCOV_EXCL_LINE */
		return -1;/* LCOV_EXCL_LINE */
	}

	if (!info->layout) {
		LOGE("invalid layout object");/* LCOV_EXCL_LINE */
		return -1;/* LCOV_EXCL_LINE */
	}

	tbm_obj = elm_object_part_content_get(info->layout, "tbm,widget");
	if (!tbm_obj) {
		LOGE("widget object is invalid");/* LCOV_EXCL_LINE */
		return -1;/* LCOV_EXCL_LINE */
	}

	return screen_connector_toolkit_evas_send_visibility(tbm_obj,
			VISIBILITY_TYPE_UNOBSCURED);
}

/* LCOV_EXCL_START */
static void __process_pending_item(struct pending_item *item, struct widget_info *info)
{
	int ret;

	if (!strcmp(item->event, "resize")) {
		ret = widget_instance_resize(info->instance_id, item->w, item->h);
	} else if (!strcmp(item->event, "resume")) {
		ret = __instance_resume(info);
	} else if (!strcmp(item->event, "pause")) {
		ret = __instance_pause(info);
	} else {
		LOGW("Unknown event(%s)", item->event);
		return;
	}

	LOGD("instance_id(%s), event(%s), result(%d)",
			info->instance_id, item->event, ret);
}
/* LCOV_EXCL_STOP */

static void __flush_pending_queue(struct widget_info *info)
{
	struct pending_item *item;

	if (!info->pending_queue) {
		LOGD("Pending queue is empty");
		return;
	}

	while (!g_queue_is_empty(info->pending_queue)) {
		item = g_queue_pop_head(info->pending_queue);
		__process_pending_item(item, info);
		__destroy_pending_item(item);
	}

	g_queue_free(info->pending_queue);
	info->pending_queue = NULL;
}

static void __push_pending_queue(struct widget_info *info,
		struct pending_item *item)
{
	if (!item) {
		LOGE("Invalid parameter");/* LCOV_EXCL_LINE */
		return;/* LCOV_EXCL_LINE */
	}

	if (!info->pending_queue)
		info->pending_queue = g_queue_new();

	if (!info->pending_queue) {
		LOGE("Out of memory");/* LCOV_EXCL_LINE */
		return;/* LCOV_EXCL_LINE */
	}

	g_queue_push_tail(info->pending_queue, item);
	LOGD("instance_id(%s), event(%s)", info->instance_id, item->event);
}

static void __add_launch_list(const char *instance_id)
{
	GList *list;
	char *dup_instance_id;

	if (!instance_id) {
		LOGE("Invalid parameter");/* LCOV_EXCL_LINE */
		return;/* LCOV_EXCL_LINE */
	}

	list = g_list_find_custom(s_info.launch_list, instance_id, __find_launch_info);
	if (list != NULL) {
		LOGW("already in launch list(%s)", instance_id);
		return;
	}

	dup_instance_id = strdup(instance_id);
	if (dup_instance_id == NULL) {
		LOGE("Out Of Memory");
		return;
	}

	s_info.launch_list = g_list_append(s_info.launch_list, dup_instance_id);
	LOGW("add launch(%s)", dup_instance_id);
}

static inline bool __is_widget_feature_enabled(void)
{
	static bool feature = false;
	static bool retrieved = false;
	int ret;

	if (retrieved == true)
		return feature;

	ret = system_info_get_platform_bool(
			"http://tizen.org/feature/shell.appwidget", &feature);
	if (ret != SYSTEM_INFO_ERROR_NONE) {
		LOGE("failed to get system info");/* LCOV_EXCL_LINE */
		return false;/* LCOV_EXCL_LINE */
	}

	retrieved = true;

	return feature;
}

#define SMACK_LABEL_LEN 255
static int __check_privilege(const char *privilege)
{
	cynara *p_cynara;

	int fd = 0;
	int ret = 0;

	char subject_label[SMACK_LABEL_LEN +1] = "";
	char uid[10] = {0, };
	char *client_session = "";

	ret = cynara_initialize(&p_cynara, NULL);
	if (ret != CYNARA_API_SUCCESS)
		return -1;

	fd = open("/proc/self/attr/current", O_RDONLY);
	if (fd < 0) {
		ret = -1;
		goto ERROR;
	}

	ret = read(fd, subject_label, SMACK_LABEL_LEN);
	if (ret < 0) {
		LOGE("read is failed");/* LCOV_EXCL_LINE */
		close(fd);/* LCOV_EXCL_LINE */
		goto ERROR;
	}
	close(fd);

	snprintf(uid, 10, "%d", getuid());

	ret = cynara_check(p_cynara, subject_label, client_session, uid, privilege);
	if (ret != CYNARA_API_ACCESS_ALLOWED) {
		ret = -1;
		goto ERROR;
	}

	ret = 0;

ERROR:
	if (p_cynara)
		cynara_finish(p_cynara);
	return ret;
}

static void __smart_callback_call(Evas_Object *obj, const char *signal, void *cbdata)
{
	if (!obj) {
		LOGD("widget is deleted, ignore smart callback call");/* LCOV_EXCL_LINE */
		return;/* LCOV_EXCL_LINE */
	}
	evas_object_smart_callback_call(obj, signal, cbdata);
}

static void __screen_connector_toolkit_evas_added_cb(const char *appid, const char *instance_id, int pid,
	Evas_Object *image, void *data)
{
	struct widget_info *info;
	int x, y, w, h;
	Evas_Object *preview;

	if (!image) {
		LOGE("Invalid parameters");/* LCOV_EXCL_LINE */
		return;/* LCOV_EXCL_LINE */
	}

	info = (struct widget_info *)data;
	if (!info) {
		LOGE("Unable to find a proper object");/* LCOV_EXCL_LINE */
		evas_object_del(image);/* LCOV_EXCL_LINE */
		return;
	}

	if ((info->instance_id == NULL) || (strcmp(info->instance_id, instance_id) != 0)) {
		LOGE("Wrong added info, appid(%s) info instance_id(%s), instance_id(%s)",/* LCOV_EXCL_LINE */
				appid, info->instance_id, instance_id);/* LCOV_EXCL_LINE */

		if (s_info.launching_instance_id && info->instance_id &&
				strcmp(s_info.launching_instance_id, info->instance_id) == 0) {
			__remove_launch_timer();
			__launch_instance();
		}
		return;/* LCOV_EXCL_LINE */
	}

	aul_widget_write_log(LOG_TAG, "[%s:%d]  %s",
			__FUNCTION__, __LINE__, info->instance_id);
	LOGD("widget added: %s, widget_id : %s, instance : %s, (%p)",
			appid, info->widget_id, info->instance_id, image);
	evas_object_geometry_get(image, &x, &y, &w, &h);
	LOGD("widget geometry:%d %d %d %d", x, y, w, h);
	elm_object_part_content_set(info->layout, "tbm,widget", image);

	if (!info->disable_preview) {
		elm_object_signal_emit(info->layout, "disable", "preview");
		preview = elm_object_part_content_unset(info->layout, "preview");
		if (preview)
			evas_object_del(preview);
	}

	if (!info->disable_loading)
		elm_object_signal_emit(info->layout, "disable", "loading,text");
	else
		elm_object_signal_emit(info->layout, "disable", "overlay,text");

	info->state = WIDGET_STATE_ATTACHED;
	info->no_response_count = 0;
	if (s_info.launching_instance_id &&
            strcmp(s_info.launching_instance_id, instance_id) == 0) {
		__remove_launch_info(instance_id);
		__remove_launch_timer();
		__launch_instance();
	}
	__flush_event_queue(info);
	/**
	 * @note
	 * After swallow this widget object to the layout,
	 * It will be automatically resized by EDJE.
	 */
}

static void __create_preview_image(Evas_Object *layout, char *widget_id)
{
	LOGI("create preview image");
	Evas_Object *preview = NULL;
	int x, y, w, h;
	char *preview_path = NULL;
	widget_size_type_e size_type;

	if (layout == NULL) {
		LOGE("layout is NULL");
		return;
	}

	preview = elm_object_part_content_unset(layout, "preview");
	if (preview)
		evas_object_del(preview);

	evas_object_geometry_get(layout, &x, &y, &w, &h);
	widget_service_get_size_type(w, h, &size_type);
	preview_path = widget_service_get_preview_image_path(widget_id, size_type);
	if (preview_path == NULL) {
		LOGE("can not get preview path");
		return;
	}

	preview = elm_image_add(layout);
	if (preview == NULL) {
		LOGE("can not get preview image");
		free(preview_path);
		return;
	}

	elm_image_file_set(preview, preview_path, NULL);
	elm_object_part_content_set(layout, "preview", preview);
	free(preview_path);
}

static void __screen_connector_toolkit_evas_removed_cb(const char *appid, const char *instance_id, int pid,
	Evas_Object *image, void *data)
{
	struct widget_evas_event_info event_info;
	struct widget_info *info;

	info = (struct widget_info *)data;
	if (!info) {
		LOGE("Unable to find a proper object");/* LCOV_EXCL_LINE */
		return;/* LCOV_EXCL_LINE */
	}

	aul_widget_write_log(LOG_TAG, "[%s:%d]  %s, restart[%d]",
			__FUNCTION__, __LINE__, instance_id, info->restart);

	if (info->restart)
		return;

	LOGD("widget removed: %s", appid);
	elm_object_part_content_set(info->layout, "tbm,widget", NULL);

	if (info->disable_preview) {
		__create_preview_image(info->layout, info->widget_id);
		elm_object_signal_emit(info->layout, "enable", "preview");
	}

	if (info->disable_loading)
		elm_object_signal_emit(info->layout, "enable", "loading,text");

	event_info.error =  WIDGET_ERROR_NONE;
	event_info.widget_app_id = info->widget_id;
	event_info.event = WIDGET_EVENT_DELETED;

	__smart_callback_call(info->layout, WIDGET_SMART_SIGNAL_WIDGET_DELETED, &event_info);
}

/* LCOV_EXCL_START */
static void __screen_connector_toolkit_evas_updated_cb(const char *appid, const char *instance_id, int pid,
	Evas_Object *image, void *data)
{
}
/* LCOV_EXCL_STOP */

/* LCOV_EXCL_START */
static void __display_overlay_text(struct widget_info *info)
{
	Evas_Object *text_block;
	Evas_Object *access_object;

	if (!info) {
		LOGE("Unable to get the info");
		return;
	}
	if (!info->disable_overlay) {
		elm_object_part_text_set(info->layout, "text",
				T_("IDS_HS_BODY_UNABLE_TO_LOAD_DATA_TAP_TO_RETRY"));
		elm_object_signal_emit(info->layout, "enable", "overlay,text");
		elm_object_signal_emit(info->layout, "disable", "preview");
		text_block = (Evas_Object*)edje_object_part_object_get(
				elm_layout_edje_get(info->layout), "text");
		if (text_block) {
			access_object = elm_access_object_register(
					text_block, info->layout);
			elm_atspi_accessible_description_set(access_object,
					T_("IDS_HS_BODY_UNABLE_TO_LOAD_DATA_TAP_TO_RETRY"));
			elm_atspi_accessible_reading_info_type_set(access_object,
					ELM_ACCESSIBLE_READING_INFO_TYPE_DESCRIPTION);
		}
	}
}
/* LCOV_EXCL_STOP */

/* LCOV_EXCL_START */
static void __display_loading_text(struct widget_info *info)
{
	if (!info->disable_preview) {
		__create_preview_image(info->layout, info->widget_id);
		elm_object_signal_emit(info->layout, "enable", "preview");
	}

	if (!info->disable_loading) {
		elm_object_part_text_set(info->layout, "text", T_("IDS_ST_POP_LOADING_ING"));
		elm_object_signal_emit(info->layout, "enable", "text");
	}
}
/* LCOV_EXCL_STOP */

static void __push_event_queue(struct widget_info *info, int event)
{
	if (info->event_queue == NULL)
		info->event_queue = g_queue_new();

	if (info->event_queue == NULL) {
		LOGE("out of memory");
		return;
	}

	LOGD("add event to queue:%d", event);
	g_queue_push_tail(info->event_queue, GINT_TO_POINTER(event));
}

static int __find_launch_info(gconstpointer a, gconstpointer b)
{
	if (a == NULL || b == NULL)
		return -1;

	if (strcmp((const char *)a, (const char *)b) == 0)
		return 0;

	return -1;
}

static void __remove_launch_info(const char *instance_id)
{
	GList *list;
	char *launch_instance_id;

	list = g_list_find_custom(s_info.launch_list, instance_id, __find_launch_info);
	if (list == NULL)
		return;
	launch_instance_id = list->data;
	s_info.launch_list = g_list_remove(s_info.launch_list, launch_instance_id);
	free(launch_instance_id);
}

static void __remove_launch_timer()
{
	if (s_info.launch_timer <= 0)
		return;

	g_source_remove(s_info.launch_timer);
	s_info.launch_timer = 0;
}

static void __launch_instance()
{
	struct widget_info *info;
	char *instance_id;
	int w, h;
	int ret;
	screen_connector_toolkit_evas_ops ops;

	if (s_info.launch_list == NULL) {
		s_info.launch_timer = 0;
		s_info.launching_instance_id = NULL;
		LOGW("null restart list");
		return;
	}

	instance_id = s_info.launch_list->data;
	info = g_hash_table_lookup(s_info.widget_table, instance_id);
	if (info == NULL) {
		LOGW("can not find instance (%s)", instance_id);
		return;
	}
	evas_object_geometry_get(info->layout, NULL, NULL, &w, &h);
	LOGD("Widget launch  %s, %d, %d", info->instance_id, w, h);

	ops.added_cb = __screen_connector_toolkit_evas_added_cb;
	ops.removed_cb = __screen_connector_toolkit_evas_removed_cb;
	ops.updated_cb = __screen_connector_toolkit_evas_updated_cb;

	if (info->toolkit_h) {
		screen_connector_toolkit_evas_remove(info->toolkit_h);
		info->toolkit_h = NULL;
	}

	info->toolkit_h = screen_connector_toolkit_evas_add(&ops,
			info->instance_id,
			SCREEN_CONNECTOR_SCREEN_TYPE_WIDGET,
			info);

	if (info->restart) {
		ret = widget_instance_terminate(info->instance_id);
		if (ret != 0)
			LOGW("widget_instance_terminate failed: %d", ret);
	}

	__display_loading_text(info);
	info->pid = widget_instance_launch(info->instance_id, info->content_info, w, h);
	info->restart = false;
	if (info->pid < 0) {
		/* LCOV_EXCL_START */
		struct widget_evas_event_info event_info;
		LOGE("Failed to launch an widget");
		event_info.error = info->pid;
		event_info.widget_app_id = info->widget_id;
		event_info.event = WIDGET_EVENT_CREATED;

		__smart_callback_call(info->layout, WIDGET_SMART_SIGNAL_WIDGET_CREATE_ABORTED, &event_info);
		__display_overlay_text(info);
		__remove_launch_info(info->instance_id);
		__launch_instance();
		/* LCOV_EXCL_STOP */
	} else {
		s_info.launching_instance_id = info->instance_id;
		s_info.launch_timer = g_timeout_add(WIDGET_LAUNCH_TIMEOUT,
				__launch_timeout_cb, info->instance_id);
	}
}

static void __set_faulted(const char *instance_id)
{
	struct widget_info *info;

	info = g_hash_table_lookup(s_info.widget_table, instance_id);
	if (!info) {
		aul_widget_write_log(LOG_TAG, "[%s:%d] fail to find info (%s)",
			__FUNCTION__, __LINE__, instance_id);
		return;
	}
	aul_widget_write_log(LOG_TAG, "[%s:%d]  set faulted for (%s)",
			__FUNCTION__, __LINE__, info->instance_id);
	info->pid = -1;
	info->is_faulted = true;
	__display_overlay_text(info);
}

static int __clear_launch_waiting(char *instance_id)
{
	struct widget_info *info;
	int ret;

	LOGW("Clear launch waiting !! %s", instance_id);
	info = g_hash_table_lookup(s_info.widget_table, instance_id);
	if (!info)
		LOGE("Unable to find a proper instance");/* LCOV_EXCL_LINE */

	__remove_launch_info(instance_id);
	if (info && info->no_response_count < MAX_NO_RESPONSE_COUNT) {
		ret = aul_terminate_pid(info->pid);
		if (ret != AUL_R_OK) {
			LOGW("Fail to term widget(%d) pid(%d) !!",
				ret, info->pid);
			return 0;
		}
		info->no_response_count++;
		LOGW("Widget is not responding(%d) !! %s(%d)",
				info->no_response_count, instance_id, info->pid);
	}
	return 0;
}

/* LCOV_EXCL_START */
static int __restart_terminated_widget(const char *widget_id)
{
	GHashTableIter iter;
	gpointer key, value;
	struct widget_info *widget_instance_info;
	int target_pid = 0;
	char *instance_id;
	int pid;
	char *id;

	g_hash_table_iter_init(&iter, s_info.widget_table);
	while (g_hash_table_iter_next(&iter, &key, &value)) {
		widget_instance_info = (struct widget_info *)value;
		if (strcmp(widget_instance_info->widget_id, widget_id) != 0)
			continue;
		if (widget_instance_info->restart) {
			target_pid = widget_instance_info->pid;
			LOGW("Prepare restart (%s)(%d)", widget_id, target_pid);
			break;
		} else {
			LOGW("There is no restart event for (%s)",
					widget_instance_info->instance_id);
		}
	}

	if (target_pid == 0) {
		LOGW("can not find widget(%s)", widget_id);
		return 0;
	}

	if (widget_instance_info->no_response_count >= MAX_NO_RESPONSE_COUNT) {
		LOGE("WIDGET IS NOT RESPONDING(%s) STOP AUTO RESTART", widget_id);
		widget_instance_info->no_response_count = 0;
		aul_widget_write_log(LOG_TAG, "[%s:%d]  not responding (%s)",
			__FUNCTION__, __LINE__, widget_id);
		return 0;
	}

	g_hash_table_iter_init(&iter, s_info.widget_table);
	while (g_hash_table_iter_next(&iter, &key, &value)) {
		widget_instance_info = (struct widget_info *)value;
		pid = widget_instance_info->pid;
		instance_id = widget_instance_info->instance_id;
		id = widget_instance_info->widget_id;
		if (!(widget_instance_info->restart && pid == target_pid &&
			(strcmp(id, widget_id) == 0)))
			continue;
		if (s_info.launching_instance_id &&
			strcmp(s_info.launching_instance_id, instance_id) == 0) {
			__remove_launch_timer();
			__clear_launch_waiting(instance_id);
		}
		__add_launch_list(instance_id);
	}
	if (s_info.launch_timer == 0)
		__launch_instance();

	return 0;
}
/* LCOV_EXCL_STOP */

static gboolean __launch_timeout_cb(gpointer user_data)
{
	char *instance_id = (char *)user_data;

	LOGW("Timeout called !! %s", instance_id);
	__clear_launch_waiting(instance_id);
	__set_faulted(instance_id);
	__launch_instance();
	return G_SOURCE_REMOVE;
}

/* LCOV_EXCL_START */
static int __handle_restart_widget_request(const char *widget_id)
{
	GHashTableIter iter;
	gpointer key, value;
	struct widget_info *widget_instance_info;
	int target_pid = 0;

	g_hash_table_iter_init(&iter, s_info.widget_table);
	while (g_hash_table_iter_next(&iter, &key, &value)) {
		widget_instance_info = (struct widget_info *)value;
		if (strcmp(widget_instance_info->widget_id, widget_id) == 0) {
			target_pid = widget_instance_info->pid;
			break;
		}
	}

	if (target_pid == 0) {
		LOGW("can not find widget (%s)", widget_id);
		return 0;
	}

	g_hash_table_iter_init(&iter, s_info.widget_table);
	while (g_hash_table_iter_next(&iter, &key, &value)) {
		widget_instance_info = (struct widget_info *)value;
		if (widget_instance_info->pid == target_pid)
			widget_instance_info->restart = true;
	}
	aul_terminate_pid_async(target_pid);
	LOGW("Widget __handle_restart_widget_request id : %s, pid : %d",
			widget_id, target_pid);
	aul_widget_write_log(LOG_TAG, "[%s:%d]  %s, targetpid : %d",
			__FUNCTION__, __LINE__, widget_id, target_pid);

	return 0;
}
/* LCOV_EXCL_STOP */

static int __instance_event_cb(const char *widget_id, const char *instance_id, int event, void *data)
{
	struct widget_info *info;
	struct widget_evas_event_info event_info;
	const char *smart_signal = NULL;
	char *content_info;
	widget_instance_h handle;
	int err;

	LOGD("widget_id : %s (%d)", widget_id, event);
	if (event == WIDGET_INSTANCE_EVENT_APP_RESTART_REQUEST) {
		__handle_restart_widget_request(widget_id);
		return 0;
	}

	info = g_hash_table_lookup(s_info.widget_table, instance_id);
	if (!info) {
		LOGE("Unable to find a proper object");/* LCOV_EXCL_LINE */
		return -1;/* LCOV_EXCL_LINE */
	}

	if (event != WIDGET_INSTANCE_EVENT_FAULT &&
			event != WIDGET_INSTANCE_EVENT_CREATE_ABORTED) {
		if (info->state == WIDGET_STATE_DETACHED) {
			__push_event_queue(info, event);
			return 0;
		}
	}

	LOGD("update: %s (%d)", instance_id, event);

	event_info.error = WIDGET_ERROR_NONE;
	event_info.widget_app_id = info->widget_id;

	switch (event) {
	case WIDGET_INSTANCE_EVENT_CREATE:
		event_info.event = WIDGET_EVENT_CREATED;
		smart_signal = WIDGET_SMART_SIGNAL_WIDGET_CREATED;
		info->is_created = true;
		__flush_pending_queue(info);
		break;
	case WIDGET_INSTANCE_EVENT_UPDATE:
		event_info.event = WIDGET_EVENT_WIDGET_UPDATED;
		smart_signal = WIDGET_SMART_SIGNAL_UPDATED;
		break;
	case WIDGET_INSTANCE_EVENT_PERIOD_CHANGED:
		event_info.event = WIDGET_EVENT_PERIOD_CHANGED;
		smart_signal = WIDGET_SMART_SIGNAL_PERIOD_CHANGED;
		break;
	case WIDGET_INSTANCE_EVENT_SIZE_CHANGED:
		event_info.event = WIDGET_EVENT_WIDGET_SIZE_CHANGED;
		smart_signal = WIDGET_SMART_SIGNAL_WIDGET_RESIZED;
		break;
	case WIDGET_INSTANCE_EVENT_EXTRA_UPDATED:
		event_info.event = WIDGET_EVENT_EXTRA_INFO_UPDATED;
		smart_signal = WIDGET_SMART_SIGNAL_EXTRA_INFO_UPDATED;
		handle = widget_instance_get_instance(info->widget_id, info->instance_id);
		if (!handle) {
			LOGE("instance handle of widget(%s) is invalid data", info->instance_id);
			break;
		}

		if (widget_instance_get_content(handle, &content_info) < 0) {
			LOGE("Failed to get content of widget(%s)", info->instance_id);
			widget_instance_unref(handle);
			break;
		}

		if (content_info) {
			if (info->content_info)
				free(info->content_info);

			info->content_info = strdup(content_info);
		}
		widget_instance_unref(handle);
		break;
	case WIDGET_INSTANCE_EVENT_FAULT:
		LOGW("widget faulted (%s)", instance_id);
		event_info.error = WIDGET_ERROR_FAULT;
		event_info.event = WIDGET_FAULT_DEACTIVATED;
		smart_signal = WIDGET_SMART_SIGNAL_WIDGET_FAULTED;
		info->state = WIDGET_STATE_DETACHED;
		__set_faulted(info->instance_id);
		break;
	case WIDGET_INSTANCE_EVENT_CREATE_ABORTED:
		LOGW("widget aborted (%s)", instance_id);
		handle = widget_instance_get_instance(info->widget_id,
				info->instance_id);
		if (!handle) {
			LOGE("instance handle of widget(%s) is invalid data",/* LCOV_EXCL_LINE */
					info->instance_id);/* LCOV_EXCL_LINE */
			break;/* LCOV_EXCL_LINE */
		}

		widget_instance_get_error_code(handle, &err);
		widget_instance_unref(handle);
		if (err == 0)
			err = WIDGET_ERROR_CANCELED;

		event_info.error = err;
		event_info.event = WIDGET_EVENT_CREATED;
		smart_signal = WIDGET_SMART_SIGNAL_WIDGET_CREATE_ABORTED;
		__set_faulted(info->instance_id);
		break;
	default:
		/* unhandled event */
		return 0;
	}

	if (!info->layout) {
		LOGE("object is not ready");
		return 0;
	}

	__smart_callback_call(info->layout, smart_signal, &event_info);

	return 0;
}

static int __lifecycle_event_cb(const char *widget_id, widget_lifecycle_event_e lifecycle_event, const char *widget_instance_id, void *data)
{
	LOGD("__lifecycle_event_cb %s (%d)", widget_id, lifecycle_event);
	if (lifecycle_event == WIDGET_LIFE_CYCLE_EVENT_APP_DEAD) {
		__restart_terminated_widget(widget_id);
	}
	return 0;
}

static void __flush_event_queue(struct widget_info *info)
{
	if (!info->event_queue) {
		LOGD("event queue is empty. nothing to flush");
		return;
	}

	while (!g_queue_is_empty(info->event_queue)) {
		int event = GPOINTER_TO_INT(g_queue_pop_head(info->event_queue));
		LOGD("call pending event %d", event);
		__instance_event_cb(info->widget_id, info->instance_id, event, NULL);
	}

	g_queue_free(info->event_queue);
	info->event_queue = NULL;
}

API int widget_viewer_evas_init(Evas_Object *win)
{
	char app_id[255];

	if (!__is_widget_feature_enabled())
		return WIDGET_ERROR_NOT_SUPPORTED;

	if (__check_privilege(WIDGET_PRIVILEGE) < 0)
		return WIDGET_ERROR_PERMISSION_DENIED;

	if (!win) {
		LOGE("win object is invalid");
		return WIDGET_ERROR_INVALID_PARAMETER;
	}

	if (s_info.initialized) {
		LOGE("Already initialized");/* LCOV_EXCL_LINE */
		return WIDGET_ERROR_ALREADY_EXIST;
	}

	if (!bindtextdomain(PKGNAME, WIDGET_VIEWER_EVAS_RESOURCE_PO))
		LOGE("bindtextdomain: %d", errno);
	else
		LOGD("%s - %s", PKGNAME, WIDGET_VIEWER_EVAS_RESOURCE_PO);

	s_info.win = win;

	screen_connector_toolkit_evas_init(win, SCREEN_CONNECTOR_SCREEN_TYPE_WIDGET);
	screen_connector_toolkit_evas_start_visibility_notify();

	if (aul_app_get_appid_bypid(getpid(), app_id, sizeof(app_id)) != AUL_R_OK) {
		LOGE("failed to get appid of pid:%d", getpid());/* LCOV_EXCL_LINE */
		return WIDGET_ERROR_INVALID_PARAMETER;
	}

	widget_instance_init(app_id);
	widget_instance_listen_event(__instance_event_cb, NULL);

	s_info.initialized = true;

	s_info.widget_table = g_hash_table_new_full(g_str_hash, g_str_equal, NULL, __destroy_widget_info);
	s_info.instance_cnt_table = g_hash_table_new_full(g_str_hash, g_str_equal, free, free);

	return WIDGET_ERROR_NONE;
}

API int widget_viewer_evas_fini(void)
{
	if (!__is_widget_feature_enabled())
		return WIDGET_ERROR_NOT_SUPPORTED;

	if (__check_privilege(WIDGET_PRIVILEGE) < 0)
		return WIDGET_ERROR_PERMISSION_DENIED;

	if (!s_info.initialized) {
		LOGE("widget viewer evas is not initialized");
		return WIDGET_ERROR_FAULT;
	}

	if (s_info.widget_table)
		g_hash_table_destroy(s_info.widget_table);

	if (s_info.instance_cnt_table)
		g_hash_table_destroy(s_info.instance_cnt_table);

	screen_connector_toolkit_evas_fini(SCREEN_CONNECTOR_SCREEN_TYPE_WIDGET);
	widget_instance_unlisten_event(__instance_event_cb);
	widget_instance_fini();

	s_info.initialized = false;

	return WIDGET_ERROR_NONE;
}

API int widget_viewer_evas_notify_resumed_status_of_viewer(void)
{
	GHashTableIter iter;
	gpointer key, value;
	struct widget_info *info;
	struct pending_item *item;
	int ret;
	Evas_Object *tbm_obj;
	bool is_visible;

	if (!__is_widget_feature_enabled())
		return WIDGET_ERROR_NOT_SUPPORTED;

	if (!s_info.initialized) {
		LOGE("widget viewer evas is not initialized");
		return WIDGET_ERROR_FAULT;
	}

	if (__check_privilege(WIDGET_PRIVILEGE) < 0)
		return WIDGET_ERROR_PERMISSION_DENIED;

	g_hash_table_iter_init(&iter, s_info.widget_table);
	while (g_hash_table_iter_next(&iter, &key, &value)) {
		info = (struct widget_info *)value;
		if (info) {
			if (info->is_created) {

				if (!info->layout) {
					LOGE("widget(%s) layout is invalid", info->instance_id);
					continue;
				}

				tbm_obj = elm_object_part_content_get(info->layout, "tbm,widget");
				if (!tbm_obj) {
					LOGE("widget(%s)(%d) object is invalid",
							info->instance_id, info->pid);
					continue;
				}

				ret = screen_connector_toolkit_evas_is_visible(tbm_obj, &is_visible);
				if (ret < 0 || !is_visible) {
					LOGE("widget(%s) object is invisible", info->instance_id);
					continue;
				}

				ret = __instance_resume(info);
				if (ret < 0) {
					LOGE("Fail to resume the widget(%s):(%d)",
							info->instance_id, ret);
				}
			} else {
				item = __create_pending_item("resume", 0, 0);
				__push_pending_queue(info, item);
			}
		}
	}

	return WIDGET_ERROR_NONE;
}

API int widget_viewer_evas_notify_paused_status_of_viewer(void)
{
	GHashTableIter iter;
	gpointer key, value;
	struct widget_info *info;
	struct pending_item *item;
	int ret;

	if (!__is_widget_feature_enabled())
		return WIDGET_ERROR_NOT_SUPPORTED;

	if (!s_info.initialized) {
		LOGE("widget viewer evas is not initialized");
		return WIDGET_ERROR_FAULT;
	}

	if (__check_privilege(WIDGET_PRIVILEGE) < 0)
		return WIDGET_ERROR_PERMISSION_DENIED;

	g_hash_table_iter_init(&iter, s_info.widget_table);
	while (g_hash_table_iter_next(&iter, &key, &value)) {
		info = (struct widget_info *)value;
		if (info) {
			if (info->is_created) {
				ret = __instance_pause(info);
				if (ret < 0) {
					LOGE("Fail to pause the widget(%s)(%d):(%d)",
							info->instance_id, info->pid, ret);
				}
			} else {
				item = __create_pending_item("pause", 0, 0);
				__push_pending_queue(info, item);
			}
		}
	}

	return WIDGET_ERROR_NONE;
}

API int widget_viewer_evas_notify_orientation_of_viewer(int orientation)
{
	if (!__is_widget_feature_enabled())
		return WIDGET_ERROR_NOT_SUPPORTED;

	if (!s_info.initialized) {
		LOGE("widget viewer evas is not initialized");
		return WIDGET_ERROR_FAULT;
	}

	if (__check_privilege(WIDGET_PRIVILEGE) < 0)
		return WIDGET_ERROR_PERMISSION_DENIED;

	if (orientation < 0 || orientation > 360) {
		LOGE("orientation is invalid parameter");
		return WIDGET_ERROR_INVALID_PARAMETER;
	}

	return WIDGET_ERROR_NONE;
}

static void __del_cb(void *data, Evas *e, Evas_Object *layout, void *event_info)
{
	struct widget_info *info = data;
	struct widget_evas_event_info evas_info;
	int *cur_instance_cnt;
	char *widget_id;
	char *instance_id;
	Evas_Object *text_block;

	LOGD("delete: layout(%p)", layout);
	aul_widget_write_log(LOG_TAG, "[%s:%d]  %s",
			__FUNCTION__, __LINE__, info->instance_id);

	evas_info.error = WIDGET_ERROR_NONE;
	evas_info.widget_app_id = info->widget_id;
	evas_info.event = WIDGET_EVENT_CREATED;

	text_block = (Evas_Object*)edje_object_part_object_get(elm_layout_edje_get(layout), "text");
	if (text_block)
		elm_access_object_unregister(text_block);

	__smart_callback_call(layout, WIDGET_SMART_SIGNAL_WIDGET_DELETED, &evas_info);
	evas_object_data_del(layout, WIDGET_INFO_TAG);
	info->layout = NULL;
	if (info->event_queue) {
		g_queue_free(info->event_queue);
		info->event_queue = NULL;
	}

	widget_id = strdup(info->widget_id);
	if (widget_id == NULL) {
		LOGE("out of memory");
		return;
	}

	instance_id = strdup(info->instance_id);
	if (instance_id == NULL) {
		LOGE("out of memory");
		free(widget_id);
		return;
	}

	/* destroy widget info first by removing from hashtable. It will send destroy command to amd */
	if (s_info.widget_table)
		g_hash_table_remove(s_info.widget_table, instance_id);

	if (s_info.instance_cnt_table) {
		cur_instance_cnt = g_hash_table_lookup(s_info.instance_cnt_table, widget_id);
		if (cur_instance_cnt != NULL) {
			(*cur_instance_cnt)--;
			LOGW("instance %s, (%d)",
					widget_id, *cur_instance_cnt);

			/* delete instance after sending destroy command */
			aul_widget_instance_del(widget_id, instance_id);
			if (*cur_instance_cnt == 0)
				g_hash_table_remove(s_info.instance_cnt_table, widget_id);
		}
	}
	__remove_launch_info(instance_id);

	free(widget_id);
	free(instance_id);
}

static void __resize_cb(void *data, Evas *e, Evas_Object *layout, void *event_info)
{
	struct widget_info *info = data;
	int x, y, w, h;
	struct pending_item *item;

	evas_object_geometry_get(layout, &x, &y, &w, &h);
	if (info->pid == 0) {
		/**
		 * @note
		 * Create a new instance in this case.
		 */
		__display_loading_text(info);
		__add_launch_list(info->instance_id);
		if (s_info.launch_timer == 0)
			__launch_instance();
	} else {
		/**
		 * @note
		 * Layout will be resized, consequently, the tbm object also will be resized.
		 */
		if (info->is_created) {
			LOGD("send resize : %s, w %d h %d", info->instance_id, w, h);
			widget_instance_resize(info->instance_id, w, h);
		} else {
			item = __create_pending_item("resize", w, h);
			__push_pending_queue(info, item);
		}
	}
}

/* LCOV_EXCL_START */
static bool __obj_is_visible(Evas_Object *layout)
{
	int x, y, w, h;
	Ecore_Wl2_Window *window = NULL;
	int window_x, window_y, window_w, window_h;
	int rotation;
	Ecore_Evas *ee = ecore_evas_ecore_evas_get(
		evas_object_evas_get(s_info.win));

	window = ecore_evas_wayland2_window_get(ee);
	ecore_wl2_window_geometry_get(window, &window_x, &window_y, &window_w, &window_h);
	evas_object_geometry_get(layout, &x, &y, &w, &h);
	rotation = ecore_wl2_window_rotation_get(window);

	if (x >= 0 && x < window_w &&
			y >= 0 && y < window_h &&
			(rotation == 0 || rotation == 180)) {
		LOGD("x %d, y %d w %d h %d, window_w %d window_h %d rotation %d",
				x, y, w, h, window_w, window_h, rotation);
		return true;
	} else if (x >= 0 && x < window_h &&
			y >= 0 && y < window_w &&
			(rotation == 90 || rotation == 270)) {
		LOGD("x %d, y %d w %d h %d, window_w %d window_h %d rotation %d",
				x, y, w, h, window_w, window_h, rotation);
		return true;
	}

	return false;
}
/* LCOV_EXCL_STOP */

/* LCOV_EXCL_START */
static void __move_cb(void *data, Evas *e, Evas_Object *layout, void *event_info)
{
	struct widget_info *info = data;

	if (!__obj_is_visible(layout) || info->pid > 0)
		return;

	info->restart = true;
	LOGW("restart! %s", info->instance_id);
	__display_loading_text(info);
	__restart_terminated_widget(info->widget_id);

}
/* LCOV_EXCL_STOP */

/* LCOV_EXCL_START */
static void _clicked_cb(void *data, Evas_Object *obj, const char *emission, const char *source)
{
	struct widget_info *info = data;

	if (info->cancel_click) {
		info->cancel_click = 0;
		return;
	}
	widget_viewer_evas_activate_faulted_widget(info->layout);
}
/* LCOV_EXCL_STOP */

static void __destroy_widget_info(gpointer data)
{
	struct widget_info *info = (struct widget_info *)data;
	if (!info)
		return;

	LOGD("__destroy_widget_info widget_id: %s instance_id %s", info->widget_id, info->instance_id);
	aul_widget_write_log(LOG_TAG, "[%s:%d]  %s",
			__FUNCTION__, __LINE__, info->instance_id);

	if (info->pending_queue)
		g_queue_free_full(info->pending_queue, __destroy_pending_item);

	if (info->layout) {
		evas_object_event_callback_del(info->layout, EVAS_CALLBACK_DEL, __del_cb);
		evas_object_event_callback_del(info->layout, EVAS_CALLBACK_RESIZE, __resize_cb);
		evas_object_event_callback_del(info->layout, EVAS_CALLBACK_MOVE, __move_cb);
		elm_object_signal_callback_del(info->layout, "clicked", "reload", _clicked_cb);
	}

	if (info->widget_id)
		free(info->widget_id);

	if (info->instance_id) {
		widget_instance_destroy(info->instance_id);
		free(info->instance_id);
	}

	if (info->content_info)
		free(info->content_info);

	screen_connector_toolkit_evas_remove(info->toolkit_h);

	free(info);
}

static struct widget_info *create_info(Evas_Object *parent, const char *widget_id, const char *instance_id, const char *content_info)
{
	struct widget_info *info;
	char err_buf[ERR_BUFFER_SIZE];

	info = (struct widget_info *)calloc(1, sizeof(*info));
	if (!info) {
		LOGE("malloc: %s", strerror_r(errno, err_buf, sizeof(err_buf)));/* LCOV_EXCL_LINE */
		goto out;/* LCOV_EXCL_LINE */
	}

	info->widget_id = strdup(widget_id);
	if (!info->widget_id)
		goto out;

	info->instance_id = strdup(instance_id);
	if (!info->instance_id)
		goto out;

	info->layout = elm_layout_add(parent);
	if (!info->layout)
		goto out;

	if (elm_layout_file_set(info->layout, WIDGET_VIEWER_EVAS_RESOURCE_EDJ, "layout") == EINA_FALSE) {
		evas_object_del(info->layout);/* LCOV_EXCL_LINE */
		goto out;
	}

	evas_object_data_set(info->layout, WIDGET_INFO_TAG, info);
	evas_object_event_callback_add(info->layout, EVAS_CALLBACK_RESIZE, __resize_cb, info);
	evas_object_event_callback_add(info->layout, EVAS_CALLBACK_MOVE, __move_cb, info);
	evas_object_event_callback_add(info->layout, EVAS_CALLBACK_DEL, __del_cb, info);
	elm_object_signal_callback_add(info->layout, "clicked", "reload", _clicked_cb, info);

	info->permanent_delete = 0;
	info->disable_preview = 0;
	info->disable_loading = 0;
	info->visibility_freeze = 0;
	info->cancel_click = 0;
	info->state = WIDGET_STATE_DETACHED;
	info->event_queue = NULL;
	info->restart = false;
	info->no_response_count = 0;

	return info;

out:
	if (info) {
		if (info->instance_id)
			free(info->instance_id);

		if (info->widget_id)
			free(info->widget_id);

		free(info);
	}

	return NULL;
}

API Evas_Object *widget_viewer_evas_add_widget(Evas_Object *parent, const char *widget_id, const char *content_info, double period)
{
	char *instance_id = NULL;
	struct widget_info *widget_instance_info = NULL;
	int max_instance_cnt;
	int *cur_instance_cnt;

	if (!__is_widget_feature_enabled()) {
		LOGE("widget is not supported");/* LCOV_EXCL_LINE */
		set_last_result(WIDGET_ERROR_NOT_SUPPORTED);/* LCOV_EXCL_LINE */
		return NULL;
	}

	if (__check_privilege(WIDGET_PRIVILEGE) < 0) {
		LOGE("Permission deny");/* LCOV_EXCL_LINE */
		set_last_result(WIDGET_ERROR_PERMISSION_DENIED);/* LCOV_EXCL_LINE */
		return NULL;
	}

	max_instance_cnt = widget_service_get_widget_max_count(widget_id);
	if (max_instance_cnt < 0) {
		set_last_result(WIDGET_ERROR_FAULT);/* LCOV_EXCL_LINE */
		LOGE("get widget max instance count fail");/* LCOV_EXCL_LINE */
		return NULL;
	}

	cur_instance_cnt = g_hash_table_lookup(s_info.instance_cnt_table, widget_id);
	if (cur_instance_cnt == NULL) {
		cur_instance_cnt = (int *)calloc(1, sizeof(int));
		if (cur_instance_cnt == NULL) {
			LOGE("Fail to calloc instance cnt");
			set_last_result(WIDGET_ERROR_FAULT);
			return NULL;
		}
		*cur_instance_cnt = 0;
		g_hash_table_insert(s_info.instance_cnt_table, strdup(widget_id), cur_instance_cnt);
	}

	if (max_instance_cnt != 0 && max_instance_cnt <= *cur_instance_cnt) {
		set_last_result(WIDGET_ERROR_MAX_EXCEEDED);
		LOGE("already have max count instances");
		return NULL;
	}

	if (!s_info.initialized) {
		set_last_result(WIDGET_ERROR_FAULT);
		LOGE("Widget viewer evas is not initialized");
		return NULL;
	}

	if (!parent) {
		set_last_result(WIDGET_ERROR_INVALID_PARAMETER);
		LOGE("parent(window) object is invalid");
		return NULL;
	}

	if (!widget_id) {
		set_last_result(WIDGET_ERROR_INVALID_PARAMETER);
		LOGE("widget package id is invalid");
		return NULL;
	}

	if (widget_instance_create(widget_id, &instance_id) < 0) {
		set_last_result(WIDGET_ERROR_FAULT);/* LCOV_EXCL_LINE */
		return NULL;
	}

	if (!instance_id) {
		set_last_result(WIDGET_ERROR_FAULT);
		LOGE("Failed to get instance_id: %s", widget_id);
		return NULL;
	}

	widget_instance_info = create_info(parent, widget_id, instance_id, content_info);
	if (!widget_instance_info) {
		set_last_result(WIDGET_ERROR_FAULT);/* LCOV_EXCL_LINE */
		LOGE("Unable to create an information object");/* LCOV_EXCL_LINE */
		widget_instance_destroy(instance_id);/* LCOV_EXCL_LINE */
		return NULL;
	}

	if (content_info)
		widget_instance_info->content_info = strdup(content_info);

	widget_instance_info->pid = 0;
	widget_instance_info->period = period;

	g_hash_table_insert(s_info.widget_table, widget_instance_info->instance_id, widget_instance_info);
	widget_service_set_lifecycle_event_cb(widget_id, __lifecycle_event_cb, NULL);
	(*cur_instance_cnt)++;

	/**
	 * @note
	 * In this case, user(developer) can get a same object using this API.
	 * If he knows the widget_id and instance_id, he can get the object of it.
	 * Same Evas_Object.
	 */
	return widget_instance_info->layout;
}

API int widget_viewer_evas_set_option(widget_evas_conf_e type, int value)
{
	if (!__is_widget_feature_enabled())
		return WIDGET_ERROR_NOT_SUPPORTED;

	if (__check_privilege(WIDGET_PRIVILEGE) < 0)
		return WIDGET_ERROR_PERMISSION_DENIED;

	if (type < WIDGET_VIEWER_EVAS_MANUAL_PAUSE_RESUME || type > WIDGET_VIEWER_EVAS_UNKNOWN) {
		LOGE("type is invalid");
		return WIDGET_ERROR_INVALID_PARAMETER;
	}

	if (type == WIDGET_VIEWER_EVAS_MANUAL_PAUSE_RESUME) {
		if (value)
			screen_connector_toolkit_evas_stop_visibility_notify();
		else
			screen_connector_toolkit_evas_start_visibility_notify();
	} else if (type == WIDGET_VIEWER_EVAS_DELAYED_RESUME) {
		if (value) {
			__is_delayed_resuming_mode = true;
			screen_connector_toolkit_evas_set_delayed_resuming_time(__default_delayed_resuming_time);
		} else {
			__is_delayed_resuming_mode = false;
			screen_connector_toolkit_evas_set_delayed_resuming_time(0);
		}
	}

	return WIDGET_ERROR_NONE;
}

API int widget_viewer_evas_pause_widget(Evas_Object *widget)
{
	struct widget_info *info;
	struct pending_item *item;
	int ret = WIDGET_ERROR_NONE;

	if (!__is_widget_feature_enabled())
		return WIDGET_ERROR_NOT_SUPPORTED;

	if (!s_info.initialized) {
		LOGE("widget viewer evas is not initialized");
		return WIDGET_ERROR_FAULT;
	}

	if (__check_privilege(WIDGET_PRIVILEGE) < 0)
		return WIDGET_ERROR_PERMISSION_DENIED;

	if (!widget) {
		LOGE("widget object is invalid");
		return WIDGET_ERROR_INVALID_PARAMETER;
	}

	info = evas_object_data_get(widget, WIDGET_INFO_TAG);
	if (!info) {
		LOGE("widget(%p) don't have the info", widget);/* LCOV_EXCL_LINE */
		return WIDGET_ERROR_INVALID_PARAMETER;/* LCOV_EXCL_LINE */
	}

	if (info->visibility_freeze) {
		LOGE("widget(%p) is freezing visibility(%d)", widget, info->visibility_freeze);
		return WIDGET_ERROR_DISABLED;
	}

	if (info->is_created) {
		ret = __instance_pause(info);
		if (ret < 0) {
			LOGE("Fail to pause the widget(%p):(%d)", widget, ret);
			return ret;
		}
	} else {
		item = __create_pending_item("pause", 0, 0);
		__push_pending_queue(info, item);
	}

	return WIDGET_ERROR_NONE;
}

API int widget_viewer_evas_resume_widget(Evas_Object *widget)
{
	struct widget_info *info;
	struct pending_item *item;
	int ret = WIDGET_ERROR_NONE;

	if (!__is_widget_feature_enabled())
		return WIDGET_ERROR_NOT_SUPPORTED;

	if (!s_info.initialized) {
		LOGE("widget viewer evas is not initialized");
		return WIDGET_ERROR_FAULT;
	}

	if (__check_privilege(WIDGET_PRIVILEGE) < 0)
		return WIDGET_ERROR_PERMISSION_DENIED;

	if (!widget) {
		LOGE("widget object is invalid");
		return WIDGET_ERROR_INVALID_PARAMETER;
	}

	info = evas_object_data_get(widget, WIDGET_INFO_TAG);
	if (!info) {
		LOGE("widget(%p) don't have the info", widget);/* LCOV_EXCL_LINE */
		return WIDGET_ERROR_INVALID_PARAMETER;/* LCOV_EXCL_LINE */
	}

	if (info->visibility_freeze) {
		LOGE("widget(%p) is freezing visibility(%d)", widget, info->visibility_freeze);
		return WIDGET_ERROR_DISABLED;
	}

	if (info->is_created) {
		ret = __instance_resume(info);
		if (ret < 0) {
			LOGE("Fail to resume the widget(%p):(%d)", widget, ret);
			return ret;
		}
	} else {
		item = __create_pending_item("resume", 0, 0);
		__push_pending_queue(info, item);
	}

	return WIDGET_ERROR_NONE;
}

API const char *widget_viewer_evas_get_content_info(Evas_Object *widget)
{
	struct widget_info *info;

	if (!__is_widget_feature_enabled()) {
		set_last_result(WIDGET_ERROR_NOT_SUPPORTED);
		return NULL;
	}

	if (!s_info.initialized) {
		set_last_result(WIDGET_ERROR_FAULT);
		LOGE("widget viewer evas is not initialized");
		return NULL;
	}

	if (__check_privilege(WIDGET_PRIVILEGE) < 0) {
		set_last_result(WIDGET_ERROR_PERMISSION_DENIED);/* LCOV_EXCL_LINE */
		LOGE("Permission deny");/* LCOV_EXCL_LINE */
		return NULL;/* LCOV_EXCL_LINE */
	}

	if (!widget) {
		LOGE("widget object is invalid");
		set_last_result(WIDGET_ERROR_INVALID_PARAMETER);
		return NULL;
	}

	info = evas_object_data_get(widget, WIDGET_INFO_TAG);
	if (!info) {
		set_last_result(WIDGET_ERROR_INVALID_PARAMETER);/* LCOV_EXCL_LINE */
		LOGE("widget(%p) don't have the info", widget);/* LCOV_EXCL_LINE */
		return NULL;
	}

	return (const char*)info->content_info;
}

API const char *widget_viewer_evas_get_title_string(Evas_Object *widget)
{
	const char *title = NULL;
	struct widget_info *info;

	if (!__is_widget_feature_enabled()) {
		set_last_result(WIDGET_ERROR_NOT_SUPPORTED);
		return NULL;
	}

	if (!s_info.initialized) {
		set_last_result(WIDGET_ERROR_FAULT);
		LOGE("widget viewer evas is not initialized");
		return NULL;
	}

	if (__check_privilege(WIDGET_PRIVILEGE) < 0) {
		set_last_result(WIDGET_ERROR_PERMISSION_DENIED);/* LCOV_EXCL_LINE */
		LOGE("Permission deny");/* LCOV_EXCL_LINE */
		return NULL;/* LCOV_EXCL_LINE */
	}

	if (!widget) {
		set_last_result(WIDGET_ERROR_INVALID_PARAMETER);
		LOGE("widget object is invalid");
		return NULL;
	}

	info = evas_object_data_get(widget, WIDGET_INFO_TAG);
	if (!info) {
		set_last_result(WIDGET_ERROR_INVALID_PARAMETER);/* LCOV_EXCL_LINE */
		LOGE("widget(%p) don't have the info", widget);/* LCOV_EXCL_LINE */
		return NULL;
	}

	/* TODO */
	title = widget_service_get_name(info->widget_id, NULL);

	info->title = title;

	return title;
}

API const char *widget_viewer_evas_get_widget_id(Evas_Object *widget)
{
	struct widget_info *info;

	if (!__is_widget_feature_enabled())
		return NULL;

	if (!s_info.initialized) {
		set_last_result(WIDGET_ERROR_FAULT);
		LOGE("widget viewer evas is not initialized");
		return NULL;
	}

	if (__check_privilege(WIDGET_PRIVILEGE) < 0) {
		set_last_result(WIDGET_ERROR_PERMISSION_DENIED);/* LCOV_EXCL_LINE */
		LOGE("Permission deny");/* LCOV_EXCL_LINE */
		return NULL;/* LCOV_EXCL_LINE */
	}

	if (!widget) {
		set_last_result(WIDGET_ERROR_INVALID_PARAMETER);
		LOGE("widget object is invalid");
		return NULL;
	}

	info = evas_object_data_get(widget, WIDGET_INFO_TAG);
	if (!info) {
		set_last_result(WIDGET_ERROR_INVALID_PARAMETER);/* LCOV_EXCL_LINE */
		LOGE("widget(%p) don't have the info", widget);/* LCOV_EXCL_LINE */
		return NULL;
	}

	return info->widget_id;
}

API double widget_viewer_evas_get_period(Evas_Object *widget)
{
	struct widget_info *info;

	if (!__is_widget_feature_enabled())
		return -1.0f;

	if (!s_info.initialized) {
		set_last_result(WIDGET_ERROR_FAULT);
		LOGE("widget viewer evas is not initialized");
		return -1.0f;
	}

	if (__check_privilege(WIDGET_PRIVILEGE) < 0) {
		set_last_result(WIDGET_ERROR_PERMISSION_DENIED);/* LCOV_EXCL_LINE */
		LOGE("Permission deny");/* LCOV_EXCL_LINE */
		return -1.0f;/* LCOV_EXCL_LINE */
	}

	if (!widget) {
		set_last_result(WIDGET_ERROR_INVALID_PARAMETER);
		LOGE("widget object is invalid");
		return -1.0f;
	}

	info = evas_object_data_get(widget, WIDGET_INFO_TAG);
	if (!info) {
		set_last_result(WIDGET_ERROR_INVALID_PARAMETER);/* LCOV_EXCL_LINE */
		LOGE("widget(%p) don't have the info", widget);/* LCOV_EXCL_LINE */
		return -1.0f;
	}

	return info->period;
}

API void widget_viewer_evas_cancel_click_event(Evas_Object *widget)
{
	Evas_Object *remote_obj;
	struct widget_info *info;

	if (!__is_widget_feature_enabled()) {
		set_last_result(WIDGET_ERROR_NOT_SUPPORTED);/* LCOV_EXCL_LINE */
		return;/* LCOV_EXCL_LINE */
	}

	if (!s_info.initialized) {
		set_last_result(WIDGET_ERROR_FAULT);
		LOGE("widget viewer evas is not initialized");
		return;
	}

	if (__check_privilege(WIDGET_PRIVILEGE) < 0) {
		set_last_result(WIDGET_ERROR_PERMISSION_DENIED);/* LCOV_EXCL_LINE */
		LOGE("Permission deny");/* LCOV_EXCL_LINE */
		return;/* LCOV_EXCL_LINE */
	}

	if (!widget) {
		set_last_result(WIDGET_ERROR_INVALID_PARAMETER);
		LOGE("widget object is invalid");
		return;
	}

	info = evas_object_data_get(widget, WIDGET_INFO_TAG);
	if (!info) {
		set_last_result(WIDGET_ERROR_INVALID_PARAMETER);/* LCOV_EXCL_LINE */
		LOGE("widget(%p) don't have the info", widget);/* LCOV_EXCL_LINE */
		return;
	}

	info->cancel_click = 1;

	remote_obj = elm_object_part_content_get(widget, "tbm,widget");
	if (!remote_obj) {
		set_last_result(WIDGET_ERROR_INVALID_PARAMETER);
		LOGE("widget object is invalid");
		return;
	}

	screen_connector_toolkit_evas_send_touch_cancel(remote_obj);

	return;
}

API int widget_viewer_evas_feed_mouse_up_event(Evas_Object *widget)
{
	Evas_Object *remote_obj;
	struct widget_info *info;

	if (!__is_widget_feature_enabled())
		return WIDGET_ERROR_NOT_SUPPORTED;

	if (!s_info.initialized) {
		LOGE("widget viewer evas is not initialized");
		return WIDGET_ERROR_FAULT;
	}

	if (__check_privilege(WIDGET_PRIVILEGE) < 0) {
		LOGE("Permission deny");/* LCOV_EXCL_LINE */
		return WIDGET_ERROR_PERMISSION_DENIED;/* LCOV_EXCL_LINE */
	}

	if (!widget) {
		LOGE("widget object is invalid");
		return WIDGET_ERROR_INVALID_PARAMETER;
	}

	info = evas_object_data_get(widget, WIDGET_INFO_TAG);
	if (!info) {
		LOGE("widget(%p) don't have the info", widget);
		return WIDGET_ERROR_INVALID_PARAMETER;
	}

	remote_obj = elm_object_part_content_get(widget, "tbm,widget");
	if (!remote_obj) {
		LOGE("widget object is invalid");/* LCOV_EXCL_LINE */
		return WIDGET_ERROR_INVALID_PARAMETER;/* LCOV_EXCL_LINE */
	}

	screen_connector_toolkit_evas_send_mouse_up(remote_obj);

	return WIDGET_ERROR_NONE;
}

/**
 * @note
 * While displaying "loading" overlay, there is a preview image.
 * But if this function is called, the preview image will not be displayed.
 * Only the help text will be there. if it is not disabled of course.
 */
API void widget_viewer_evas_disable_preview(Evas_Object *widget)
{
	struct widget_info *info;

	if (!__is_widget_feature_enabled()) {
		set_last_result(WIDGET_ERROR_NOT_SUPPORTED);/* LCOV_EXCL_LINE */
		return;/* LCOV_EXCL_LINE */
	}

	if (!s_info.initialized) {
		set_last_result(WIDGET_ERROR_FAULT);
		LOGE("widget viewer evas is not initialized");
		return;
	}

	if (__check_privilege(WIDGET_PRIVILEGE) < 0) {
		LOGE("Permission deny");/* LCOV_EXCL_LINE */
		return;/* LCOV_EXCL_LINE */
	}

	if (!widget) {
		set_last_result(WIDGET_ERROR_INVALID_PARAMETER);
		LOGE("widget object is invalid");
		return;
	}

	info = evas_object_data_get(widget, WIDGET_INFO_TAG);
	if (!info) {
		set_last_result(WIDGET_ERROR_INVALID_PARAMETER);/* LCOV_EXCL_LINE */
		LOGE("widget(%p) don't have the info", widget);/* LCOV_EXCL_LINE */
		return;/* LCOV_EXCL_LINE */
	}

	info->disable_preview = 1;
	elm_object_signal_emit(info->layout, "disable", "preview");

	return;
}

/**
 * @note
 * While displaying "loading" overlay, there is a help text.
 * But if this function is called, the text help message will not be displayed.
 * Only the preview image will be there. if it is not disabled of course.
 */
API void widget_viewer_evas_disable_overlay_text(Evas_Object *widget)
{
	struct widget_info *info;

	if (!__is_widget_feature_enabled()) {
		set_last_result(WIDGET_ERROR_NOT_SUPPORTED);/* LCOV_EXCL_LINE */
		return;/* LCOV_EXCL_LINE */
	}

	if (!s_info.initialized) {
		set_last_result(WIDGET_ERROR_FAULT);
		LOGE("widget viewer evas is not initialized");
		return;
	}

	if (__check_privilege(WIDGET_PRIVILEGE) < 0) {
		LOGE("Permission deny");/* LCOV_EXCL_LINE */
		return;/* LCOV_EXCL_LINE */
	}

	if (!widget) {
		set_last_result(WIDGET_ERROR_INVALID_PARAMETER);
		LOGE("widget object is invalid");
		return;
	}

	info = evas_object_data_get(widget, WIDGET_INFO_TAG);
	if (!info) {
		set_last_result(WIDGET_ERROR_INVALID_PARAMETER);/* LCOV_EXCL_LINE */
		LOGE("widget(%p) don't have the info", widget);/* LCOV_EXCL_LINE */
		return;
	}

	info->disable_overlay = 1;
	elm_object_signal_emit(info->layout, "disable", "overlay,text");
	return;
}

/**
 * @note
 * This function should be called before resize the widget object after create it.
 * If this function is called, the widget object will not display "preview" image.
 * Instead of it, the widget content object will be displayed immediately.
 */
API void widget_viewer_evas_disable_loading(Evas_Object *widget)
{
	struct widget_info *info;

	if (!__is_widget_feature_enabled()) {
		set_last_result(WIDGET_ERROR_NOT_SUPPORTED);/* LCOV_EXCL_LINE */
		return;/* LCOV_EXCL_LINE */
	}

	if (!s_info.initialized) {
		set_last_result(WIDGET_ERROR_FAULT);
		LOGE("widget viewer evas is not initialized");
		return;
	}

	if (__check_privilege(WIDGET_PRIVILEGE) < 0) {
		LOGE("Permission deny");/* LCOV_EXCL_LINE */
		return;/* LCOV_EXCL_LINE */
	}

	if (!widget) {
		set_last_result(WIDGET_ERROR_INVALID_PARAMETER);
		LOGE("widget object is invalid");
		return;
	}

	info = evas_object_data_get(widget, WIDGET_INFO_TAG);
	if (!info) {
		set_last_result(WIDGET_ERROR_INVALID_PARAMETER);/* LCOV_EXCL_LINE */
		LOGE("widget(%p) don't have the info", widget);/* LCOV_EXCL_LINE */
		return;
	}

	info->disable_loading = 1;
	elm_object_signal_emit(info->layout, "disable", "loading,text");
	return;
}

API void widget_viewer_evas_activate_faulted_widget(Evas_Object *widget)
{
	struct widget_info *info;
	int w;
	int h;

	if (!__is_widget_feature_enabled()) {
		set_last_result(WIDGET_ERROR_NOT_SUPPORTED);/* LCOV_EXCL_LINE */
		return;/* LCOV_EXCL_LINE */
	}

	if (!s_info.initialized) {
		set_last_result(WIDGET_ERROR_FAULT);
		LOGE("widget viewer evas is not initialized");
		return;
	}

	if (__check_privilege(WIDGET_PRIVILEGE) < 0) {
		set_last_result(WIDGET_ERROR_PERMISSION_DENIED);/* LCOV_EXCL_LINE */
		LOGE("Permission deny");/* LCOV_EXCL_LINE */
		return;/* LCOV_EXCL_LINE */
	}

	if (!widget) {
		set_last_result(WIDGET_ERROR_INVALID_PARAMETER);
		LOGE("widget object is invalid");
		return;
	}

	info = evas_object_data_get(widget, WIDGET_INFO_TAG);
	if (!info) {
		set_last_result(WIDGET_ERROR_INVALID_PARAMETER);/* LCOV_EXCL_LINE */
		LOGE("widget(%p) don't have the info", widget);/* LCOV_EXCL_LINE */
		return;
	}

	aul_widget_write_log(LOG_TAG, "[%s:%d] %s pid : %d",
			__FUNCTION__, __LINE__, info->instance_id, info->pid);
	if (info->pid < 0) {
		evas_object_geometry_get(info->layout, NULL, NULL, &w, &h);
		__display_loading_text(info);
		__add_launch_list(info->instance_id);
		__launch_instance();
	} else {
		/**
		 * @note
		 * Widget process is running well...
		 */
	}
	return;
}

API bool widget_viewer_evas_is_faulted(Evas_Object *widget)
{
	struct widget_info *info;

	if (!__is_widget_feature_enabled()) {
		set_last_result(WIDGET_ERROR_NOT_SUPPORTED);/* LCOV_EXCL_LINE */
		return false;/* LCOV_EXCL_LINE */
	}

	if (!s_info.initialized) {
		set_last_result(WIDGET_ERROR_FAULT);
		LOGE("widget viewer evas is not initialized");
		return false;
	}

	if (__check_privilege(WIDGET_PRIVILEGE) < 0) {
		set_last_result(WIDGET_ERROR_PERMISSION_DENIED);/* LCOV_EXCL_LINE */
		LOGE("Permission deny");/* LCOV_EXCL_LINE */
		return false;/* LCOV_EXCL_LINE */
	}

	if (!widget) {
		set_last_result(WIDGET_ERROR_INVALID_PARAMETER);
		LOGE("widget object is invalid");
		return false;
	}

	info = evas_object_data_get(widget, WIDGET_INFO_TAG);
	if (!info) {
		set_last_result(WIDGET_ERROR_INVALID_PARAMETER);/* LCOV_EXCL_LINE */
		LOGE("widget(%p) don't have the info", widget);/* LCOV_EXCL_LINE */
		return false;/* LCOV_EXCL_LINE */
	}

	return info->is_faulted;
}

API int widget_viewer_evas_freeze_visibility(Evas_Object *widget, widget_visibility_status_e status)
{
	struct widget_info *info;
	int ret = 0;
	Evas_Object *tbm_obj;

	if (!__is_widget_feature_enabled())
		return WIDGET_ERROR_NOT_SUPPORTED;

	if (!s_info.initialized) {
		LOGE("widget viewer evas is not initialized");/* LCOV_EXCL_LINE */
		return WIDGET_ERROR_FAULT;/* LCOV_EXCL_LINE */
	}

	if (__check_privilege(WIDGET_PRIVILEGE) < 0) {
		LOGE("Permission deny");/* LCOV_EXCL_LINE */
		return WIDGET_ERROR_PERMISSION_DENIED;/* LCOV_EXCL_LINE */
	}

	if (!widget) {
		LOGE("widget object is invalid");
		return WIDGET_ERROR_INVALID_PARAMETER;
	}

	info = evas_object_data_get(widget, WIDGET_INFO_TAG);
	if (!info) {
		LOGE("widget(%p) don't have the info", widget);
		return WIDGET_ERROR_INVALID_PARAMETER;
	}

	tbm_obj = elm_object_part_content_get(widget, "tbm,widget");
	if (!tbm_obj) {
		LOGE("widget object is invalid\n");/* LCOV_EXCL_LINE */
		return WIDGET_ERROR_INVALID_PARAMETER;/* LCOV_EXCL_LINE */
	}

	if (status == WIDGET_VISIBILITY_STATUS_SHOW_FIXED) {
		ret = screen_connector_toolkit_evas_freeze_visibility(tbm_obj, VISIBILITY_TYPE_UNOBSCURED);
		if (ret < 0) {
			LOGE("Fail to resume the widget(%p):(%d)", widget, ret);
			return ret;
		}
		info->visibility_freeze = status;
	} else if (status == WIDGET_VISIBILITY_STATUS_HIDE_FIXED) {
		ret = screen_connector_toolkit_evas_freeze_visibility(tbm_obj, VISIBILITY_TYPE_FULLY_OBSCURED);
		if (ret < 0) {
			LOGE("Fail to pause the widget(%p):(%d)", widget, ret);
			return ret;
		}
		info->visibility_freeze = status;
	} else {

		LOGE("status value is invalid");
		return WIDGET_ERROR_INVALID_PARAMETER;
	}

	return WIDGET_ERROR_NONE;
}

API int widget_viewer_evas_thaw_visibility(Evas_Object *widget)
{
	struct widget_info *info;
	Evas_Object *tbm_obj;

	if (!__is_widget_feature_enabled())
		return WIDGET_ERROR_NOT_SUPPORTED;

	if (!s_info.initialized) {
		LOGE("widget viewer evas is not initialized");
		return WIDGET_ERROR_FAULT;
	}

	if (__check_privilege(WIDGET_PRIVILEGE) < 0) {
		LOGE("Permission deny");/* LCOV_EXCL_LINE */
		return WIDGET_ERROR_PERMISSION_DENIED;/* LCOV_EXCL_LINE */
	}

	if (!widget) {
		LOGE("widget object is invalid");
		return WIDGET_ERROR_INVALID_PARAMETER;
	}

	info = evas_object_data_get(widget, WIDGET_INFO_TAG);
	if (!info) {
		LOGE("widget(%p) don't have the info", widget);/* LCOV_EXCL_LINE */
		return WIDGET_ERROR_INVALID_PARAMETER;/* LCOV_EXCL_LINE */
	}

	tbm_obj = elm_object_part_content_get(widget, "tbm,widget");
	if (!tbm_obj) {
		LOGE("widget object is invalid\n");/* LCOV_EXCL_LINE */
		return WIDGET_ERROR_INVALID_PARAMETER;/* LCOV_EXCL_LINE */
	}

	screen_connector_toolkit_evas_thaw_visibility(tbm_obj);
	info->visibility_freeze = 0;

	return WIDGET_ERROR_NONE;
}

API bool widget_viewer_evas_is_visibility_frozen(Evas_Object *widget)
{
	struct widget_info *info;

	if (!__is_widget_feature_enabled()) {
		set_last_result(WIDGET_ERROR_NOT_SUPPORTED);/* LCOV_EXCL_LINE */
		return false;/* LCOV_EXCL_LINE */
	}

	if (!s_info.initialized) {
		set_last_result(WIDGET_ERROR_FAULT);
		LOGE("widget viewer evas is not initialized");
		return false;
	}

	if (__check_privilege(WIDGET_PRIVILEGE) < 0) {
		set_last_result(WIDGET_ERROR_PERMISSION_DENIED);/* LCOV_EXCL_LINE */
		LOGE("Permission deny");/* LCOV_EXCL_LINE */
		return false;/* LCOV_EXCL_LINE */
	}

	if (!widget) {
		set_last_result(WIDGET_ERROR_INVALID_PARAMETER);
		LOGE("widget object is invalid");
		return false;
	}

	info = evas_object_data_get(widget, WIDGET_INFO_TAG);
	if (!info) {
		set_last_result(WIDGET_ERROR_INVALID_PARAMETER);/* LCOV_EXCL_LINE */
		LOGE("widget(%p) don't have the info", widget);/* LCOV_EXCL_LINE */
		return false;/* LCOV_EXCL_LINE */
	}

	if (info->visibility_freeze)
		return true;

	return false;
}

API bool widget_viewer_evas_is_widget(Evas_Object *widget)
{
	struct widget_info *info;

	if (!__is_widget_feature_enabled()) {
		set_last_result(WIDGET_ERROR_NOT_SUPPORTED);/* LCOV_EXCL_LINE */
		return false;/* LCOV_EXCL_LINE */
	}

	if (!s_info.initialized) {
		set_last_result(WIDGET_ERROR_FAULT);
		LOGE("widget viewer evas is not initialized");
		return false;
	}

	if (__check_privilege(WIDGET_PRIVILEGE) < 0) {
		set_last_result(WIDGET_ERROR_PERMISSION_DENIED);/* LCOV_EXCL_LINE */
		LOGE("Permission deny");/* LCOV_EXCL_LINE */
		return false;/* LCOV_EXCL_LINE */
	}

	if (!widget) {
		set_last_result(WIDGET_ERROR_INVALID_PARAMETER);
		LOGE("widget object is invalid");
		return false;
	}

	info = evas_object_data_get(widget, WIDGET_INFO_TAG);
	if (!info) {
		set_last_result(WIDGET_ERROR_INVALID_PARAMETER);/* LCOV_EXCL_LINE */
		LOGE("widget(%p) don't have the info", widget);/* LCOV_EXCL_LINE */
		return false;/* LCOV_EXCL_LINE */
	}

	return true;
}

API void widget_viewer_evas_set_permanent_delete(Evas_Object *widget, int flag)
{
	struct widget_info *info;

	if (!__is_widget_feature_enabled()) {
		set_last_result(WIDGET_ERROR_NOT_SUPPORTED);/* LCOV_EXCL_LINE */
		return;/* LCOV_EXCL_LINE */
	}

	if (!s_info.initialized) {
		set_last_result(WIDGET_ERROR_FAULT);
		LOGE("widget viewer evas is not initialized");
		return;
	}

	if (__check_privilege(WIDGET_PRIVILEGE) < 0) {
		set_last_result(WIDGET_ERROR_PERMISSION_DENIED);/* LCOV_EXCL_LINE */
		LOGE("Permission deny");/* LCOV_EXCL_LINE */
		return;/* LCOV_EXCL_LINE */
	}

	if (!widget) {
		set_last_result(WIDGET_ERROR_INVALID_PARAMETER);
		LOGE("widget object is invalid");
		return;
	}

	info = evas_object_data_get(widget, WIDGET_INFO_TAG);
	if (!info) {
		set_last_result(WIDGET_ERROR_INVALID_PARAMETER);/* LCOV_EXCL_LINE */
		LOGE("widget(%p) don't have the info", widget);/* LCOV_EXCL_LINE */
		return;/* LCOV_EXCL_LINE */
	}

	info->permanent_delete = flag;
	return;
}

API const char *widget_viewer_evas_get_widget_instance_id(Evas_Object *widget)
{
	struct widget_info *info;

	if (!__is_widget_feature_enabled()) {
		set_last_result(WIDGET_ERROR_NOT_SUPPORTED);/* LCOV_EXCL_LINE */
		return NULL;/* LCOV_EXCL_LINE */
	}

	if (__check_privilege("http://tizen.org/privilege/widget.viewer") < 0) {
		set_last_result(WIDGET_ERROR_PERMISSION_DENIED);/* LCOV_EXCL_LINE */
		LOGE("Permission deny");/* LCOV_EXCL_LINE */
		return NULL;/* LCOV_EXCL_LINE */
	}

	if (!s_info.initialized) {
		set_last_result(WIDGET_ERROR_FAULT);
		LOGE("widget viewer evas is not initialized");
		return NULL;
	}

	if (!widget) {
		set_last_result(WIDGET_ERROR_INVALID_PARAMETER);
		LOGE("widget object is invalid");
		return NULL;
	}

	info = evas_object_data_get(widget, WIDGET_INFO_TAG);
	if (!info) {
		set_last_result(WIDGET_ERROR_INVALID_PARAMETER);/* LCOV_EXCL_LINE */
		LOGE("widget(%p) don't have the info", widget);/* LCOV_EXCL_LINE */
		return NULL;/* LCOV_EXCL_LINE */
	}

	return (const char *)info->instance_id;
}

API int widget_viewer_evas_set_delayed_resuming_time(int time)
{
	if (time < 0 || time > 10000)
		return WIDGET_ERROR_INVALID_PARAMETER;

	__default_delayed_resuming_time = time;
	if (__is_delayed_resuming_mode)
		screen_connector_toolkit_evas_set_delayed_resuming_time(time);

	return WIDGET_ERROR_NONE;
}

API int widget_viewer_evas_launch_setup_app(Evas_Object *widget)
{
	char *setup_appid;
	const char *widget_id;
	const char *instance_id;
	const char *content_info;
	app_control_h service;

	widget_id = widget_viewer_evas_get_widget_id(widget);
	if (!widget_id) {
		LOGE("cannot get widget id from object");
		return WIDGET_ERROR_INVALID_PARAMETER;
	}

	setup_appid = widget_service_get_app_id_of_setup_app(widget_id);
	if (!setup_appid) {
		LOGE("widget[%s] does not have setup appid", widget_id);
		return WIDGET_ERROR_NOT_EXIST;
	}

	content_info = widget_viewer_evas_get_content_info(widget);
	instance_id = widget_viewer_evas_get_widget_instance_id(widget);
	app_control_create(&service);
	app_control_set_operation(service, APP_CONTROL_OPERATION_EDIT);
	app_control_set_app_id(service, setup_appid);
	app_control_add_extra_data(service, APP_CONTROL_DATA_WIDGET_APP_ID, widget_id);
	app_control_add_extra_data(service, APP_CONTROL_DATA_WIDGET_INSTANCE_ID, instance_id);
	app_control_add_extra_data(service, APP_CONTROL_DATA_WIDGET_CONTENT, content_info);

	if (app_control_send_launch_request(service, NULL, NULL) != APP_CONTROL_ERROR_NONE)
		LOGE("Failed to launch setup app[%s]", setup_appid);

	app_control_destroy(service);
	free(setup_appid);

	return WIDGET_ERROR_NONE;
}

/* End of a file */
