/*
 * Copyright 2014  Samsung Electronics Co., Ltd
 *
 * Licensed under the Flora License, Version 1.1 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://floralicense.org/license/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define _GNU_SOURCE
#include <glib.h>
#include <Elementary.h>
#include <app.h>
#include <unistd.h>
#include <errno.h>
#include <widget_errno.h>
#include <widget_service.h>
#include <widget_service_internal.h>
#include <widget_viewer_evas.h>
#include <vconf.h>
#include <bundle.h>
#include <dlog.h>
#include <string.h>
#include <app_control.h>
#include <app_control_internal.h>
#include <efl_extension.h>
#include <aul.h>
#include <aul_debug_info.h>
#include <appsvc.h>
#include <watch_control.h>
#include <pkgmgr-info.h>


int errno;

#include "main.h"
#include "debug.h"

#define ERR_BUFFER_SIZE		1024
#define WATCH_CATEGORY "http://tizen.org/category/wearable_clock"
#define LAZY_WATCH_TIME 3.0f
#define CR 13
#define LF 10
#if !defined(VCONFKEY_WMS_CLOCKS_SET_IDLE)
#define VCONFKEY_WMS_CLOCKS_SET_IDLE "db/wms/clocks_set_idle"
#endif

#define LAZY_LOADER "__LAZY_LOADER__"
#define DEFAULT_WATCH "org.tizen.idle-clock-digital"

typedef struct _widget_info {
	char *widget_id;
	char *content_info;
	char *title;
	double period;
	Evas_Object *widget;
	Evas_Object *widget_layout;
	int *size_types;
	int count_of_size_type;
} widget_info;

typedef struct _widget_size_list_item {
	widget_info *info;
	int idx;
}widget_size_list_item;

static struct info {
	int w;
	int h;
	Evas_Object *win;
	Evas_Object *bg;
	Evas_Object *box;
	Evas_Object *layout;
	Ecore_Timer *long_press;
	GList *widget_info_list;
	Ecore_Timer *lazy_widget_loader;
	double lazy_watch_time;
	char *default_watch;
	char *watch_category;
} s_info = {
	.w = 0,
	.h = 0,
	.win = NULL,
	.layout = NULL,
	.long_press = NULL,
	.widget_info_list = NULL,
	.lazy_widget_loader = NULL,
	.lazy_watch_time = LAZY_WATCH_TIME,
	.default_watch = DEFAULT_WATCH,
	.watch_category = WATCH_CATEGORY,
};

#define LONG_PRESS 1.0f
#define CONF_FNAME "/etc/widget_viewer_sdk.conf"

static char *__watch_pkgid;
static Evas_Object *__clock;
static bool __debug_info_initialized;

static void back_key_cb(void *data, Evas_Object *obj, void *event_info)
{
	ui_app_exit();
}

static void hide_widget_info_cb(void *data, Evas_Object *obj, const char *emission, const char *source)
{
	DbgPrint("Hide info panel\n");
	elm_object_signal_emit(s_info.layout, "hide", "info");
}

static char *list_item_text_get_cb(void *data, Evas_Object *obj, const char *part)
{
	char buffer[256];
	int ret;
	int w;
	int h;
	widget_size_list_item *item = (widget_size_list_item *)data;

	ret = widget_service_get_size(item->info->size_types[item->idx], &w, &h);
	if (ret != WIDGET_ERROR_NONE)
		return strdup("Invalid");

	snprintf(buffer, sizeof(buffer), "%dx%d", w, h);
	DbgPrint("Size: [%s]\n", buffer);
	return strdup(buffer);
}

static Evas_Object *list_item_content_get_cb(void *data, Evas_Object *obj, const char *part)
{
	Evas_Object *icon = NULL;
	char *icon_filename;
	const char *widget_id;
	widget_size_list_item *item = (widget_size_list_item *)data;

	widget_id = widget_viewer_evas_get_widget_id(item->info->widget);

	icon_filename = widget_service_get_preview_image_path(widget_id, item->info->size_types[item->idx]);
	if (icon_filename) {
		icon = elm_icon_add(s_info.win);
		if (icon) {
			elm_image_file_set(icon, icon_filename, NULL);
			elm_image_resizable_set(icon, EINA_TRUE, EINA_TRUE);
			evas_object_size_hint_max_set(icon, 100, 100);
			DbgPrint("Icon: %s\n", icon_filename);
		}
		free(icon_filename);
	} else {
		icon = NULL;
	}

	return icon;
}

static void list_item_clicked_cb(void *data, Evas_Object *obj, void *event_info)
{
	int w;
	int h;
	widget_size_list_item *item;

	item = (widget_size_list_item *)data;
	if (item == NULL)
		return;

	if (widget_service_get_size(item->info->size_types[item->idx], &w, &h) != WIDGET_ERROR_NONE)
		return;

	evas_object_resize(item->info->widget, w, h);
	evas_object_size_hint_min_set(item->info->widget, w, h);
	evas_object_size_hint_max_set(item->info->widget, w, h);
}

static Eina_Bool show_widget_info_cb(void *data)
{
	DbgPrint("Show info panel\n");

	int i;
	int w;
	int h;
	Evas_Object *size_list;
	widget_size_list_item *item;
	Elm_Object_Item *elm_item;
	widget_info *info = (widget_info *)data;
	static Elm_Genlist_Item_Class class = {
		.item_style = "1text.1icon",
		.func = {
			.text_get = list_item_text_get_cb,
			.content_get = list_item_content_get_cb,
			.state_get = NULL,
			.del = NULL,
		}
	};

	DbgPrint("======== %d", info->count_of_size_type);

	if (info->count_of_size_type <= 1) {
		elm_object_signal_emit(s_info.layout, "hide", "size,list");
		return ECORE_CALLBACK_CANCEL;
	} else {
		elm_object_signal_emit(s_info.layout, "show", "size,list");
	}

	size_list = elm_object_part_content_get(s_info.layout, "widget,size,list");
	if (size_list) {
		elm_genlist_clear(size_list);
	} else {
		size_list = elm_genlist_add(s_info.win);
		if (!size_list) {
			ErrPrint("Failed to create a genlist\n");
			return ECORE_CALLBACK_RENEW;
		}

		elm_object_part_content_set(s_info.layout, "widget,size,list", size_list);
	}

	DbgPrint("========\nDump supported size types \n");
	for (i = 0; i < info->count_of_size_type; i++) {
		if (widget_service_get_size(info->size_types[i], &w, &h) != WIDGET_ERROR_NONE || w == 0 || h == 0) {
			DbgPrint("[%X] is not supported (%dx%d)\n", info->size_types[i], w, h);
			continue;
		}
		DbgPrint("Size[%X]\n", info->size_types[i]);

		item = (widget_size_list_item *)calloc(1, sizeof(widget_size_list_item));
		if (item == NULL) {
			ErrPrint("Fail to calloc");
			continue;
		}

		item->info = info;
		item->idx = i;
		elm_item = elm_genlist_item_append(size_list, &class, (void *)item, NULL, ELM_GENLIST_ITEM_NONE, list_item_clicked_cb, (void *)item);
		if (!elm_item)
			ErrPrint("Failed to add a new size elm_item\n");
	}
	DbgPrint("========\nEnd of a dump\n");
	elm_object_signal_emit(s_info.layout, "show", "info");

	s_info.long_press = NULL;
	return ECORE_CALLBACK_CANCEL;
}

static void layout_down_cb(void *data, Evas *e, Evas_Object *obj, void *event_info)
{
	if (s_info.long_press) {
		ecore_timer_del(s_info.long_press);
		s_info.long_press = NULL;
	}

	s_info.long_press = ecore_timer_add(LONG_PRESS, show_widget_info_cb, data);
	if (!s_info.long_press)
		ErrPrint("Failed to add a timer\n");
}

static void layout_up_cb(void *data, Evas *e, Evas_Object *obj, void *event_info)
{
	if (s_info.long_press) {
		ecore_timer_del(s_info.long_press);
		s_info.long_press = NULL;
	}
}

static void __watch_added(void *data, Evas_Object *obj, void *event_info)
{
	LOGD("watch added");
	__clock = (Evas_Object *)event_info;

	evas_object_resize(s_info.win, s_info.w, s_info.h);
	evas_object_resize(__clock, s_info.w, s_info.h);

	evas_object_show(__clock);
	evas_object_show(s_info.win);
}

static void __watch_removed(void *data, Evas_Object *obj, void *event_info)
{
	LOGD("watch removed");
}

void  __clock_vconf_cb(keynode_t *node, void *data)
{
	char *cur_watch = vconf_get_str(VCONFKEY_WMS_CLOCKS_SET_IDLE);

	LOGD("watch sdk [%s], home [%s]",
			__watch_pkgid, cur_watch);
	if (cur_watch != NULL && __watch_pkgid != NULL
			&& strcmp(cur_watch, __watch_pkgid) == 0) {
		ui_app_exit();
		LOGW("same watch displayed on the other viewer EXIT");
	}
	free(cur_watch);
}

static bool app_create(void *data)
{

	Evas_Object *scroller;
	s_info.win = elm_win_add(NULL, "Widget Viewer (SDK)", ELM_WIN_BASIC);
	if (!s_info.win) {
		LOGD("Failed to create a window\n");
		return false;
	}

	elm_app_base_scale_set(2.6);
	elm_win_screen_size_get(s_info.win, NULL, NULL, &s_info.w, &s_info.h);
	LOGD("Window: %dx%d\n", s_info.w, s_info.h);

	elm_win_alpha_set(s_info.win, EINA_FALSE);
	elm_win_indicator_mode_set(s_info.win, ELM_WIN_INDICATOR_SHOW);

	s_info.layout = elm_layout_add(s_info.win);
	if (!s_info.layout) {
		evas_object_del(s_info.box);
		evas_object_del(s_info.bg);
		evas_object_del(s_info.win);
		s_info.box = NULL;
		s_info.win = NULL;
		s_info.bg = NULL;
		return false;
	}

	if (elm_layout_file_set(s_info.layout, LAYOUT_EDJ, "layout") != EINA_TRUE) {
		LOGE("Failed to load an edje\n");
		evas_object_del(s_info.bg);
		evas_object_del(s_info.layout);
		evas_object_del(s_info.win);
		s_info.bg = NULL;
		s_info.layout = NULL;
		s_info.win = NULL;
		return false;
	}
	evas_object_size_hint_weight_set(s_info.layout, EVAS_HINT_EXPAND, EVAS_HINT_EXPAND);
	elm_win_resize_object_add(s_info.win, s_info.layout);
	evas_object_show(s_info.layout);
	elm_object_signal_callback_add(s_info.layout, "mouse,clicked,1", "widget,info,bg", hide_widget_info_cb, NULL);

	/* Initialize scroller */
	scroller = elm_scroller_add(s_info.layout);
	evas_object_size_hint_weight_set(scroller, EVAS_HINT_EXPAND, EVAS_HINT_EXPAND);
	elm_win_resize_object_add(s_info.layout, scroller);
	evas_object_show(scroller);


	/* Initialize box */
	s_info.box = elm_box_add(scroller);
	if (!s_info.box) {
		evas_object_del(s_info.bg);
		evas_object_del(s_info.win);
		s_info.bg = NULL;
		s_info.win = NULL;
		return false;
	}
	evas_object_size_hint_fill_set(s_info.box, EVAS_HINT_FILL, EVAS_HINT_FILL);
	evas_object_size_hint_weight_set(s_info.box, EVAS_HINT_EXPAND, EVAS_HINT_EXPAND);
	elm_box_align_set(s_info.box, 0.0, 0.0);


	elm_object_content_set(scroller, s_info.box);
	elm_scroller_bounce_set(scroller, EINA_TRUE, EINA_FALSE);
	elm_scroller_policy_set(scroller, ELM_SCROLLER_POLICY_OFF, ELM_SCROLLER_POLICY_OFF);
	elm_scroller_propagate_events_set(scroller, EINA_TRUE);
	elm_scroller_page_relative_set(scroller, 0, 1);
	elm_scroller_region_show(scroller, 50, 50, 200, 200);

	elm_object_scroll_lock_x_set(scroller, EINA_FALSE);
	elm_object_part_content_set(s_info.layout, "widget", scroller);


	/* WIDGET init */
	widget_viewer_evas_set_option(WIDGET_VIEWER_EVAS_DIRECT_UPDATE, 1);
	widget_viewer_evas_set_option(WIDGET_VIEWER_EVAS_EVENT_AUTO_FEED, 1);
	widget_viewer_evas_init(s_info.win);

	eext_object_event_callback_add(s_info.win, EEXT_CALLBACK_BACK, back_key_cb, NULL);
	evas_object_show(s_info.win);
	vconf_notify_key_changed(VCONFKEY_WMS_CLOCKS_SET_IDLE, __clock_vconf_cb, NULL);

	return true;
}

static void free_widget_info(gpointer data)
{
	const char *tmp;
	widget_info *info = (widget_info *)data;

	if (info == NULL)
		return;

	tmp = widget_viewer_evas_get_widget_id(info->widget);
	DbgPrint("Unload previous widget: %s\n", tmp);

	elm_box_unpack(s_info.box, info->widget_layout);
	evas_object_del(info->widget);
	evas_object_del(info->widget_layout);
	if (info->title)
		free(info->title);
	if (info->content_info)
		free(info->content_info);
	if (info->size_types)
		free(info->size_types);
	if (info->widget_id)
		free(info->widget_id);

	info->widget = NULL;
	info->widget_layout = NULL;
	info->title = NULL;
	info->content_info = NULL;
	info->size_types = NULL;
	info->period = WIDGET_VIEWER_EVAS_DEFAULT_PERIOD;
	info->count_of_size_type = 20;
	free(info);
}

static int unload_widget(void)
{
	int widget_cnt;

	widget_cnt = g_list_length(s_info.widget_info_list);
	if (widget_cnt == 0)
		return WIDGET_ERROR_NOT_EXIST;

	g_list_free_full(s_info.widget_info_list, free_widget_info);
	s_info.widget_info_list = NULL;

	elm_object_part_text_set(s_info.layout, "widget,id", "");
	elm_object_part_text_set(s_info.layout, "widget,content,info", "");
	elm_object_part_text_set(s_info.layout, "widget,title", "");
	elm_object_part_text_set(s_info.layout, "widget,period", "");

	return WIDGET_ERROR_NONE;
}

static void app_terminate(void *data)
{
	eext_object_event_callback_del(s_info.win, EEXT_CALLBACK_BACK, back_key_cb);

	if (__debug_info_initialized) {
		aul_debug_info_fini();
		__debug_info_initialized = false;
	}
	unload_widget();

	/* WIDGET fini */
	widget_viewer_evas_fini();

	if (s_info.win) {
		evas_object_del(s_info.win);
		s_info.win = NULL;
	}
}

static void app_pause(void *data)
{
	/* WIDGET pause */
	widget_viewer_evas_notify_paused_status_of_viewer();
	watch_manager_notify_paused_status_of_viewer();
}

static void __resume_process()
{
	widget_viewer_evas_notify_resumed_status_of_viewer();
	watch_manager_notify_resumed_status_of_viewer();
}

static void app_resume(void *data)
{
	__resume_process();
}

static void updated_cb(void *data, Evas_Object *obj, void *event_info)
{
	/* Updated */
	DbgPrint("Widget updated\n");
}

static void extra_updated_cb(void *data, Evas_Object *obj, void *event_info)
{
	const char *string;
	char *tmp;
	widget_info *info = (widget_info *)data;

	string = widget_viewer_evas_get_title_string(obj);
	if (string) {
		tmp = strdup(string);
		if (!tmp)
			return;

		free(info->title);
		info->title = tmp;

		elm_object_part_text_set(s_info.layout, "widget,title", info->title);
	}

	string = widget_viewer_evas_get_content_info(obj);
	if (string) {
		tmp = strdup(string);
		if (!tmp)
			return;

		free(info->content_info);
		info->content_info = tmp;

		elm_object_part_text_set(s_info.layout, "widget,content,info", info->content_info);
		DbgPrint("Content updated: [%s]\n", info->content_info);
	}
}

static void period_changed_cb(void *data, Evas_Object *obj, void *event_info)
{
	char buffer[96];
	widget_info *info = (widget_info *)data;

	info->period = widget_viewer_evas_get_period(obj);
	snprintf(buffer, sizeof(buffer), "%lf", info->period);
	elm_object_part_text_set(s_info.layout, "widget,period", buffer);
	DbgPrint("Period updated: %s\n", buffer);
}

static void update_message(const char *fmt, ...)
{
	if (!fmt) {
		elm_object_signal_emit(s_info.layout, "hide", "message");
	} else {
		va_list ap;
		char buffer[1024];

		va_start(ap, fmt);
		vsnprintf(buffer, sizeof(buffer) - 1, fmt, ap);
		va_end(ap);

		elm_object_part_text_set(s_info.layout, "message", buffer);
		elm_object_signal_emit(s_info.layout, "show", "message");
		ErrPrint("%s", buffer);
	}
}

static void widget_create_aborted_cb(void *data, Evas_Object *obj, void *event_info)
{
	update_message("Create aborted");
}

static void widget_created_cb(void *data, Evas_Object *obj, void *event_info)
{
	period_changed_cb(data, obj, NULL);
	extra_updated_cb(data, obj, NULL);
}

static int load_widget(const char *widget_id, widget_info *info)
{
	Evas_Object *bg = NULL;
	int w = 0;
	int h = 0;
	int i;

	for (i = 0; i < info->count_of_size_type; i++) {
		if (widget_service_get_size(info->size_types[i], &w, &h) != WIDGET_ERROR_NONE || w == 0 || h == 0)
			continue;
		break;
	}

	info->widget_layout = elm_layout_add(s_info.win);
	elm_layout_file_set(info->widget_layout, LAYOUT_EDJ, "widget");
	evas_object_size_hint_weight_set(info->widget_layout, EVAS_HINT_EXPAND, EVAS_HINT_EXPAND);
	evas_object_size_hint_align_set(info->widget, 0.0, 0.5);
	evas_object_show(info->widget_layout);

	bg = evas_object_rectangle_add(evas_object_evas_get(s_info.layout));
	evas_object_resize(bg, w, h);
	evas_object_size_hint_min_set(bg, w, h);
	evas_object_color_set(bg, 255, 255, 255, 0);
	evas_object_show(bg);
	elm_object_part_content_set(info->widget_layout, "bg", bg);

	info->widget = widget_viewer_evas_add_widget(s_info.win, widget_id, info->content_info, info->period);
	if (!info->widget) {
		update_message("Failed to create a new widget");
		return WIDGET_ERROR_FAULT;
	}

	DbgPrint("Resize the widget(%s) to [%X] %dx%d\n", widget_id, info->size_types[0], w, h);

	evas_object_smart_callback_add(info->widget, WIDGET_SMART_SIGNAL_UPDATED, updated_cb, info);
	evas_object_smart_callback_add(info->widget, WIDGET_SMART_SIGNAL_EXTRA_INFO_UPDATED, extra_updated_cb, info);
	evas_object_smart_callback_add(info->widget, WIDGET_SMART_SIGNAL_PERIOD_CHANGED, period_changed_cb, info);
	evas_object_smart_callback_add(info->widget, WIDGET_SMART_SIGNAL_WIDGET_CREATED, widget_created_cb, info);
	evas_object_smart_callback_add(info->widget, WIDGET_SMART_SIGNAL_WIDGET_CREATE_ABORTED, widget_create_aborted_cb, info);

	evas_object_resize(info->widget, w, h);
	evas_object_size_hint_align_set(info->widget, 0.0, 0.5);
	evas_object_show(info->widget);
	elm_object_part_content_set(info->widget_layout, "widget", info->widget);

	elm_box_pack_end(s_info.box, info->widget_layout);
	elm_object_part_text_set(s_info.layout, "widget,id", widget_id);

	update_message(NULL);
	elm_box_padding_set(s_info.box, 0, 5);

	DbgPrint("Found valid size[%X]: %dx%d\n", info->size_types[i], w, h);
	evas_object_event_callback_add(info->widget, EVAS_CALLBACK_MOUSE_DOWN, layout_down_cb, info);
	evas_object_event_callback_add(info->widget, EVAS_CALLBACK_MOUSE_UP, layout_up_cb, NULL);

	s_info.widget_info_list = g_list_append(s_info.widget_info_list, info);

	return WIDGET_ERROR_NONE;
}

static int prepare_widget(const char *widget_id, app_control_h control, widget_info *info)
{
	int ret;
	bundle *b;

	if (app_control_export_as_bundle(control, &b) == APP_CONTROL_ERROR_NONE) {
		bundle_raw *r;
		int len;

		if (bundle_encode(b, &r, &len) == BUNDLE_ERROR_NONE) {
			info->content_info = strdup((const char *)r);
			if (!info->content_info)
				ErrPrint("strdup: %d\n", errno);
			else
				DbgPrint("Encoded content_info: [%s]\n", info->content_info);

			free((char *)r);
		} else {
			ErrPrint("Failed to encode a bundle\n");
		}

		bundle_free(b);
	}

	info->count_of_size_type = 20;
	ret = widget_service_get_supported_size_types(widget_id, &info->count_of_size_type, &info->size_types);
	if (ret != WIDGET_ERROR_NONE)
		ErrPrint("Failed to load an widget\n");

	DbgPrint("[%s] %d\n", widget_id, info->count_of_size_type);
	return ret;
}

static Eina_Bool lazy_widget_loader_cb(void *data)
{
	widget_info *info = (widget_info *)data;

	DbgPrint("Lazy loader expired. load widget [%s]\n", info->widget_id);
	(void)load_widget(info->widget_id, info);
	s_info.lazy_widget_loader = NULL;

	return ECORE_CALLBACK_CANCEL;
}

static void _run_widget(char *widget_id, app_control_h service, char *lazy_loader)
{
	widget_info *info = (widget_info *)calloc(1, sizeof(widget_info));

	if (info == NULL) {
		ErrPrint("Fail to create info");
		return;
	}
	info->widget_id = strdup(widget_id);
	if (info->widget_id == NULL) {
		ErrPrint("Fail to strdup widget_id");
		free(info);
		return;
	}

	info->period = WIDGET_VIEWER_EVAS_DEFAULT_PERIOD;
	info->count_of_size_type = 20;

	prepare_widget(widget_id, service, info);
	if (s_info.lazy_widget_loader) {
		char *tmp;
		tmp = ecore_timer_del(s_info.lazy_widget_loader);
		free(tmp);
	}

	if (lazy_loader != NULL && strcmp(lazy_loader, "true") == 0) {
		DbgPrint("Load a watch after some time later (%lf)\n", s_info.lazy_watch_time);
		s_info.lazy_widget_loader = ecore_timer_add(s_info.lazy_watch_time, lazy_widget_loader_cb, info);
		if (!s_info.lazy_widget_loader) {
			ErrPrint("Unable to fire the timer\n");
			lazy_widget_loader_cb(info);
		}
		/**
		 * @note
		 * widget_id will be deleted from lazy_widget_loader_cb. or return value of ecore_timer_del().
		 */
	} else {
		DbgPrint("Immediately loads the watch[%s]\n", widget_id);
		lazy_widget_loader_cb(info);
	}
}

static void __set_bundle_for_watchapp(bundle *app_control_b, bundle *b)
{
	char *sdk_util = NULL;
	char *value = NULL;
	char buf[32];

	bundle_get_str(app_control_b, AUL_K_DEBUG, &value);
	if (value) {
		bundle_add_str(b, AUL_K_DEBUG, value);
		value = NULL;
	}

	bundle_get_str(app_control_b, AUL_K_SEQ_NUM, &value);
	if (value) {
		bundle_add_str(b, AUL_K_SEQ_NUM, value);
		value = NULL;
	}

	bundle_get_str(app_control_b, AUL_K_WAIT_RESULT, &value);
	if (value) {
		bundle_add_str(b, AUL_K_WAIT_RESULT, value);
		value = NULL;
	}

	bundle_get_str(app_control_b, AUL_K_CALLER_PID, &value);
	if (value) {
		bundle_add_str(b, AUL_K_CALLER_PID, value);
		value = NULL;
	} else {
		snprintf(buf, sizeof(buf), "%d", getpid());
		bundle_add_str(b, AUL_K_CALLER_PID, buf);
	}

	bundle_get_str(app_control_b, AUL_K_CALLER_UID, &value);
	if (value) {
		bundle_add_str(b, AUL_K_CALLER_UID, value);
		value = NULL;
	} else {
		snprintf(buf, sizeof(buf), "%d", getuid());
		bundle_add_str(b, AUL_K_CALLER_UID, buf);
	}

	bundle_get_str(app_control_b, AUL_K_SDK, &sdk_util);
	if (sdk_util != NULL) {
		aul_debug_info_init();
		aul_debug_info_set(app_control_b, b);
		__debug_info_initialized = true;
		DbgPrint("util %s", sdk_util);
	}
}

static char *__get_pkgid_by_appid(char *appid)
{
	pkgmgrinfo_appinfo_h handle;
	int ret;
	char *pkgid = NULL;
	char *ret_pkgid = NULL;

	ret = pkgmgrinfo_appinfo_get_appinfo(appid, &handle);
	if (ret != PMINFO_R_OK)
		return NULL;

	ret = pkgmgrinfo_appinfo_get_pkgid(handle, &pkgid);
	if (ret != PMINFO_R_OK) {
		pkgmgrinfo_appinfo_destroy_appinfo(handle);
		return NULL;
	}

	ret_pkgid = strdup(pkgid);
	pkgmgrinfo_appinfo_destroy_appinfo(handle);

	return ret_pkgid;
}

static void _app_control(app_control_h service, void *data)
{
	char *widget_id = NULL;
	int ret;
	char *app_type = NULL;
	pkgmgrinfo_appinfo_h app_info;
	app_control_h watch_control = NULL;
	bundle *b;
	bundle *app_control_b;
	char *lazy_loader = NULL;
	bool is_array;
	int array_len;
	char **widget_id_arr;
	char *operation = NULL;
	int i;
	char *cur_watch = NULL;

	app_control_get_extra_data(service, WIDGET_APPID, &widget_id);
	if (widget_id) {
		DbgPrint("Loading a widget: [%s]\n", widget_id);
		ret = pkgmgrinfo_appinfo_get_appinfo(widget_id, &app_info);
		if (PMINFO_R_OK != ret)
			ErrPrint("fail to get %s appinfo %d", widget_id, ret);
		else
			pkgmgrinfo_appinfo_get_component_type(app_info, &app_type);
	}

	if (app_type != NULL && strcmp(app_type, "watchapp") == 0) {
		cur_watch = vconf_get_str(VCONFKEY_WMS_CLOCKS_SET_IDLE);
		if (widget_id != NULL) {
			if (__watch_pkgid) {
				free(__watch_pkgid);
				__watch_pkgid = NULL;
			}
			__watch_pkgid = __get_pkgid_by_appid(widget_id);

			if (cur_watch != NULL && __watch_pkgid != NULL &&
					strcmp(cur_watch, __watch_pkgid) == 0) {
				vconf_set_str(VCONFKEY_WMS_CLOCKS_SET_IDLE,
						DEFAULT_WATCH);
			}
		}

		evas_object_resize(s_info.win, s_info.w, s_info.h);
		watch_manager_init(s_info.win);
		evas_object_smart_callback_add(s_info.win, WATCH_SMART_SIGNAL_ADDED, __watch_added, NULL);
		evas_object_smart_callback_add(s_info.win, WATCH_SMART_SIGNAL_REMOVED, __watch_removed, NULL);

		watch_manager_get_app_control(widget_id, &watch_control);

		app_control_get_operation(service, &operation);
		if (operation) {
			app_control_set_operation(watch_control, operation);
			free(operation);
		}

		app_control_to_bundle(watch_control, &b);
		app_control_to_bundle(service, &app_control_b);

		__set_bundle_for_watchapp(app_control_b, b);

		ret = aul_forward_app(widget_id, b);
		DbgPrint("aul_forward_app returns [%d]", ret);
		app_control_destroy(watch_control);
	} else {
		ret = app_control_get_extra_data(service, LAZY_LOADER, &lazy_loader);
		app_control_to_bundle(service, &b);
		widget_service_set_sdk_util(b);
		ret = unload_widget();
		ret = app_control_is_extra_data_array(service, WIDGET_APPID, &is_array);
		if (is_array) {
			app_control_get_extra_data_array(service, WIDGET_APPID, &widget_id_arr, &array_len);
			for (i = 0; i < array_len; ++i)
				_run_widget(widget_id_arr[i], service, lazy_loader);

		} else if (widget_id != NULL) {
			_run_widget(widget_id, service, lazy_loader);
		} else {
			__resume_process();
			elm_win_activate(s_info.win);
		}
	}

	if (widget_id)
		free(widget_id);
}

int main(int argc, char *argv[])
{
	char err_buf[ERR_BUFFER_SIZE];
	ui_app_lifecycle_callback_s event_callback;

	event_callback.create = app_create;
	event_callback.terminate = app_terminate;
	event_callback.pause = app_pause;
	event_callback.resume = app_resume;
	event_callback.app_control = _app_control;

	if (setenv("BUFMGR_LOCK_TYPE", "once", 0) < 0)
		LOGE("setenv(BUFMGR_LOCK_TYPE) is failed: %s", strerror_r(errno, err_buf, sizeof(err_buf)));

	if (setenv("BUFMGR_MAP_CACHE", "true", 0) < 0)
		LOGE("setenv(BUFMGR_MAP_CACHE) is failed: %s", strerror_r(errno, err_buf, sizeof(err_buf)));

	return ui_app_main(argc, argv, &event_callback, NULL);
}

/* End of a file */
